{===============================================================}
{    Oxygen Plan Library, v1.3  1999-2001 Oxygen Software      }
{---------------------------------------------------------------}
{        Written by Oleg Fyodorov, delphi@oxygensoftware.com    }
{                http://www.oxygensoftware.com                  }
{===============================================================}

unit O2Plan;

{$Include O2Plan.inc}
{$OPTIMIZATION OFF}

interface

  uses
    Windows, Classes, Controls, Graphics, StrF, ExtCtrls, SysUtils, PlanUtils, {$IFDEF USE_RX} TimerLst, RxGif, VCLUtils,{$ENDIF}
    jpeg, PlanObjects, Printers, ShellApi, Dialogs;

  resourcestring
    // Plan commands
    cmScale             = 'Scale ';
    cmSetPenColor       = 'SetPenColor ';
    cmSetBrushColor     = 'SetBrushColor ';
    cmSetLineWidth      = 'SetLineWidth ';
    cmSetFont           = 'SetFont ';
    cmStandardPicture   = 'StandardPicture ';
    cmStandardAniGif    = 'StandardAniGif ';
    cmBackgroundPicture = 'BackgroundPicture ';
    cmLayer             = 'NewLayer ';

    sDefaultLayer       = 'Default';

    sFileNotFound       = 'File %s not found';
    sInvalidOwner       = 'Plan object owner is not TOxygenPlan';
    sUnknownPlanObject  = 'Unknown plan object: %s';

  type

  TOxygenPlan = class;

  EOxygenPlanException = class(Exception);

  TPlanStandardPicture = class(TPicture)
    public
      Name : String;
      function GetDescrString:String;
  end;

{$IFDEF USE_RX}
  TPlanStandardAniGif = class(TGifImage)
    public
      Name : String;
      function GetDescrString:String;
  end;
{$ENDIF}

  TPlanOption = (poAutoAnimate, poSavePictures, poSaveStandardPictures, poSaveBackgroundPicture, poChildrenClipping, poPrintProportional);
  TPlanOptions = set of TPlanOption;

  TPlanPrintMethod = (ppMetafile, ppBitmap);

  TPrintMargins = class
    pmLeft,
    pmTop,
    pmRight,
    pmBottom : Integer;
  end;
  TObjectClickEvent    = procedure (Plan : TOxygenPlan; PlanObject : TPlanObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer) of object;
  TPlanNotifyEvent = procedure (Plan : TOxygenPlan; PlanObject : TPlanObject) of object;
  TParseProgressEvent  = procedure (Plan : TOxygenPlan; const PercentsCompleted : Integer) of object;
  TNeedStdPictureEvent = function  (Plan : TOxygenPlan; const PictureName : String) : Boolean of object;
  TNeedPictureEvent    = procedure  (Plan : TOxygenPlan; PlanPicture : TPlanPicture) of object;
{$IFDEF USE_RX}
  TNeedStdAniGifEvent  = function  (Plan : TOxygenPlan; const AniGifName : String) : Boolean of object;
  TNeedAniGifEvent     = procedure  (Plan : TOxygenPlan; PlanAniGif : TPlanAniGif) of object;
{$ENDIF}

  TSelectedObjectList = class
    private
      FList : TList;
      function  Get(Index : Integer) : TPlanObject;
      procedure Put(Index : Integer; Item : TPlanObject);
      function GetCount : Integer;
    public
      property Count : Integer read GetCount;
      property Items[Index : Integer] : TPlanObject read Get write Put; default;
      constructor Create;
      procedure Free;
      function Add(Item : TPlanObject) : Integer;
      function IndexOf(Item : TPlanObject) : Integer;
      procedure Delete(Index : Integer);
      procedure DeleteObject(Item : TPlanObject);
      procedure EditMove(APoint : TPoint);
      procedure Clear;
  end;

  TOxygenPlan = class(TCustomPanel)
    private
      FScaleX, FScaleY      : Integer;
      FCurX,FCurY           : Integer;
      FStartCoord           : TPoint;
      FActiveObject         : TPlanObject;
      FActiveMarginType     : TMarginType;
      FBackgroundPicture    : TPicture;
      FReadOnly             : Boolean;
      FPlanStrings          : TStrings;
      FNewClickNum          : Integer;    // click number when drawing new object
      FSelectColor          : TColor;     // small selection boxes color
      FCurLayer             : TPlanLayer; // current layer
      FCurDrawingObject     : TPlanObject; // current drawing object
      FCurDrawingPointNum   : Integer;     // current drawing object's point number
      FPen                  : TPen;
      FBrush                : TBrush;
      NewObjUM, OldObjUM    : TPlanObject;
      FDisableRepaintCount  : Integer;
      FReadyToMoving        : Boolean;
      FTransparent          : Boolean;
      FPlanPrintMethod      : TPlanPrintMethod;

{$IFDEF USE_RX}
      FTimerList            : TRxTimerList;
      FStandardAniGifsList  : TList;
{$ENDIF}
      FStandardPicturesList : TList;
      FOptions              : TPlanOptions;
      FPrinting             : Boolean;
      FPrintMargins         : TPrintMargins;
      FSelectedObjects      : TSelectedObjectList;
      FWasMove              : Boolean;
{$IFDEF VER100}
      FPaintBitmap          : TBitmap;
      FDoubleBuffered       : Boolean; 
{$ENDIF}

      FObjectClick          : TObjectClickEvent;
      FObjectDblClick       : TPlanNotifyEvent;
      FObjectMouseMove      : TPlanNotifyEvent;
      FObjectMouseEnter     : TPlanNotifyEvent;
      FObjectMouseLeave     : TPlanNotifyEvent;
      FAfterDrawNewObject   : TPlanNotifyEvent;
      FBeforeDrawNewObject  : TPlanNotifyEvent;
      FParseProgress        : TParseProgressEvent;
      FAfterPaint           : TNotifyEvent;
      FNeedStdPicture       : TNeedStdPictureEvent;
      FNeedPicture          : TNeedPictureEvent;
      FObjectBlink          : TPlanNotifyEvent;
{$IFDEF USE_RX}
      FNeedAniGif           : TNeedAniGifEvent;
      FNeedStdAniGif        : TNeedStdAniGifEvent;
{$ENDIF}

      function  GetPlanStrings:TStrings;
      procedure SetPlanStrings(Value:TStrings);
      procedure SetBackgroundPicture(Value:TPicture);
      procedure SetScaleX(Value:Integer);
      procedure SetScaleY(Value:Integer);
      procedure SetReadOnly(Value:Boolean);
      function  GetBackgroundPictureType:String;
      procedure SetCurLayer(Value:TPlanLayer);
      function  GetPicture(Index:Integer):TPlanStandardPicture;
      procedure ClearPictures;
{$IFDEF USE_RX}
      function  GetAniGif(Index:Integer):TPlanStandardAniGif;
      procedure ClearAniGifs;
      function  GetAniGifCount : Integer;
      procedure StdAniGifChange(Sender : TObject);
{$ENDIF}
      function  GetPictureCount : Integer;
      procedure SetActiveObject(PlanObject:TPlanObject);
      procedure ReadStringData(Reader: TReader);
      procedure StdPictureChange(Sender : TObject);
      procedure SetOptions(Value : TPlanOptions);
      function  GetObjectCount : Integer;
      procedure SetBrush(Value : TBrush);
      procedure SetPen(Value : TPen);
      procedure SetTransparent(const Value : Boolean);

    protected
      procedure DoObjectClick(PlanObject : TPlanObject; Button : TMouseButton; Shift : TShiftState; X, Y : Integer); virtual;
      procedure DoObjectDblClick(PlanObject : TPlanObject); virtual;
      procedure DoObjectMouseMove(PlanObject : TPlanObject); virtual;
      procedure DoObjectMouseEnter(PlanObject : TPlanObject); virtual;
      procedure DoObjectMouseLeave(PlanObject : TPlanObject); virtual;
      procedure DoAfterDrawNewObject(PlanObject : TPlanObject); virtual;
      procedure DoBeforeDrawNewObject(PlanObject : TPlanObject); virtual;
      procedure DoParseProgress(const PercentsCompleted : Integer); virtual;
      procedure DoAfterPaint; virtual;
      function  DoNeedStdPicture(const PictureName : String) : Boolean; virtual;
      procedure DoNeedPicture(PlanPicture : TPlanPicture); virtual;
{$IFDEF USE_RX}
      function  DoNeedStdAniGif(const AniGifName : String) : Boolean; virtual;
      procedure DoNeedAniGif(PlanAniGif : TPlanAniGif); virtual;
{$ENDIF}
      procedure DefineProperties(Filer: TFiler); override;
      procedure Loaded; override;

      procedure Resize; override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
      procedure DblClick; override;
      procedure Paint; override;

    public
      Layers : TPlanLayers;

      property ActiveMarginType: TMarginType read FActiveMarginType; // 0.95
      property ActiveObject:TPlanObject read FActiveObject write SetActiveObject; // 0.95
{$IFDEF USE_RX}
      property AniGifs[Index : Integer] : TPlanStandardAniGif read GetAniGif;
      property AniGifCount : Integer read GetAniGifCount;
{$ENDIF}
      property BackgroundPictureType : String read GetBackgroundPictureType;
      property Canvas;
      property CurDrawingPointNum : Integer read FCurDrawingPointNum; // 0.95
      property CurrentLayer : TPlanLayer read FCurLayer write SetCurLayer stored False;
      property DisableRepaintCount : Integer read FDisableRepaintCount; // 0.95
      property Printing : Boolean read FPrinting write FPrinting; // 0.96
      property ObjectCount : Integer read GetObjectCount;
      property Pictures[Index : Integer] : TPlanStandardPicture read GetPicture;
      property PictureCount : Integer read GetPictureCount;
      property SelectedObjects : TSelectedObjectList read FSelectedObjects; // 1.01
      property StartCoord  : TPoint read FStartCoord; // 0.95

{$IFDEF USE_RX}
      function    AddTimerEvent(AOnTimer: TNotifyEvent; AInterval: Longint; ACycled: Boolean): Longint; // 0.95
      procedure   DeleteTimerEvent(AHandle: Longint); // 0.95
      function    TimerByHandle(AHandle: Longint): TRxTimerEvent; // 0.95

      function    AddAniGif(const AniGifName : String; AniGif : TGIFImage):Integer;
      function    AniGifIndex(const AniGifName:String):Integer;
      procedure   DeleteAniGif(Index:Integer);
      procedure   LoadStdAniGif(const AniGifName : String; PlanAniGif : TPlanAniGif);
      procedure   LoadAniGif(PlanAniGif : TPlanAniGif);
{$ENDIF}
      function    AddPicture(const PictureName : String; Picture : TPicture):Integer;
      procedure   DeletePicture(Index : Integer);
      procedure   LoadStdPicture(const PictureName : String; PlanPicture : TPlanPicture);
      procedure   LoadPicture(PlanPicture : TPlanPicture);
      function    PictureIndex(const PictureName:String):Integer;

      procedure   Assign(Source : TPersistent); override;
      constructor Create(AOwner : TComponent); override;
      procedure   CancelDrawNewObject;
      procedure   Clear;
      procedure   EnableRepaint;
      procedure   DisableRepaint;
      procedure   DoObjectBlink(PlanObject : TPlanObject); virtual;
      procedure   ExportToMetafile(const FileName : String); // 0.91
      procedure   ExportToBitmap(const FileName : String);  // 0.91
      procedure   ExportToJpeg(const FileName : String);  // 0.91
{$IFDEF USE_RX}
      procedure   ExportToGif(const FileName : String);  // 0.91
{$ENDIF}
      procedure   DrawNewObject(PlanObjectClass : TPlanObjectClass);
      function    FindObject(ObjectClass : TPlanObjectClass; const ObjectName : String; const ObjectID : Integer):TPlanObject;
      function    GetBitmap : TBitmap; // 0.96
      function    GetMetafile : TMetafile; // 0.96
      procedure   Initialize; virtual; // 0.95
      procedure   LoadFromFile(FileName : TFileName);
      procedure   LoadFromStream(Stream : TStream);
      function    ObjectByID(ObjectClass:TPlanObjectClass; const ObjectID:Integer):TPlanObject;
      function    ObjectByName(ObjectClass : TPlanObjectClass; const ObjectName : String):TPlanObject;
      function    ObjectByCustomProperty(ObjectClass : TPlanObjectClass; const PropertyName, PropertyValue : String) : TPlanObject;
      function    ObjectUnderMouse(var MarginType : TMarginType) : TPlanObject;
      procedure   PaintToCanvas(ACanvas : TCanvas; const CWidth, CHeight : Integer); // 0.96
      procedure   PaintToDesktop;
      procedure   Parse(PlanData : TStrings);
      procedure   Print(APrinter : TPrinter);
      class procedure RegisterPlanObjectClass(const ACmdString : String; PlanObjectClass : TPlanObjectClass); // 0.95
      procedure   SaveTo(PlanData : TStrings);
      procedure   SaveToFile(FileName:String);
      procedure   SaveToStream(Stream:TStream);
      destructor  Destroy; override;

    published
      property Align;
      //property Alignment;
      // Start animating TPlanAniGif and TPlanAniPicture automatically
      property BackgroundPicture     : TPicture read FBackgroundPicture write SetBackgroundPicture;
      property BevelInner;
      property BevelOuter;
      property BevelWidth;
      property BorderWidth;
      property BorderStyle;
      property Brush                 : TBrush read FBrush write SetBrush;
      {$IFNDEF VER100}
      property Constraints;
      {$ENDIF}
      property DragCursor;
      property DragMode;
      {$IFNDEF VER100}
      property DragKind;
      property DockSite;
      {$ENDIF}
      property DoubleBuffered {$IFDEF VER100} : Boolean read FDoubleBuffered write FDoubleBuffered {$ENDIF};

      property Enabled;
      property Color;
      property Ctl3D;
      property Font;
      property Locked;
      property Options               : TPlanOptions read FOptions write SetOptions;
      property ParentColor;
      property ParentCtl3D;
      property ParentFont;
      property ParentShowHint;
      property Pen                   : TPen read FPen write SetPen;
      property PlanPrintMethod       : TPlanPrintMethod read FPlanPrintMethod write FPlanPrintMethod;
      property PlanStrings           : TStrings read GetPlanStrings write SetPlanStrings;
      property PopupMenu;
      property PrintMargins          : TPrintMargins read FPrintMargins write FPrintMargins;
      property ReadOnly              : Boolean read FReadOnly write SetReadOnly;
      property SelectColor           : TColor read FSelectColor write FSelectColor;
      property ScaleX                : Integer read FScaleX write SetScaleX;
      property ScaleY                : Integer read FScaleY write SetScaleY;
      property ShowHint;
      property TabOrder;
      property TabStop;
      property Transparent           : Boolean read FTransparent write SetTransparent;
      property Visible;
      {$IFNDEF VER100}
      property UseDockManager default True;
      {$ENDIF}

      // events
      property OnClick;
      property OnDblClick;
      {$IFNDEF VER100}
      property OnCanResize;
      property OnConstrainedResize;
      property OnDockDrop;
      property OnDockOver;
      property OnStartDock;
      property OnEndDock;
      property OnUnDock;
      property OnGetSiteInfo;
      {$ENDIF}
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnEnter;
      property OnExit;
      property OnKeyDown;
      property OnKeyPress;
      property OnKeyUp;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      {$IFNDEF VER100}
      property OnMouseWheel;
      property OnMouseWheelDown;
      property OnMouseWheelUp;
      {$ENDIF}
      property OnResize;
      property OnStartDrag;

      property OnAfterPaint         : TNotifyEvent read FAfterPaint write FAfterPaint;
      property OnAfterDrawNewObject : TPlanNotifyEvent read FAfterDrawNewObject write FAfterDrawNewObject;
      property OnBeforeDrawNewObject: TPlanNotifyEvent read FBeforeDrawNewObject write FBeforeDrawNewObject;
      property OnNeedStdPicture     : TNeedStdPictureEvent read FNeedStdPicture write FNeedStdPicture;
      property OnNeedPicture        : TNeedPictureEvent read FNeedPicture write FNeedPicture;
{$IFDEF USE_RX}
      property OnNeedStdAniGif      : TNeedStdAniGifEvent read FNeedStdAniGif write FNeedStdAniGif;
      property OnNeedAniGif         : TNeedAniGifEvent read FNeedAniGif write FNeedAniGif;
{$ENDIF}
      property OnObjectBlink        : TPlanNotifyEvent read FObjectBlink write FObjectBlink;
      property OnObjectClick        : TObjectClickEvent read FObjectClick write FObjectClick;
      property OnObjectDblClick     : TPlanNotifyEvent read FObjectDblClick write FObjectDblClick;
      property OnObjectMouseMove    : TPlanNotifyEvent read FObjectMouseMove write FObjectMouseMove;
      property OnObjectMouseEnter   : TPlanNotifyEvent read FObjectMouseEnter write FObjectMouseEnter;
      property OnObjectMouseLeave   : TPlanNotifyEvent read FObjectMouseLeave write FObjectMouseLeave;
      property OnParseProgress      : TParseProgressEvent read FParseProgress write FParseProgress;
  end;

  procedure RenameStdPicture(Plan : TOxygenPlan; const OldName,NewName : String);
  function  StandardPictureReferenceCount(Plan : TOxygenPlan; const StdPictureName : String) : Integer;
{$IFDEF USE_RX}
  procedure RenameStdAniGif(Plan : TOxygenPlan; const OldName,NewName : String);
  function  StandardAniGifReferenceCount(Plan : TOxygenPlan; const StdAniGifName : String) : Integer;
{$ENDIF}

procedure Register;

implementation

type
  PPlanObjectInfo = ^TPlanObjectInfo;
  TPlanObjectInfo = record
    PlanObjectClass : TPlanObjectClass;
    CmdString : string;
  end;

  TPlanObjectInfoList = class(TList)
    public
      constructor Create;
      destructor Destroy; override;
      procedure Add(const ACmdString : String; AClass: TPlanObjectClass);
      function FindCmdString(ACmdString : String): TPlanObjectClass;
      {function FindClassName(const ClassName : String): TPlanObjectClass;
      procedure Remove(AClass: TPlanObjectClass);}
  end;

// TPlanObjectInfoList
constructor TPlanObjectInfoList.Create;
begin
  inherited Create;
  Add(ocLine, TPlanLine);
  Add(ocBox, TPlanBox);
  Add(ocCircle, TPlanCircle);
  Add(ocEllipse, TPlanEllipse);
  Add(ocText, TPlanText);
  Add(ocPicture, TPlanPicture);
  Add(ocPolygon, TPlanPolygon);
  Add(ocPolyLine, TPlanPolyLine);
  Add(ocTraceLine, TPlanTraceLine);
  Add(ocCompositeObject, TPlanCompositeObject);
{$IFDEF USE_RX}
  Add(ocAniGif, TPlanAniGif);
  Add(ocAniPicture, TPlanAniPicture);
{$ENDIF}  
end;

destructor TPlanObjectInfoList.Destroy;
  var  i : Integer;
begin
  for i := 0 to Count - 1 do Dispose(PPlanObjectInfo(Items[i]));
  inherited Destroy;
end;

procedure TPlanObjectInfoList.Add(const ACmdString : String; AClass: TPlanObjectClass);
  var NewRec: PPlanObjectInfo;
begin
  New(NewRec);
  with NewRec^ do begin
    CmdString := AnsiUpperCase(ACmdString);
    PlanObjectClass := AClass;
  end;
  inherited Add(NewRec);
end;

function TPlanObjectInfoList.FindCmdString(ACmdString : String): TPlanObjectClass;
 var i : Integer;
     s : String;
begin
  Result := nil;
  s := AnsiUpperCase(ACmdString);
  for i := Count - 1 downto 0 do
    with PPlanObjectInfo(Items[i])^ do if (CmdString = s) then begin
      Result := PlanObjectClass;
      Exit;
    end;
end;

{function TPlanObjectInfoList.FindClassName(const ClassName : String): TPlanObjectClass;
  var i : Integer;
begin
  for i := Count - 1 downto 0 do begin
    Result := PPlanObjectInfo(Items[i])^.PlanObjectClass;
    if (Result.ClassName = ClassName) then Exit;
  end;
  Result := nil;
end;

procedure TPlanObjectInfoList.Remove(AClass: TPlanObjectClass);
  var i : Integer;
      P : PPlanObjectInfo;
begin
  for i := Count - 1 downto 0 do begin
    P := PPlanObjectInfo(Items[i]);
    if P^.PlanObjectClass.InheritsFrom(AClass) then begin
      Dispose(P);
      Delete(i);
    end;
  end;
end;}

  var PlanObjectsInfo: TPlanObjectInfoList = nil;

// general functions
procedure RenameStdPicture(Plan : TOxygenPlan; const OldName,NewName : String);
  var l,i : Integer;
      pp : TPlanPicture;
begin
  for l := 1 to Plan.Layers.Count do
    for i := 1 to Plan.Layers[l-1].Count do
      if (Plan.Layers[l-1].Objects[i-1] is TPlanPicture) then begin
        pp := (Plan.Layers[l-1].Objects[i-1] as TPlanPicture);
        if (AnsiUpperCase(pp.StdPictureName) = AnsiUpperCase(OldName)) then pp.StdPictureName := NewName;
      end;
end;

function StandardPictureReferenceCount(Plan : TOxygenPlan; const StdPictureName : String) : Integer;
  var l,i : Integer;
      pp : TPlanPicture;
begin
  Result := 0;
  for l := 1 to Plan.Layers.Count do
    for i := 1 to Plan.Layers[l-1].Count do
      if (Plan.Layers[l-1].Objects[i-1] is TPlanPicture) then begin
        pp := (Plan.Layers[l-1].Objects[i-1] as TPlanPicture);
        if (AnsiUpperCase(pp.StdPictureName) = AnsiUpperCase(StdPictureName)) then Inc(Result);
      end;

end;

{$IFDEF USE_RX}
procedure RenameStdAniGif(Plan : TOxygenPlan; const OldName,NewName : String);
  var l,i : Integer;
      pp : TPlanAniGif;
begin
  for l := 1 to Plan.Layers.Count do
    for i := 1 to Plan.Layers[l-1].Count do
      if (Plan.Layers[l-1].Objects[i-1] is TPlanAniGif) then begin
        pp := (Plan.Layers[l-1].Objects[i-1] as TPlanAniGif);
        if (AnsiUpperCase(pp.StdAniGifName) = AnsiUpperCase(OldName)) then pp.StdAniGifName := NewName;
      end;
end;

function StandardAniGifReferenceCount(Plan : TOxygenPlan; const StdAniGifName : String) : Integer;
  var l,i : Integer;
      pp : TPlanAniGif;
begin
  Result := 0;
  for l := 1 to Plan.Layers.Count do
    for i := 1 to Plan.Layers[l-1].Count do
      if (Plan.Layers[l-1].Objects[i-1] is TPlanAniGif) then begin
        pp := (Plan.Layers[l-1].Objects[i-1] as TPlanAniGif);
        if (AnsiUpperCase(pp.StdAniGifName) = AnsiUpperCase(StdAniGifName)) then Inc(Result);
      end;

end;
{$ENDIF}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanStandardPicture
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TPlanStandardPicture.GetDescrString:String;
  var ms  : TMemoryStream;
      s   : String;
      pic : TPicture;
begin
  Result:='';
  if ((Graphic=nil) or Graphic.Empty) then Exit;
  pic:=TPicture.Create;
  pic.Assign(Self);
  Result:=cmStandardPicture+Name+',';
  ms:=TMemoryStream.Create;
  Graphic.SaveToStream(ms);
  s:=GetPictureType(pic);
  Result:=Result+s+','+IntToStr(Integer(Graphic.Transparent))+','+Stream2String(ms);
  ms.Free;
  Self.Assign(pic);
  pic.Free;
end;


{$IFDEF USE_RX}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanStandardAniGif
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TPlanStandardAniGif.GetDescrString:String;
  var ms:TMemoryStream;
      s:String;
      gif:TGifImage;
begin
  Result:='';
  if Empty then Exit;
  gif:=TGifImage.Create;
  gif.Assign(Self);
  Result:=cmStandardAniGif+Name+',';
  ms:=TMemoryStream.Create;
  gif.SaveToStream(ms);
  Result:=Result+s+Stream2String(ms);
  ms.Free;
  gif.Free;
end;
{$ENDIF}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//    TSelectedObjectList
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TSelectedObjectList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

procedure TSelectedObjectList.Free;
begin
  Clear;
  FList.Free;
  inherited Free;
end;

function TSelectedObjectList.GetCount : Integer;
begin
  Result := FList.Count;
end;

function TSelectedObjectList.Get(Index : Integer) : TPlanObject;
begin
  Result := TPlanObject(FList[Index]);
end;

procedure TSelectedObjectList.Put(Index : Integer; Item : TPlanObject);
begin
  FList[Index] := Item;
end;

function TSelectedObjectList.Add(Item : TPlanObject) : Integer;
begin
  if (IndexOf(Item) = -1) then Result := FList.Add(Item) else Result := -1;
end;

function TSelectedObjectList.IndexOf(Item : TPlanObject) : Integer;
begin
  Result := FList.IndexOf(Item);
end;

procedure TSelectedObjectList.Delete(Index : Integer);
  var po : TPlanObject;
begin
  try
    po := Items[Index];
    FList.Delete(Index);
    if (po <> nil) then if po.Selected then po.UnSelect;
  except end;
end;

procedure TSelectedObjectList.DeleteObject(Item : TPlanObject);
  var i : Integer;
begin
  i := IndexOf(Item);
  if (i >= 0) then FList.Delete(i);
end;

procedure TSelectedObjectList.EditMove(APoint : TPoint);
  var i : Integer;
begin
  for i := 1 to Count do Items[i-1].EditMove(APoint);
end;

procedure TSelectedObjectList.Clear;
begin
  while (Count > 0) do Delete(0);
  FList.Clear;
end;


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//    TOxygenPlan
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TOxygenPlan.Create(AOwner: TComponent);
begin
  Layers := TPlanLayers.Create(Self);
  FPrintMargins:=TPrintMargins.Create;
  FCurLayer := Layers[Layers.Add(sDefaultLayer)];
  FPen:=TPen.Create;
  FBrush:=TBrush.Create;
  Pen.Color := clWhite;
  Pen.Style := psSolid;
  Pen.Width := 1;
  Brush.Color := clBlack;
  Brush.Style := bsSolid;
  inherited Create(AOwner);
  inherited Caption := '';
  inherited Color := clBlack;
{$IFDEF VER100}
  FPaintBitmap := TBitmap.Create;
{$ENDIF}
  Font.Color := clWhite;
  DoubleBuffered := True;
  FCurDrawingObject := nil;
  FCurDrawingPointNum := 0;
  FBackgroundPicture:=TPicture.Create;
  FCurX:=0;
  FCurY:=0;
  FActiveObject:=nil;
  FActiveMarginType:=mtNone;
  FObjectClick:=nil;
  FObjectDblClick:=nil;
  FObjectMouseMove:=nil;
  FObjectMouseEnter:=nil;
  FObjectMouseLeave:=nil;
  FAfterDrawNewObject := nil;
  FBeforeDrawNewObject := nil;
  FParseProgress := nil;
  FAfterPaint := nil;
  FNeedStdPicture := nil;
  FNeedPicture := nil;
  FObjectBlink := nil;
{$IFDEF USE_RX}
  FNeedStdAniGif := nil;
  FNeedAniGif := nil;
{$ENDIF}
  FReadOnly:=False;
  FNewClickNum:=-1;
  FPlanStrings:=TStringList.Create;
  FSelectColor:=clSilver;
  NewObjUM:=nil;
  OldObjUM:=nil;
  FDisableRepaintCount:=0;
{$IFDEF USE_RX}
  FTimerList:=TRxTimerList.Create(Self);
  FStandardAniGifsList:=TList.Create;
{$ENDIF}
  FStandardPicturesList:=TList.Create;
  FOptions := [poAutoAnimate];
  FReadyToMoving := False;
  FTransparent := False;
  FPrinting := False;
  FOptions := [poAutoAnimate, poSavePictures, poSaveStandardPictures];
  with FPrintMargins do begin
    pmLeft := 50;
    pmTop := 50;
    pmRight := 50;
    pmBottom := 50;
  end;
  FSelectedObjects := TSelectedObjectList.Create;
  FWasMove := False;
  FPlanPrintMethod := ppMetafile;
{$IFDEF O2_SW}
  if (MessageDlg('This version of Oxygen Plan Library is NOT REGISTERED. '+#13#10+
                 'Press Ok to visit http://www.oxygensoftware.com and register.',
                 mtWarning,[mbOk,mbCancel],0) = mrOk) then ShellExecute(0,'open','http://www.oxygensoftware.com',nil,nil,SW_SHOWNORMAL);
{$ENDIF}
end;

destructor TOxygenPlan.Destroy;
begin
  Destroying;
FPrintMargins.Free;
  Layers.Clear;
  FSelectedObjects.Free;
  //Layers.Free;
  FPen.Free;
  FBrush.Free;
  FPlanStrings.Free;
  FBackgroundPicture.Free;
{$IFDEF USE_RX}
  FTimerList.Free;
  ClearAniGifs;
  FStandardAniGifsList.Free;
{$ENDIF}
  ClearPictures;
  FStandardPicturesList.Free;
{$IFDEF VER100}
  FPaintBitmap.Free;
{$ENDIF}
  inherited Destroy;
end;

{$IFDEF USE_RX}
function TOxygenPlan.AddTimerEvent(AOnTimer: TNotifyEvent; AInterval: Longint; ACycled: Boolean): Longint;
begin
  Result := FTimerList.Add(AOnTimer, AInterval, ACycled);
  FTimerList.Activate;
end;

procedure TOxygenPlan.DeleteTimerEvent(AHandle: Longint);
begin
  FTimerList.Delete(AHandle)
end;

function TOxygenPlan.TimerByHandle(AHandle: Longint): TRxTimerEvent;
begin
  Result := FTimerList.ItemByHandle(AHandle);
end;
{$ENDIF}

function TOxygenPlan.AddPicture(const PictureName:String; Picture:TPicture):Integer;
  var pp:TPlanStandardPicture;
begin
  if (PictureIndex(PictureName) >= 0) then raise EOxygenPlanException.Create('Standard Picture "' + PictureName + '" already exists');
  pp:=TPlanStandardPicture.Create;
  pp.Assign(Picture);
  if (Picture.Graphic <> nil) and Picture.Graphic.Transparent then begin
    pp.Graphic.Transparent := True;
    if (pp.Graphic is TBitmap) and (pp.Bitmap <> nil) then pp.Bitmap.TransparentColor := pp.Bitmap.Canvas.Pixels[0,0];
  end;
  pp.Name:=PictureName;
  pp.OnChange := StdPictureChange;
  Result:=FStandardPicturesList.Add(pp);
end;

{$IFDEF USE_RX}
function TOxygenPlan.AddAniGif(const AniGifName:String; AniGif:TGIFImage):Integer;
  var pag:TPlanStandardAniGif;
begin
  if (AniGifIndex(AniGifName) >= 0) then raise EOxygenPlanException.Create('Standard AniGif "' + AniGifName + '" already exists');
  pag:=TPlanStandardAniGif.Create;
  pag.Assign(AniGif);
  pag.Name:=AniGifName;
  pag.OnChange := StdAniGifChange;
  Result:=FStandardAniGifsList.Add(pag);
end;

function TOxygenPlan.AniGifIndex(const AniGifName:String):Integer;
  var i:Integer;
begin
  Result:=-1;
  for i:=1 to FStandardAniGifsList.Count do
    if (AnsiUpperCase(TPlanStandardAniGif(FStandardAniGifsList[i-1]).Name)=AnsiUpperCase(AniGifName)) then begin
      Result:=i-1;
      Exit;
    end;
end;
{$ENDIF}

procedure TOxygenPlan.Assign(Source : TPersistent);
  var src:TOxygenPlan;
      i:Integer;
      sp : TPlanStandardPicture;
{$IFDEF USE_RX}
      sa : TPlanStandardAniGif;
{$ENDIF}
      sl : TStringList;
      po : TPlanOptions;
begin
  DisableRepaint;
  sl := TStringList.Create;
  if (Source is TOxygenPlan) then try
    src := (Source as TOxygenPlan);
    Clear;
    FBackgroundPicture.Assign(src.BackgroundPicture);
    for i:=1 to src.PictureCount do begin
      sp := src.Pictures[i-1];
      AddPicture(sp.Name,sp);
    end;
{$IFDEF USE_RX}
    for i:=1 to src.AniGifCount do begin
      sa := src.AniGifs[i-1];
      AddAniGif(sa.Name,sa);
    end;
{$ENDIF}
    DoubleBuffered := src.DoubleBuffered;
    ScaleX := src.ScaleX;
    ScaleY := src.ScaleY;
    Options := src.Options;
    Brush.Assign(src.Brush);
    Pen.Assign(src.Pen);
    Font.Assign(src.Font);
    SelectColor := src.SelectColor;
    po := src.Options;
    src.Options := [poSavePictures];
    src.SaveTo(sl);
    src.Options := po;
    Parse(sl);
    try CurrentLayer := Layers.LayerByName(src.CurrentLayer.Name) except end;
  except end;
  EnableRepaint;
  sl.Free;
end;

procedure TOxygenPlan.ReadStringData(Reader: TReader);
begin
  Reader.ReadString;
end;

procedure TOxygenPlan.StdPictureChange(Sender : TObject);
  var l,i : Integer;
      pp : TPlanPicture;
      sp : TPlanStandardPicture;
begin
  sp := (Sender as TPlanStandardPicture);
  DisableRepaint;
  try
    for l := 1 to Layers.Count do
      for i := 1 to Layers[l-1].Count do
        if (Layers[l-1].Objects[i-1] is TPlanPicture) then begin
          pp := (Layers[l-1].Objects[i-1] as TPlanPicture);
          if (AnsiUpperCase(pp.StdPictureName) = AnsiUpperCase(sp.Name)) then pp.Picture.Assign(sp);
        end;
  finally
    EnableRepaint;
  end;
end;

{$IFDEF USE_RX}
procedure TOxygenPlan.StdAniGifChange(Sender : TObject);
  var l,i : Integer;
      pp : TPlanAniGif;
      sp : TPlanStandardAniGif;
begin
  sp := (Sender as TPlanStandardAniGif);
  DisableRepaint;
  try
    for l := 1 to Layers.Count do
      for i := 1 to Layers[l-1].Count do
        if (Layers[l-1].Objects[i-1] is TPlanAniGif) then begin
          pp := (Layers[l-1].Objects[i-1] as TPlanAniGif);
          if (AnsiUpperCase(pp.StdAniGifName) = AnsiUpperCase(sp.Name)) then pp.GIFImage.Assign(sp);
        end;
  finally
    EnableRepaint;
  end;
end;
{$ENDIF}

procedure TOxygenPlan.SetOptions(Value : TPlanOptions);
begin
  if (Value <> FOptions) then FOptions := Value;
  Repaint;
end;

function TOxygenPlan.GetObjectCount : Integer;
  var i : Integer;
begin
  Result := 0;
  for i := 1 to Layers.Count do Inc(Result, Layers[i-1].Count);
end;

procedure TOxygenPlan.SetBrush(Value : TBrush);
begin
  FBrush.Assign(Value);
end;

procedure TOxygenPlan.SetPen(Value : TPen);
begin
  FPen.Assign(Value);
end;

procedure TOxygenPlan.SetTransparent(const Value : Boolean);
begin
  if (Value <> FTransparent) then begin
    FTransparent := Value;
    Invalidate;
  end;
end;

procedure TOxygenPlan.DefineProperties(Filer: TFiler);
begin
  if Filer is TReader then inherited DefineProperties(Filer);
  Filer.DefineProperty('PlanStrings', ReadStringData, nil, False);
end;

procedure TOxygenPlan.Loaded;
begin
  inherited Loaded;
  Parse(FPlanStrings);
end;

procedure TOxygenPlan.SetActiveObject(PlanObject:TPlanObject);
 { var i:Integer;
      po:TPlanObject;}
begin
  FActiveObject := PlanObject;
  CurrentLayer.SetTopObject(PlanObject)
  {i:=CurrentLayer.IndexOf(PlanObject);
  if (i=-1) then Exit;
  po:=CurrentLayer.Objects[i];
  CurrentLayer.ObjectList.Delete(i);
  CurrentLayer.ObjectList.Add(po);}
end;

function TOxygenPlan.GetPlanStrings:TStrings;
begin
  SaveTo(FPlanStrings);
  Result:=FPlanStrings;
end;

procedure TOxygenPlan.SetPlanStrings(Value:TStrings);
begin
  FPlanStrings.Clear;
  FPlanStrings.AddStrings(Value);
  Parse(FPlanStrings);
  Repaint;
end;

procedure TOxygenPlan.SetBackgroundPicture(Value:TPicture);
begin
  FBackgroundPicture.Assign(Value);
  Repaint;
end;

procedure TOxygenPlan.SetScaleX(Value:Integer);
begin
  if (Value > 0) then FScaleX := Value;
  Resize;
end;

procedure TOxygenPlan.SetScaleY(Value:Integer);
begin
  if (Value > 0) then FScaleY := Value;
  Resize;
end;

procedure TOxygenPlan.SetReadOnly(Value:Boolean);
begin
  FReadOnly := Value;
  if Value then if Assigned(FActiveObject) then FActiveObject.UnSelect;
  Repaint;
end;

function TOxygenPlan.GetBackgroundPictureType:String;
begin
  Result:=GetPictureType(FBackgroundPicture);
end;

procedure TOxygenPlan.SetCurLayer(Value:TPlanLayer);
begin
  if Assigned(FActiveObject) then FActiveObject.UnSelect;
  FCurLayer:=Value;
  Repaint;
end;

function TOxygenPlan.GetPicture(Index:Integer):TPlanStandardPicture;
begin
  Result:=nil;
  if (Index>=FStandardPicturesList.Count) then Exit;
  Result:=TPlanStandardPicture(FStandardPicturesList[Index]);
end;

procedure TOxygenPlan.ClearPictures;
begin
  DisableRepaint;
  while (FStandardPicturesList.Count>0) do DeletePicture(0);
  EnableRepaint;
end;

{$IFDEF USE_RX}
function TOxygenPlan.GetAniGif(Index:Integer):TPlanStandardAniGif;
begin
  Result:=nil;
  if (Index>=FStandardAniGifsList.Count) then Exit;
  Result:=TPlanStandardAniGif(FStandardAniGifsList[Index]);
end;

procedure TOxygenPlan.ClearAniGifs;
begin
  DisableRepaint;
  while (FStandardAniGifsList.Count>0) do DeleteAniGif(0);
  EnableRepaint;
end;

function TOxygenPlan.GetAniGifCount : Integer;
begin
  Result:=Self.FStandardAniGifsList.Count;
end;
{$ENDIF}

function TOxygenPlan.GetPictureCount : Integer;
begin
  Result:=FStandardPicturesList.Count;
end;

procedure TOxygenPlan.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  var mt : TMarginType;
      p  : TPlanObject;
begin
  inherited MouseDown(Button,Shift,X,Y);
  {if not (ssCtrl in Shift) then begin
    FSelectedObjects.Clear;
    if Assigned(FActiveObject) then FActiveObject.UnSelect;
  end;}
  if FReadOnly or CurrentLayer.ReadOnly then begin
    FActiveObject := ObjectUnderMouse(mt);
    FActiveMarginType := mtNone;
    Exit;
  end;
  if (FCurDrawingObject <> nil) then begin
    if (FCurDrawingPointNum = 0) then begin
      p := ObjectUnderMouse(mt);
      if (p <> nil) then FCurDrawingObject.Parent := p;
      FCurDrawingObject.Visible:=True;
      //FCurDrawingObject.Layer.Add(FCurDrawingObject);
      FCurDrawingObject.Properties.Add('Layer='+IntToStr(FCurDrawingObject.Layer.Index));
    end;
    try
      DoBeforeDrawNewObject(FCurDrawingObject);
    except
      FCurDrawingObject.Free;
      FCurDrawingObject := nil;
      FCurDrawingPointNum := 0;
      Exit;
    end;
    FActiveObject := FCurDrawingObject;
    if FCurDrawingObject.AddPoint(X,Y,chMouseDown) then Inc(FCurDrawingPointNum)
    else begin
      FCurDrawingObject.FillObjectProperties;
      FSelectedObjects.Clear;
      FCurDrawingObject.Select;
      DoAfterDrawNewObject(FCurDrawingObject);
      FCurDrawingObject := nil;
      FCurDrawingPointNum := 0;
    end;
  end
  else begin
    FActiveObject := ObjectUnderMouse(mt);
    FActiveMarginType := mt;
    if (FActiveObject <> nil) then begin
      FStartCoord := Point(X,Y);
      with FActiveObject do begin
        if (ssCtrl in Shift) then Selected := not Selected
        else begin
          if not Selected then FSelectedObjects.Clear;
          Selected := True;
        end;
      end;
      {if not FWasMove then begin
        if not (ssCtrl in Shift) then begin
          FSelectedObjects.Clear;
          if Assigned(FActiveObject) then FActiveObject.UnSelect;
        end;
        with ActiveObject do if (ssCtrl in Shift) then Selected := not Selected else Selected := True;
      end;
  }
      FWasMove := False;
      FReadyToMoving := (ssLeft in Shift);
    end;
  end;
end;

procedure TOxygenPlan.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  var po:TPlanObject;
      mt : TMarginType;
begin
  if (FCurDrawingObject <> nil) then begin
    if not FCurDrawingObject.AddPoint(X,Y,chMouseUp) then begin
      FCurDrawingObject.FillObjectProperties;
      FSelectedObjects.Clear;
      FCurDrawingObject.Select;
      DoAfterDrawNewObject(FCurDrawingObject);
      FCurDrawingObject := nil;
      FCurDrawingPointNum := 0;
      Repaint;
    end;
  end;
  if not FWasMove then if not (ssCtrl in Shift) then begin
    {FSelectedObjects.Clear;
    if Assigned(FActiveObject) then FActiveObject.UnSelect;}
    FSelectedObjects.Clear;
    if Assigned(FActiveObject) then FActiveObject.Select;
  end;
  po := ObjectUnderMouse(mt);
  if (po <> nil) then begin
    {if not (FReadOnly or CurrentLayer.ReadOnly) then begin
      if not FWasMove then begin
        with po do if (ssCtrl in Shift) then Selected := not Selected else Selected := True;
      end;
    end;}
    DoObjectClick(po,Button,Shift,X,Y);
  end;
  FReadyToMoving := False;
  inherited MouseUp(Button,Shift,X,Y);
  FWasMove := False;
end;

procedure TOxygenPlan.MouseMove(Shift: TShiftState; X, Y: Integer);
  var mt : TMarginType;
begin
  inherited MouseMove(Shift,X,Y);
  FCurX := X;
  FCurY := Y;
  NewObjUM := ObjectUnderMouse(mt);
  if (NewObjUM <> OldObjUM) then begin
    if (OldObjUM <> nil) then DoObjectMouseLeave(OldObjUM);
    if (NewObjUM <> nil) then DoObjectMouseEnter(NewObjUM);
  end;
  if (NewObjUM <> nil) then DoObjectMouseMove(NewObjUM);
  OldObjUM := NewObjUM;
  if (FReadOnly or CurrentLayer.ReadOnly) then begin
    if (NewObjUM <> nil) then Cursor := NewObjUM.Cursor else Cursor := crDefault;
    Exit;
  end;
  if (FActiveObject <> nil) then if FActiveObject.ReadOnly then begin
    if (NewObjUM <> nil) then Cursor := NewObjUM.Cursor else Cursor := crDefault;
    Exit;
  end;
  if (FCurDrawingObject <> nil) and (FCurDrawingPointNum > 0) then begin
    Cursor := crDefault;
    if not FCurDrawingObject.AddPoint(X,Y,chMouseMove) then begin
      FCurDrawingObject.FillObjectProperties;
      FCurDrawingObject.Select;
      DoAfterDrawNewObject(FCurDrawingObject);
      FCurDrawingObject := nil;
      FCurDrawingPointNum := 0;
    end;
    Exit;
  end;
  if ((ssLeft in Shift) and (FActiveObject <> nil)) then begin
    if (FActiveMarginType <> mtNone) then FActiveObject.EditResize(Point(X,Y))
    else if FReadyToMoving then begin
      //FActiveObject.EditMove(Point(X,Y));
      FSelectedObjects.EditMove(Point(X,Y));
      FWasMove := True;
    end;
  end
  else begin
    if ((NewObjUM <> nil) and (FCurDrawingObject = nil)) then begin
      if (not NewObjUM.ReadOnly) and NewObjUM.CanResize then
        case mt of
          mtLeft, mtRight:          Cursor:=crSizeWE;
          mtTop, mtBottom:          Cursor:=crSizeNS;
          mtTopLeft, mtBottomRight: Cursor:=crSizeNWSE;
          mtTopRight, mtBottomLeft: Cursor:=crSizeNESW;
          mtMovePoint             : Cursor:=crDrag;
          else Cursor:=crDefault;
        end
      else Cursor := NewObjUM.Cursor; //crDefault;
    end
    else Cursor:=crDefault;
  end;
  FReadyToMoving := (ssLeft in Shift);
end;

procedure TOxygenPlan.DblClick;
begin
  if (FActiveObject<>nil) then DoObjectDblClick(FActiveObject);
  inherited DblClick;
end;

procedure TOxygenPlan.DoObjectClick(PlanObject:TPlanObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FObjectClick) then FObjectClick(Self,PlanObject,Button,Shift,X,Y);
end;

procedure TOxygenPlan.DoObjectDblClick(PlanObject : TPlanObject);
begin
  if Assigned(FObjectDblClick) then FObjectDblClick(Self, PlanObject);
end;

procedure TOxygenPlan.DoObjectMouseMove(PlanObject : TPlanObject);
begin
  if Assigned(FObjectMouseMove) then FObjectMouseMove(Self, PlanObject);
end;

procedure TOxygenPlan.DoObjectMouseEnter(PlanObject:TPlanObject);
begin
  if Assigned(FObjectMouseEnter) then FObjectMouseEnter(Self, PlanObject);
end;

procedure TOxygenPlan.DoObjectMouseLeave(PlanObject:TPlanObject);
begin
  if Assigned(FObjectMouseLeave) then FObjectMouseLeave(Self, PlanObject);
end;

procedure TOxygenPlan.DoAfterDrawNewObject(PlanObject:TPlanObject);
begin
  if Assigned(FAfterDrawNewObject) then FAfterDrawNewObject(Self, PlanObject);
end;

procedure TOxygenPlan.DoBeforeDrawNewObject(PlanObject : TPlanObject);
begin
  if Assigned(FBeforeDrawNewObject) then FBeforeDrawNewObject(Self, PlanObject);
end;

procedure TOxygenPlan.DoParseProgress(const PercentsCompleted : Integer);
begin
  if Assigned(FParseProgress) then FParseProgress(Self, PercentsCompleted);
end;

procedure TOxygenPlan.DoAfterPaint;
begin
  if Assigned(FAfterPaint) then FAfterPaint(Self);
end;

{$IFDEF USE_RX}
function TOxygenPlan.DoNeedStdAniGif(const AniGifName : String) : Boolean;
begin
  if Assigned(FNeedStdAniGif) then Result := FNeedStdAniGif(Self, AniGifName) else Result:=False;
end;

procedure TOxygenPlan.DoNeedAniGif(PlanAniGif : TPlanAniGif);
begin
  if Assigned(FNeedAniGif) then FNeedAniGif(Self, PlanAniGif);
end;
{$ENDIF}

function TOxygenPlan.DoNeedStdPicture(const PictureName : String) : Boolean;
begin
  if Assigned(FNeedStdPicture) then Result := FNeedStdPicture(Self, PictureName) else Result:=False;
end;

procedure TOxygenPlan.DoNeedPicture(PlanPicture : TPlanPicture);
begin
  if Assigned(FNeedPicture) then FNeedPicture(Self, PlanPicture);
end;

procedure TOxygenPlan.DoObjectBlink(PlanObject : TPlanObject);
begin
  if Assigned(FObjectBlink) then FObjectBlink(Self, PlanObject);
end;

function TOxygenPlan.ObjectUnderMouse(var MarginType:TMarginType):TPlanObject;
begin
  Result:=nil;
  if not CurrentLayer.Visible then Exit;
  Result:=CurrentLayer.ObjectFromCoord(FCurX,FCurY,MarginType);
end;

function TOxygenPlan.ObjectByID(ObjectClass : TPlanObjectClass; const ObjectID : Integer):TPlanObject;
  var i,j:Integer;
      po:TPlanObject;
begin
  Result:=nil;
  for i:=1 to Layers.Count do
    for j:=1 to Layers[i-1].Count do begin
      po:=Layers[i-1].Objects[j-1];
      if ((po is ObjectClass) and (po.ID=ObjectID)) then begin
        Result:=po;
        Exit;
      end;
    end;
end;

function TOxygenPlan.ObjectByName(ObjectClass : TPlanObjectClass; const ObjectName : String):TPlanObject;
  var i,j : Integer;
      po : TPlanObject;
begin
  Result := nil;
  for i := 1 to Layers.Count do
    for j := 1 to Layers[i-1].Count do begin
      po := Layers[i-1].Objects[j-1];
      if ((po is ObjectClass) and (UpperCase(po.Name) = UpperCase(ObjectName))) then begin
        Result := po;
        Exit;
      end;
    end;
end;

function TOxygenPlan.ObjectByCustomProperty(ObjectClass : TPlanObjectClass; const PropertyName, PropertyValue : String) : TPlanObject;
  var i,j : Integer;
      po : TPlanObject;
begin
  Result := nil;
  for i := 1 to Layers.Count do
    for j := 1 to Layers[i-1].Count do begin
      po := Layers[i-1].Objects[j-1];
      if ((po is ObjectClass) and (UpperCase(po.Properties.Values[PropertyName]) = UpperCase(PropertyValue))) then begin
        Result := po;
        Exit;
      end;
    end;
end;

function TOxygenPlan.FindObject(ObjectClass : TPlanObjectClass; const ObjectName : String; const ObjectID : Integer) : TPlanObject;
  var i,j:Integer;
      po:TPlanObject;
begin
  Result := nil;
  for i := 1 to Layers.Count do
    for j := 1 to Layers[i-1].Count do begin
      po := Layers[i-1].Objects[j-1];
      if ((po is ObjectClass) and (UpperCase(po.Name) = UpperCase(ObjectName)) and (po.ID = ObjectID)) then begin
        Result := po;
        Exit;
      end;
    end;
end;


procedure TOxygenPlan.CancelDrawNewObject;
begin
  if (FCurDrawingObject<>nil) then try FCurDrawingObject.Free except end;
  FCurDrawingObject:=nil;
  FCurDrawingPointNum:=0;
end;

procedure TOxygenPlan.Clear;
begin
  CancelDrawNewObject;
  Layers.Clear;
  ClearPictures;
{$IFDEF USE_RX}
  ClearAniGifs;
{$ENDIF}
  FCurLayer:=Layers[Layers.Add(sDefaultLayer)];
  FPlanStrings.Clear;
  FBackgroundPicture.Assign(nil);
  Repaint;
end;

{$IFDEF USE_RX}
procedure TOxygenPlan.DeleteAniGif(Index:Integer);
begin
  if (Index>=FStandardAniGifsList.Count) then Exit;
  with TPlanStandardAniGif(FStandardAniGifsList[Index]) do try
    OnChange := nil;
    Free;
  except end;
  FStandardAniGifsList.Delete(Index);
end;
{$ENDIF}

procedure TOxygenPlan.DeletePicture(Index:Integer);
begin
  if (Index>=FStandardPicturesList.Count) then Exit;
  with TPlanStandardPicture(FStandardPicturesList[Index]) do try
    OnChange := nil;
    Free;
  except end;
  FStandardPicturesList.Delete(Index);
end;

procedure TOxygenPlan.EnableRepaint;
begin
  if (FDisableRepaintCount>0) then Dec(FDisableRepaintCount);
  if (FDisableRepaintCount=0) then Repaint;
end;

procedure TOxygenPlan.DisableRepaint;
begin
  Inc(FDisableRepaintCount);
end;

procedure TOxygenPlan.ExportToMetafile(const FileName : String); // 0.91
  var mf : TMetafile;
      mfc : TMetafileCanvas;
begin
  mf := TMetafile.Create;
  try
    mf.Width := Self.Width;
    mf.Height := Self.Height;
    mfc := TMetafileCanvas.Create(mf, 0);
    try
      PaintToCanvas(mfc, Width, Height);
    finally
      mfc.Free;
    end;
    mf.SaveToFile(FileName);
  finally
    mf.Free;
  end;
end;

procedure TOxygenPlan.ExportToBitmap(const FileName : String);  // 0.91
  var b : TBitmap;
begin
  b := GetBitmap;
  try
    b.SaveToFile(FileName);
  finally
    b.Free;
  end;
end;

procedure TOxygenPlan.ExportToJpeg(const FileName : String);  // 0.91
  var b : TBitmap;
      jpg : TJpegImage;
begin
  b := TBitmap.Create;
  jpg := TJpegImage.Create;
  try
    b.Width := Width;
    b.Height := Height;
    PaintToCanvas(b.Canvas, Width, Height);
    jpg.Assign(b);
    jpg.SaveToFile(FileName);
  finally
    b.Free;
    jpg.Free;
  end;
end;

{$IFDEF USE_RX}
procedure TOxygenPlan.ExportToGif(const FileName : String);  // 0.91
  var b : TBitmap;
      gif : TGifImage;
begin
  b := TBitmap.Create;
  gif := TGifImage.Create;
  try
    b.Width := Width;
    b.Height := Height;
    PaintToCanvas(b.Canvas, Width, Height);
    gif.Assign(b);
    gif.SaveToFile(FileName);
  finally
    b.Free;
    gif.Free;
  end;
end;
{$ENDIF}

function TOxygenPlan.GetBitmap : TBitmap; // 0.96
begin
  Result := TBitmap.Create;
  with Result do begin
    Width := Self.Width;
    Height := Self.Height;
    PaintToCanvas(Canvas, Width, Height);
  end;
end;

function TOxygenPlan.GetMetafile : TMetafile; // 0.96
  var mfc : TMetafileCanvas;
begin
  Result := TMetafile.Create;
  with Result do begin
    Width := Self.Width;
    Height := Self.Height;
    mfc := TMetafileCanvas.Create(Result, 0);
    try
      PaintToCanvas(mfc, Self.Width, Self.Height);
    finally
      mfc.Free;
    end;
  end;
end;

procedure TOxygenPlan.Initialize;
begin
  // Do nothing
end;

procedure TOxygenPlan.DrawNewObject(PlanObjectClass:TPlanObjectClass);
begin
  CancelDrawNewObject;
  FCurDrawingObject := TPlanObject(PlanObjectClass.NewInstance);
  FCurDrawingObject.Create(Self);
  FCurDrawingObject.Visible := False;
end;

{$IFDEF USE_RX}
procedure TOxygenPlan.LoadStdAniGif(const AniGifName : String; PlanAniGif : TPlanAniGif);
begin
  if DoNeedStdAniGif(AniGifName) then PlanAniGif.StdAniGifName := AniGifName;
end;

procedure TOxygenPlan.LoadAniGif(PlanAniGif : TPlanAniGif);
begin
  DoNeedAniGif(PlanAniGif);
end;
{$ENDIF}

procedure TOxygenPlan.LoadStdPicture(const PictureName : String; PlanPicture : TPlanPicture);
begin
  if DoNeedStdPicture(PictureName) then PlanPicture.StdPictureName:=PictureName;
end;

procedure TOxygenPlan.LoadPicture(PlanPicture : TPlanPicture);
begin
  DoNeedPicture(PlanPicture);
end;

procedure TOxygenPlan.Parse(PlanData:TStrings);
  label NextLine;
  var i,ps:Integer;
      ParentObjects:TList;
      s,st:String;
      l:TPlanLayer;
      CurObject:TPlanObject;
      p:TPlanStandardPicture;
      ms:TMemoryStream;
      b:Boolean;
      PlanObjectClass : TPlanObjectClass;
{$IFDEF USE_RX}
      ag:TPlanStandardAniGif;
{$ENDIF}
begin
  if (PlanData.Count=0) then Exit;
  CancelDrawNewObject;
  Layers.Clear;
  {ClearPictures;
  ClearAniGifs;}
  //FCurLayer := Layers[Layers.Add(sDefaultLayer)];
  //FBackgroundPicture.Assign(nil);
  CurObject:=nil;
  ParentObjects:=TList.Create;
  try
    for i:=1 to PlanData.Count do begin
      DoParseProgress(Trunc((i*100)/PlanData.Count));
      s := Trim(PlanData[i-1]);
      ps := Pos(';',s);
      if (ps > 0) then s := System.Copy(s,1,ps-1);
      if (s='') then goto NextLine;
      // -------------------------------------- Scale
      if (Pos(UpperCase(cmScale),AnsiUpperCase(s))=1) then begin
        s:=Copy(s,Length(cmScale)+1,255);
        if (Pos(',',s)>0) then try
          FScaleX:=StrToInt(Copy(s,1,Pos(',',s)-1));
          FScaleY:=StrToInt(Copy(s,Pos(',',s)+1,255));
          if (FScaleX<=0) then FScaleX:=500;
          if (FScaleY<=0) then FScaleY:=500;
        except end;
      end
      // -------------------------------------- SetPenColor
      else if (Pos(UpperCase(cmSetPenColor),AnsiUpperCase(s))=1) then begin
        s := Copy(s,Length(cmSetPenColor)+1,1000);
        Pen.Color := ReadColor(s);
        Canvas.Pen.Color := Pen.Color;
      end
      // -------------------------------------- SetBrushColor
      else if (Pos(UpperCase(cmSetBrushColor),AnsiUpperCase(s))=1) then begin
        s := Copy(s,Length(cmSetBrushColor)+1,1000);
        Brush.Color := ReadColor(s);
        Canvas.Brush.Color := Brush.Color;
      end
      // -------------------------------------- SetLineWidth
      else if (Pos(UpperCase(cmSetLineWidth),AnsiUpperCase(s))=1) then begin
        s := Copy(s,Length(cmSetLineWidth)+1,1000);
        try Pen.Width := StrToInt(s) except end;
        if (Pen.Width = 0) then Pen.Width := 1;
      end
      // -------------------------------------- SetFont
      else if (Pos(UpperCase(cmSetFont),AnsiUpperCase(s))=1) then begin
        s:=Copy(s,Length(cmSetFont)+1,1000);
        ps:=Pos(',',s);
        if (ps>0) then try
          Font.Name:=Copy(s,1,ps-1);
          s:=Copy(s,ps+1,255);
          ps:=Pos(',',s);
          if (ps=0) then ps:=Length(s)+1;
          Font.Size:=StrToInt(Copy(s,1,ps-1));
          Font.Style:=[];
          s:=Copy(s,ps+1,255);
          if (s<>'') then begin
            if (Pos('I',s)>0) then with Font do Style:=Style+[fsItalic];
            if (Pos('B',s)>0) then with Font do Style:=Style+[fsBold];
            if (Pos('U',s)>0) then with Font do Style:=Style+[fsUnderline];
            if (Pos('S',s)>0) then with Font do Style:=Style+[fsStrikeOut];
          end;
        except end;
      end
      // -------------------------------------- BackgroundPicture
      else if (Pos(UpperCase(cmBackgroundPicture),AnsiUpperCase(s))=1) then begin
        System.Delete(s,1,Length(cmBackgroundPicture));
        if (Pos(',',s)>0) then try
          st := Copy(s,1,Pos(',',s)-1); // format
          System.Delete(s,1,Pos(',',s));
          if (Pos(',',s)>0) then begin
            b:=(Copy(s,1,Pos(',',s)-1)<>'0');
            System.Delete(s,1,Pos(',',s));
          end
          else b:=False;
          ms:=TMemoryStream.Create;
          try
            String2Stream(s,ms);
            LoadPictureFromStream(st, FBackgroundPicture, ms);
          except end;
          ms.Free;
          if not ((FBackgroundPicture.Graphic=nil) or FBackgroundPicture.Graphic.Empty) then FBackgroundPicture.Graphic.Transparent:=b;
        except end;
      end
      // -------------------------------------- StandardPicture
      else if (Pos(UpperCase(cmStandardPicture),AnsiUpperCase(s))=1) then begin
        System.Delete(s,1,Length(cmStandardPicture));
        p:=TPlanStandardPicture.Create;
        p.Assign(nil);
        if (Pos(',',s)>0) then try
          p.Name:=Copy(s,1,Pos(',',s)-1);
          System.Delete(s,1,Pos(',',s));
          if (Pos(',',s)>0) then try
            st:=Copy(s,1,Pos(',',s)-1);
            System.Delete(s,1,Pos(',',s));
            if (Pos(',',s)>0) then begin
              b:=(Copy(s,1,Pos(',',s)-1)<>'0');
              System.Delete(s,1,Pos(',',s));
            end
            else b:=False;
            ms:=TMemoryStream.Create;
            try
              String2Stream(s,ms);
              LoadPictureFromStream(st, p, ms);
            except end;
            ms.Free;
            if not ((p.Graphic=nil) or p.Graphic.Empty) then begin
              p.Graphic.Transparent:=b;
              if ((p.Graphic is TBitmap) and (p.Bitmap <> nil) and b) then p.Bitmap.TransparentColor := p.Bitmap.Canvas.Pixels[0,0];
            end;
            AddPicture(p.Name,p);
          except end;
        except end;
        p.Free;
      end
{$IFDEF USE_RX}
      // -------------------------------------- StandardAniGif
      else if (Pos(UpperCase(cmStandardAniGif),AnsiUpperCase(s))=1) then begin
        System.Delete(s,1,Length(cmStandardAniGif));
        ag:=TPlanStandardAniGif.Create;
        ag.Assign(nil);
        if (Pos(',',s)>0) then try
          ag.Name:=Copy(s,1,Pos(',',s)-1);
          System.Delete(s,1,Pos(',',s));
          ms:=TMemoryStream.Create;
          try
            String2Stream(s,ms);
            ag.LoadFromStream(ms)
          except end;
          ms.Free;
          AddAniGif(ag.Name,ag);
        except end;
        ag.Free;
      end
{$ENDIF}
      // -------------------------------------- Layer
      else if (Pos(UpperCase(cmLayer),AnsiUpperCase(s))=1) then begin
        s:=Copy(s,Length(cmLayer)+1,255);
        ps:=Pos(',',s);
        if (ps>0) then try
          st:=Copy(s,1,ps-1);
          l:=Layers.LayerByName(st);
          if (l=nil) then l:=Layers[Layers.Add(st)];
          if (Layers.Count = 1) then FCurLayer := l;
          s:=Copy(s,ps+1,255);
          ps:=Pos(',',s);
          if (ps=0) then ps:=Length(s)+1;
          l.Visible:=(Copy(s,1,ps-1)<>'0');
          s:=Copy(s,ps+1,255);
          if (s<>'') then begin // ReadOnly
            ps:=Pos(',',s);
            if (ps=0) then ps:=Length(s)+1;
            l.ReadOnly:=(Copy(s,1,ps-1)<>'0');
          end;
        except end;
      end
      // -------------------------------------- Start Object
      else if (Pos(UpperCase(sStartObject),AnsiUpperCase(s))=1) then begin
        if (Layers.Count = 0) then FCurLayer := Layers[Layers.Add(sDefaultLayer)];
        if (CurObject <> nil) then CurObject.UpdateObjectProperties;
        s := Copy(s,Length(sStartObject)+1,255);
        PlanObjectClass := PlanObjectsInfo.FindCmdString(UpperCase(s));
        if (PlanObjectClass = nil) then raise EOxygenPlanException.Create(Format(sUnknownPlanObject,[s]));
        CurObject := PlanObjectClass.Create(Self);
        if (CurObject <> nil) then begin
          if (ParentObjects.Count = 0) then CurObject.Parent := nil else CurObject.Parent := TPlanObject(ParentObjects[ParentObjects.Count-1]);
          ParentObjects.Add(CurObject);
        end;
      end
      // -------------------------------------- End Object
      else if (Pos(UpperCase(sEndObject),AnsiUpperCase(s))=1) then begin
        if (ParentObjects.Count > 0) then ParentObjects.Delete(ParentObjects.Count-1);
      end
      else if (CurObject<>nil) then CurObject.Properties.Add(s);
NextLine:
    end;
  finally
    DoParseProgress(100);
    if (Layers.Count = 0) then FCurLayer := Layers[Layers.Add(sDefaultLayer)];
    if (CurObject<>nil) then CurObject.UpdateObjectProperties;
    ParentObjects.Free;
  end;
end;

class procedure TOxygenPlan.RegisterPlanObjectClass(const ACmdString : String; PlanObjectClass : TPlanObjectClass);
begin
  PlanObjectsInfo.Add(ACmdString,PlanObjectClass);
end;

function TOxygenPlan.PictureIndex(const PictureName:String):Integer;
  var i:Integer;
begin
  Result:=-1;
  for i:=1 to FStandardPicturesList.Count do
    if (AnsiUpperCase(TPlanStandardPicture(FStandardPicturesList[i-1]).Name)=AnsiUpperCase(PictureName)) then begin
      Result:=i-1;
      Exit;
    end;
end;

procedure TOxygenPlan.LoadFromFile(FileName:TFileName);
begin
  if not FileExists(FileName) then begin
    raise EOxygenPlanException.Create(Format(sFileNotFound,[FileName]));
    Exit;
  end;
  DisableRepaint;
  try
    Clear;
    (FPlanStrings as TStringList).LoadFromFile(FileName);
    Parse(FPlanStrings);
  finally
    EnableRepaint;
  end;
end;

procedure TOxygenPlan.SaveTo(PlanData:TStrings);
  var i,j:Integer;
      po:TPlanObject;
      s:String;
begin
  PlanData.Clear;
  PlanData.Add(cmScale+IntToStr(ScaleX)+','+IntToStr(ScaleY));
  PlanData.Add(cmSetPenColor+EncodeColor(Pen.Color));
  PlanData.Add(cmSetBrushColor+EncodeColor(Brush.Color));
  PlanData.Add(cmSetLineWidth+IntToStr(Pen.Width));
  if (poSaveBackgroundPicture in Options) then if (FBackgroundPicture <> nil) then
    if (FBackgroundPicture.Graphic <> nil) then if not FBackgroundPicture.Graphic.Empty then begin
      s := Trim(GetPictureStreamString(FBackgroundPicture));
      if (s <> '') then PlanData.Add(cmBackgroundPicture + s);
    end;
  Layers.AddLayersDescr(PlanData);
  if (poSaveStandardPictures in Options) then begin
    for i := 1 to FStandardPicturesList.Count do begin
      s:=TPlanStandardPicture(FStandardPicturesList[i-1]).GetDescrString;
      if (s<>'') then PlanData.Add(s);
    end;
{$IFDEF USE_RX}
    for i := 1 to FStandardAniGifsList.Count do begin
      s := TPlanStandardAniGif(FStandardAniGifsList[i-1]).GetDescrString;
      if (s<>'') then PlanData.Add(s);
    end;
{$ENDIF}
  end;
  for i := 1 to Layers.Count do
    for j := 1 to Layers[i-1].Count do begin
      po := Layers[i-1].Objects[j-1];
      if (po.Parent = nil) then po.AddObjectDescr(0,PlanData);
    end;
end;

procedure TOxygenPlan.SaveToFile(FileName:String);
begin
  SaveTo(FPlanStrings);
  (FPlanStrings as TStringList).SaveToFile(FileName);
end;

procedure TOxygenPlan.LoadFromStream(Stream:TStream);
begin
  DisableRepaint;
  try
    Clear;
    (FPlanStrings as TStringList).LoadFromStream(Stream);
    Parse(FPlanStrings);
  finally
    EnableRepaint;
  end;
end;

procedure TOxygenPlan.SaveToStream(Stream:TStream);
begin
  SaveTo(FPlanStrings);
  (FPlanStrings as TStringList).SaveToStream(Stream);
end;

procedure TOxygenPlan.Resize;
  var i,j:Integer;
begin
  if (csDestroying in ComponentState) then Exit;
  inherited Resize;
  for i:=1 to Layers.Count do
    for j:=1 to Layers[i-1].Count do Layers[i-1].Objects[j-1].ParentResized;
  Repaint;
end;

procedure TOxygenPlan.PaintToCanvas(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i : Integer;
      l : TPlanLayer;
{$IFDEF O2_SW}
      s : String;
{$ENDIF}
begin
  if ((csDestroying in ComponentState) or (FDisableRepaintCount>0)) then Exit;
  ACanvas.Lock;
  ACanvas.Font := Font;
  ACanvas.Pen := Pen;
  //SetBkMode(ACanvas.Handle, Windows.TRANSPARENT);
  ACanvas.Brush := Brush;
  ACanvas.Brush.Color := Color;
  ACanvas.Brush.Style := bsSolid;
  try
    if not FPrinting then ACanvas.FillRect(ACanvas.ClipRect);
    if FTransparent then CopyParentImage(Self, ACanvas);
    if (FBackgroundPicture.Graphic <> nil) then if not FBackgroundPicture.Graphic.Empty then ACanvas.StretchDraw(Rect(0,0,Width,Height),FBackgroundPicture.Graphic);
    if (Layers.Count > 0) then begin
      for i:=1 to Layers.Count do begin
        l:=Layers[i-1];
        if ((l<>CurrentLayer) and (l.Visible)) then l.Draw(ACanvas, CWidth, CHeight);
      end;
      if CurrentLayer.Visible then CurrentLayer.Draw(ACanvas, CWidth, CHeight);
    end;
{$IFDEF O2_SW}
    s := 'Oxygen Plan Library is not registered';
    ACanvas.Font.Name := 'Arial';
    ACanvas.Font.Size := 8;
    ACanvas.Font.Color := clSilver;
    ACanvas.Brush.Style := bsClear;
    ACanvas.TextOut(CWidth - ACanvas.TextWidth(s) - 2, CHeight - ACanvas.TextHeight(s) - 2, s);
{$ENDIF}
  finally
    ACanvas.Unlock;
  end;
end;

procedure TOxygenPlan.PaintToDesktop;
  var i : Integer;
      l : TPlanLayer;
      dc : HDC;
      cnv : TCanvas;
begin
  if ((csDestroying in ComponentState) or (FDisableRepaintCount>0)) then Exit;
  RedrawWindow(0,nil,0,RDW_ERASE+RDW_INTERNALPAINT+RDW_INVALIDATE+RDW_ERASENOW+RDW_UPDATENOW+RDW_ALLCHILDREN);
  Sleep(50);
  dc := GetDC(0);
  cnv := TCanvas.Create;
  cnv.Handle := dc;
  cnv.Lock;
  cnv.Font := Font;
  cnv.Pen := Pen;
  cnv.Brush := Brush;
  cnv.Brush.Color := Color;
  cnv.Brush.Style := bsSolid;
  try
    if (Layers.Count > 0) then begin
      for i := 1 to Layers.Count do begin
        l := Layers[i-1];
        if ((l <> CurrentLayer) and (l.Visible)) then l.Draw(cnv, iScreenWidth, iScreenHeight);
      end;
      if CurrentLayer.Visible then CurrentLayer.Draw(cnv, iScreenWidth, iScreenHeight);
    end;
  finally
    cnv.Unlock;
  end;
  cnv.Free;
  ReleaseDC(0,dc);
end;

procedure TOxygenPlan.Paint;
begin
{$IFDEF VER100}
  if FDoubleBuffered then begin
    FPaintBitmap.FreeImage;
    FPaintBitmap.Width := Width;
    FPaintBitmap.Height := Height;
    PaintToCanvas(FPaintBitmap.Canvas, Width, Height);
    Canvas.Draw(0,0,FPaintBitmap);
  end
  else PaintToCanvas(Canvas, Width, Height);
{$ELSE}
  PaintToCanvas(Canvas, Width, Height);
{$ENDIF}
  DoAfterPaint;
end;


procedure TOxygenPlan.Print(APrinter : TPrinter);
  var w, h : Integer;
      mf : TMetafile;
      mfc : TMetafileCanvas;
      b : TBitmap;
      r : TRect;
{$IFDEF O2_SW}
      s : String;
{$ENDIF}
begin
  FPrinting := True;
  if (FPlanPrintMethod = ppBitmap) then begin // printing as Bitmap
    b := GetBitmap;
    try
      APrinter.BeginDoc;
      w := APrinter.PageWidth;
      h := APrinter.PageHeight;
      r := Rect(FPrintMargins.pmLeft, FPrintMargins.pmTop, w - FPrintMargins.pmRight, h - FPrintMargins.pmBottom);
      if (poPrintProportional in Options) then begin
        while ((r.Right - r.Left)/(r.Bottom - r.Top) > Width/Height) do Dec(r.Right);
        while ((r.Right - r.Left)/(r.Bottom - r.Top) < Width/Height) do Dec(r.Bottom);

        {r := Rect(FPrintMargins.pmLeft, FPrintMargins.pmTop, Width, Height);
        if (r. > APrinter.PageWidth) or (h > APrinter.PageHeight) then repeat
          Dec(w);
          Dec(h);
        until (w <= APrinter.PageWidth) and (h <= APrinter.PageHeight)
        else if (w < APrinter.PageWidth) and (h < APrinter.PageHeight) then repeat
          Inc(w);
          Inc(h);
        until (w <= APrinter.PageWidth) and (h <= APrinter.PageHeight)}
      end;
      with APrinter.Canvas do begin
        StretchDraw(r, b);
        Pen.Width := 1;
        Pen.Color := clBlack;
        Brush.Style := bsClear;
        Rectangle(FPrintMargins.pmLeft, FPrintMargins.pmTop, w - PrintMargins.pmRight, h - FPrintMargins.pmBottom);
{$IFDEF O2_SW}
        s := 'Oxygen Plan Library is not registered';
        Font.Name := 'Arial';
        Font.Size := 40;
        Font.Color := clSilver;
        Brush.Style := bsClear;
        TextOut((w - TextWidth(s)) div 2, h div 2,s);
{$ENDIF}
      end;
      APrinter.EndDoc;
    finally
      FPrinting := False;
      b.Free;
    end;
  end
  else begin // printing as metafile
    mf := TMetafile.Create;
    try
      APrinter.BeginDoc;
      mf.Width := Width;
      mf.Height := Height;
      w := APrinter.PageWidth;
      h := APrinter.PageHeight;
      mfc := TMetafileCanvas.Create(mf, APrinter.Handle);
      try
        PaintToCanvas(mfc, Width, Height);
      finally
        mfc.Free;
      end;
      r := Rect(FPrintMargins.pmLeft, FPrintMargins.pmTop, w - FPrintMargins.pmRight, h - FPrintMargins.pmBottom);
      if (poPrintProportional in Options) then begin
        while ((r.Right - r.Left)/(r.Bottom - r.Top) > Width/Height) do Dec(r.Right);
        while ((r.Right - r.Left)/(r.Bottom - r.Top) < Width/Height) do Dec(r.Bottom);
      end;
      with APrinter.Canvas do begin
        StretchDraw(r, mf);
        Pen.Width := 1;
        Pen.Color := clBlack;
        Brush.Style := bsClear;
        Rectangle(FPrintMargins.pmLeft, FPrintMargins.pmTop, w - PrintMargins.pmRight, h - FPrintMargins.pmBottom);
{$IFDEF O2_SW}
        s := 'Oxygen Plan Library is not registered';
        Font.Name := 'Arial';
        Font.Size := 40;
        Font.Color := clSilver;
        Brush.Style := bsClear;
        TextOut((w - TextWidth(s)) div 2, h div 2,s);
{$ENDIF}
      end;
      APrinter.EndDoc;
    finally
      FPrinting := False;
      mf.Free;
    end;
  end;
end;


procedure Register;
begin
  RegisterComponents('Oxygen', [TOxygenPlan]);
end;


initialization
  PlanObjectsInfo := TPlanObjectInfoList.Create;

finalization
  PlanObjectsInfo.Destroy;

end.




