{===============================================================}
{    Oxygen Plan Library, v1.3  1999-2001 Oxygen Software      }
{---------------------------------------------------------------}
{        Written by Oleg Fyodorov, delphi@oxygensoftware.com    }
{                http://www.oxygensoftware.com                  }
{===============================================================}

unit PlanObjects;

{$Include O2Plan.inc}
{$OPTIMIZATION OFF}

interface

  uses Windows, Graphics, Classes, Controls, SysUtils, StrF, {$IFDEF USE_RX} TimerLst, RxGif, VCLUtils, {$ENDIF} jpeg,
       PlanUtils;

  resourcestring
    // Plan objects
    ocLine              = 'Line';
    ocBox               = 'Box';
    ocCircle            = 'Circle';
    ocEllipse           = 'Ellipse';
    ocText              = 'Text';
    ocPicture           = 'Picture';
    ocPolygon           = 'Polygon';
    ocPolyLine          = 'PolyLine';
    ocTraceLine         = 'TraceLine';
    ocCompositeObject   = 'Composite';
    ocAniGif            = 'AniGif';
    ocAniPicture        = 'AniPicture';

    sStartObject = 'Object ';
    sEndObject   = 'End';

type
  TPlanObject = class;
  TPlanObjectClass = class of TPlanObject;
  TMarginType = (mtNone, mtLeft, mtTopLeft, mtTop, mtTopRight, mtRight, mtBottomRight, mtBottom, mtBottomLeft, mtMovePoint);
  TCallHandler = (chMouseDown, chMouseMove, chMouseUp);

{$IFDEF USE_RX}
  TBlinkType = (btNone, btPen, btBrush, btPicture, btFont, btText);
{$ENDIF}

  PPointArray = ^TPointArray;
  TPointArray = array[0..10000] of TPoint;

  TGradientRecord = class
    StartColor,
    EndColor  : TColor;
    Direction : TGradientFillDirection;
    Colors    : Byte;
    Active    : Boolean;
  end;

  TPlanLayer = class
    private
      FOwner      : TCustomControl;
      FObjectList : TList;
      FVisible    : Boolean;
      HRegion     : HRGN;
      FReadOnly   : Boolean;
      function  GetObject(Index:Integer):TPlanObject;
      function  GetObjectsCount : Integer;
      procedure SetVisible(Value:Boolean);
      procedure SetReadOnly(const Value : Boolean);
    public
      Name    : String;
      Index   : Integer;
      property Objects[Index : Integer] : TPlanObject read GetObject; default;
      property Count:Integer read GetObjectsCount;
      property ReadOnly : Boolean read FReadOnly write SetReadOnly;
      property Visible : Boolean read FVisible write SetVisible;
      constructor Create(AOwner : TCustomControl);
      procedure Free;
      function  Add(PlanObject : TPlanObject):Integer;
      function  IndexOf(PlanObject : TPlanObject):Integer;
      procedure Delete(Index:Integer);
      procedure Clear;
      procedure Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
      function  ObjectFromCoord(X, Y : Integer; var MarginType : TMarginType):TPlanObject;
      procedure SetTopObject(PlanObject : TPlanObject);
  end;

  TPlanLayers = class
    private
      FOwner : TCustomControl;
      FLayersList:TList;
      function NewLayer(const LName:String):TPlanLayer;
      function Get(Index:Integer):TPlanLayer;
      procedure Put(Index:Integer; Layer:TPlanLayer);
      function GetCount:Integer;
    public
      property Layers[Index:Integer]:TPlanLayer read Get write Put; default;
      property Count:Integer read GetCount;
      constructor Create(AOwner : TCustomControl);
      procedure Free;
      function  Add(const LayerName:String):Integer;
      procedure Delete(Index:Integer);
      function  IndexOf(const LayerName:String):Integer;
      procedure Clear;
      function  LayerByName(const LayerName:String):TPlanLayer;
      procedure AddLayersDescr(sl:TStrings);
  end;

  TPlanObject = class(TObject)
    private
      FOwner       : TCustomControl;
      FParent      : TPlanObject;
      FChildren    : TList;
      FLayer       : TPlanLayer;
      FID          : Integer;
      FName        : String;
      FSelected    : Boolean;
      FPen         : TPen;
      FBrush       : TBrush;
      FSelectColor : TColor;
      FVisible     : Boolean;
      FCanResize   : Boolean;
      FClippingRgn : HRGN;
      FDestroying  : Boolean;
      FReadOnly    : Boolean;
      FProperties  : TStrings;
      FCursor      : TCursor;
      FGradientRecord : TGradientRecord;

{$IFDEF USE_RX}
      // blinking
      FBlinkTimerHandle       : Integer;
      FBlinkType              : TBlinkType;
      FBlinkPen, FSavePen     : TPen;
      FBlinkBrush, FSaveBrush : TBrush;
      FBlinkIndex             : Integer; // 0 - Save, 1 - Blink
{$ENDIF}

      function  GetChildrenCount : Integer;
      procedure SetPen(Value : TPen); virtual;
      procedure SetBrush(Value : TBrush); virtual;
      procedure SetLayer(Value : TPlanLayer); virtual;
      procedure SetName(Value : String);
      function  GetObjectsCount : Integer;
      function  GetObject(Index : Integer) : TPlanObject;
      procedure SetParent(Value : TPlanObject);
      procedure SetSelected(Value : Boolean); virtual;
      function  GetSelectedByColor : Boolean;
      procedure SetVisible(const Value : Boolean);
{$IFDEF USE_RX}
      function  GetBlinking : Boolean;
{$ENDIF}
   protected
      function  GetObjectTypeDescr : String; virtual;
      procedure SaveCoords; virtual;
      function  GetDisplayRect:TRect; virtual; abstract;
      procedure DrawChildren(ACanvas : TCanvas; const CWidth, CHeight : Integer); virtual;
{$IFDEF USE_RX}
      procedure BlinkTimer(Sender: TObject); virtual;  // blinking handler
{$ENDIF}
      function  GetClippingRgn : HRGN; virtual;
      procedure SetReadOnly(const Value : Boolean);
      procedure SetProperties(Value : TStrings);
    public
      property Properties      : TStrings read FProperties write SetProperties;
{$IFDEF USE_RX}
      property Blinking        : Boolean read GetBlinking;
      property BlinkType       : TBlinkType read FBlinkType write FBlinkType;
      property BlinkPen        : TPen read FBlinkPen;
      property BlinkBrush      : TBrush read FBlinkBrush;
{$ENDIF}
      property Brush           : TBrush read FBrush write SetBrush;
      property CanResize       : Boolean read FCanResize write FCanResize;
      property ChildrenCount   : Integer read GetChildrenCount;
      property Cursor          : TCursor read FCursor write FCursor;
      property DisplayRect     : TRect read GetDisplayRect;
      property GradientRecord  : TGradientRecord read FGradientRecord;
      property ID              : Integer read FID write FID;
      property Name            : String read FName write SetName;
      property Layer           : TPlanLayer read FLayer write SetLayer;
      property Objects[Index : Integer] : TPlanObject read GetObject;
      property ObjectsCount    : Integer read GetObjectsCount;
      property ObjectTypeDescr : String read GetObjectTypeDescr;
      property Owner           : TCustomControl read FOwner;
      property Parent          : TPlanObject read FParent write SetParent;
      property Pen             : TPen read FPen write SetPen;
      property ReadOnly        : Boolean read FReadOnly write SetReadOnly;
      property Selected        : Boolean read FSelected write SetSelected;
      property SelectedByColor : Boolean read GetSelectedByColor;
      property Visible : Boolean read FVisible write SetVisible;

      procedure   AddObjectDescr(const SpaceNum : Integer; sl : TStrings); virtual;
      function    AddPoint(const CoordX,CoordY : Integer; CallHandler : TCallHandler {$IFNDEF VER100} = chMouseMove {$ENDIF}) : Boolean; virtual; abstract;// returns False when drawing is finished
      function    CanBeParent : Boolean; virtual; abstract;
      function    Clone(Parent : TPlanObject) : TPlanObject; virtual; abstract;
      constructor Create(Owner:TCustomControl); virtual;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); virtual; abstract;
      procedure   EditMove(APoint : TPoint); virtual;
      procedure   EditResize(APoint:TPoint); virtual; abstract;
      procedure   FillObjectProperties; virtual;
      procedure   Free; virtual;
      function    HasInternalPoint(X,Y:Integer):Boolean; virtual; abstract;
      function    HasInternalProperty(const PropertyName : String) : Boolean; virtual;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; virtual; abstract;
      procedure   Move(const DeltaX, DeltaY : Integer); virtual;
      procedure   Resize(const DeltaX, DeltaY : Integer); virtual; abstract;
      procedure   NeedRepaint; virtual; abstract;
      function    ObjectFromCoord(X, Y:Integer; var MarginType : TMarginType):TPlanObject; virtual;
      procedure   ParentResized; virtual;
      procedure   Select; virtual;
      procedure   SelectByColor(const AColor:TColor); virtual;
      procedure   SetTopObject(PlanObject : TPlanObject);
{$IFDEF USE_RX}
      procedure   StartBlink(const Interval:Integer); virtual;
      procedure   StopBlink; virtual;
{$ENDIF}
      procedure   UnSelect; virtual;
      procedure   UnSelectColor; virtual;
      procedure   UpdateObjectProperties; virtual;
  end;



  TPlanLine = class(TPlanObject)
    private
      Fx1,Fy1,Fx2,Fy2   :Integer;
      ScrSave1,ScrSave2 :TPoint;
      procedure SetX1(Value:Integer);
      procedure SetX2(Value:Integer);
      procedure SetY1(Value:Integer);
      procedure SetY2(Value:Integer);
      function  GetScrPoint1:TPoint;
      function  GetScrPoint2:TPoint;
      procedure SetScrPoint1(p:TPoint);
      procedure SetScrPoint2(p:TPoint);
    protected
      function  GetObjectTypeDescr : String; override;
      procedure SaveCoords; override;
      function  GetClippingRgn : HRGN; override;
    public
      property X1 : Integer read Fx1 write SetX1;
      property Y1 : Integer read Fy1 write SetY1;
      property X2 : Integer read Fx2 write SetX2;
      property Y2 : Integer read Fy2 write SetY2;
      property ScrPoint1 : TPoint read GetScrPoint1 write SetScrPoint1;
      property ScrPoint2 : TPoint read GetScrPoint2 write SetScrPoint2;

      function    AddPoint(const CoordX,CoordY:Integer; CallHandler : TCallHandler {$IFNDEF VER100}= chMouseMove{$ENDIF}):Boolean; override;
      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      function    GetDisplayRect:TRect; override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; override;
      procedure   EditMove(APoint:TPoint); override;
      procedure   EditResize(APoint:TPoint); override;
      procedure   NeedRepaint; override;
      procedure   Move(const DeltaX, DeltaY : Integer); override;
      procedure   Resize(const DeltaX, DeltaY : Integer); override;
      procedure   UpdateObjectProperties; override;
  end;

  TPlanBox = class(TPlanObject)
    private
      Fx1,Fy1,Fx2,Fy2   : Integer;
      ScrSave1,ScrSave2 : TPoint;
      procedure SetX1(Value:Integer);
      procedure SetX2(Value:Integer);
      procedure SetY1(Value:Integer);
      procedure SetY2(Value:Integer);
      function  GetScrPoint1:TPoint;
      function  GetScrPoint2:TPoint;
      procedure SetScrPoint1(p:TPoint);
      procedure SetScrPoint2(p:TPoint);
    protected
      function  GetObjectTypeDescr : String; override;
      procedure SaveCoords; override;
    public
      property X1 : Integer read Fx1 write SetX1;
      property Y1 : Integer read Fy1 write SetY1;
      property X2 : Integer read Fx2 write SetX2;
      property Y2 : Integer read Fy2 write SetY2;
      property ScrPoint1 : TPoint read GetScrPoint1 write SetScrPoint1;
      property ScrPoint2 : TPoint read GetScrPoint2 write SetScrPoint2;

      function    AddPoint(const CoordX,CoordY:Integer; CallHandler : TCallHandler {$IFNDEF VER100}= chMouseMove{$ENDIF}):Boolean; override;
      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      function    GetDisplayRect:TRect; override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; override;
      procedure   EditMove(APoint:TPoint); override;
      procedure   EditResize(APoint:TPoint); override;
      procedure   NeedRepaint; override;
      procedure   Move(const DeltaX, DeltaY : Integer); override;
      procedure   Resize(const DeltaX, DeltaY : Integer); override;
      procedure   UpdateObjectProperties; override;
  end;

  TPlanEllipse = class(TPlanBox)
    protected
      function  GetClippingRgn : HRGN; override;
      function  GetObjectTypeDescr : String; override;
    public
      property X1;
      property Y1;
      property X2;
      property Y2;
      property ScrPoint1;
      property ScrPoint2;

      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; override;
      procedure   EditMove(APoint:TPoint); override;
      procedure   NeedRepaint; override;
      procedure   EditResize(APoint:TPoint); override;
      procedure   Move(const DeltaX, DeltaY : Integer); override;
      procedure   Resize(const DeltaX, DeltaY : Integer); override;
  end;

  TPlanCircle = class(TPlanObject)
    private
      Fx,Fy,FRadius : Integer;
      ScrSaveCenter : TPoint;
      ScrSaveRadius : Integer;
      procedure SetX(Value:Integer);
      procedure SetY(Value:Integer);
      procedure SetRadius(Value:Integer);
      function  GetScrCenter:TPoint;
      function  GetScrRadius:Integer;
      procedure SetScrCenter(p:TPoint);
      procedure SetScrRadius(r:Integer);
    protected
      function  GetObjectTypeDescr : String; override;
      procedure SaveCoords; override;
      function  GetClippingRgn : HRGN; override;
    public
      property X : Integer read Fx write SetX;
      property Y : Integer read Fy write SetY;
      property Radius : Integer read FRadius write SetRadius;
      property ScrCenter : TPoint read GetScrCenter write SetScrCenter;
      property ScrRadius : Integer read GetScrRadius write SetScrRadius;

      function    AddPoint(const CoordX,CoordY:Integer; CallHandler : TCallHandler {$IFNDEF VER100}= chMouseMove{$ENDIF}):Boolean; override;
      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      function    GetDisplayRect:TRect; override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; override;
      procedure   EditMove(APoint:TPoint); override;
      procedure   NeedRepaint; override;
      procedure   EditResize(APoint:TPoint); override;
      procedure   Move(const DeltaX, DeltaY : Integer); override;
      procedure   Resize(const DeltaX, DeltaY : Integer); override;
      procedure   UpdateObjectProperties; override;
  end;

  TPlanText = class(TPlanBox)
    private
      FText, FSaveText, FBlinkText : String;
      FFont, FSaveFont, FBlinkFont : TFont;
      FAlignment : TAlignment;
      FEscapement : Integer;
      procedure SetText(Value:String);
      procedure SetFont(Value:TFont);
      procedure SetAlignment(Value:TAlignment);
      procedure SetEscapement(const Value : Integer);
    protected
      function  GetObjectTypeDescr : String; override;
{$IFDEF USE_RX}
      procedure BlinkTimer(Sender: TObject); override;
{$ENDIF}
    public
{$IFDEF USE_RX}
      property BlinkFont : TFont read FBlinkFont;
      property BlinkText : String read FBlinkText write FBlinkText;
{$ENDIF}
      property X1;
      property Y1;
      property X2;
      property Y2;
      property Text : String read FText write SetText;
      property Font : TFont read FFont write SetFont;
      property Alignment : TAlignment read FAlignment write SetAlignment;
      property Escapement : Integer read FEscapement write SetEscapement;
      property ScrPoint1;
      property ScrPoint2;

      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner : TCustomControl); override;
      procedure   Free; override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
{$IFDEF USE_RX}
      procedure   StartBlink(const Interval:Integer); override;
      procedure   StopBlink; override;
{$ENDIF}
      procedure   UpdateObjectProperties; override;
  end;

  TPlanPicture = class(TPlanBox)
    private
      FPicture {$IFDEF USE_RX}, FSavePicture, FBlinkPicture {$ENDIF} : TPicture;
      FStdPictureName : String;
      procedure SetStdPictureName(const Value : String);
      function  GetTransparent : Boolean;
      procedure SetTransparent(const Value : Boolean);
    protected
      function  GetObjectTypeDescr : String; override;
{$IFDEF USE_RX}
      procedure BlinkTimer(Sender: TObject); override;
{$ENDIF}
    public
{$IFDEF USE_RX}
      property BlinkPicture : TPicture read FBlinkPicture;
{$ENDIF}
      property X1;
      property Y1;
      property X2;
      property Y2;
      property ScrPoint1;
      property ScrPoint2;
      property StdPictureName : String read FStdPictureName write SetStdPictureName;
      property Picture : TPicture read FPicture; {write FPicture;}
      property Transparent : Boolean read GetTransparent write SetTransparent;

      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      procedure   Free; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
{$IFDEF USE_RX}
      procedure   StartBlink(const Interval:Integer); override;
      procedure   StopBlink; override;
{$ENDIF}
      procedure   UpdateObjectProperties; override;
  end;

{$IFDEF USE_RX}
  TPlanAniPicture = class(TPlanPicture)
    private
      FImageIndex : Integer;
      FImageCount : Integer;
      FCurBitmap  : TBitmap;
      FInterval   : Integer;
      procedure SetImageIndex(const Value : Integer);
      procedure SetImageCount(const Value : Integer);
    protected
      procedure BlinkTimer(Sender: TObject); override;
      function  GetObjectTypeDescr : String; override;
    public
      property ImageCount : Integer read FImageCount write SetImageCount;
      property ImageIndex : Integer read FImageIndex write SetImageIndex;
      property Interval : Integer read FInterval write FInterval;

      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Animate(const Interval:Integer);
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      procedure   Free; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      procedure   StartBlink(const Interval:Integer); override;
      procedure   Stop;
      procedure   StopBlink; override;
      procedure   UpdateObjectProperties; override;
  end;
{$ENDIF}


  TPlanPolygon = class(TPlanObject)
    private
      FPointCount:Integer;
      FPointArray:PPointArray;
      FScrPointArray : PPointArray;
      FScrSaveArray  : PPointArray;
      FDrawArray : PPointArray;
      FCurMovingPointIndex : Integer;
      //FCurDrawingPointNum : Integer;
      procedure SetPointCount(Value:Integer);
      function  GetPoint(Index:Integer):TPoint;
      procedure SetPoint(Index:Integer; Value:TPoint);
    protected
      function  GetObjectTypeDescr : String; override;
      procedure SaveCoords; override;
      function  GetClippingRgn : HRGN; override;
    public
      property PointCount:Integer read FPointCount write SetPointCount;
      property Points[Index:Integer]:TPoint read GetPoint write SetPoint;

      function    AddPoint(const CoordX,CoordY:Integer; CallHandler : TCallHandler {$IFNDEF VER100}= chMouseMove{$ENDIF}):Boolean; override;
      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   ParentResized; override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   DeletePoint(const Index : Integer);
      procedure   FillObjectProperties; override;
      procedure   Free; override;
      function    GetDisplayRect:TRect; override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      function    HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean; override;
      procedure   EditMove(APoint:TPoint); override;
      procedure   NeedRepaint; override;
      procedure   EditResize(APoint:TPoint); override;
      procedure   Move(const DeltaX, DeltaY : Integer); override;
      procedure   Resize(const DeltaX, DeltaY : Integer); override;
      procedure   UpdateObjectProperties; override;
  end;

  TPlanPolyLine = class(TPlanPolygon)
    protected
      function  GetObjectTypeDescr : String; override;
      function  GetClippingRgn : HRGN; override;
    public
      property PointCount;
      property Points;

      function    AddPoint(const CoordX,CoordY : Integer; CallHandler : TCallHandler) : Boolean; override;
      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      function    HasInternalPoint(X,Y:Integer):Boolean; override;
  end;

  TPlanTraceLine = class(TPlanPolyLine)
    protected
      function  GetObjectTypeDescr : String; override;
    public
      property  PointCount;
      property  Points;

      function  Clone(Parent : TPlanObject) : TPlanObject; override;
      procedure Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      function  GetDisplayRect:TRect; override;
  end;

  TPlanCompositeObject = class(TPlanBox)
    private
      procedure SetPen(Value:TPen); override;
      procedure SetBrush(Value:TBrush); override;
      procedure SetSelected(Value:Boolean); override;
   protected
      function  GetObjectTypeDescr : String; override;
    public
      function  CanBeParent : Boolean; override;
      procedure Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure Select; override;
      procedure UnSelect; override;
      procedure SelectByColor(const AColor:TColor); override;
      procedure UnSelectColor; override;
  end;

{$IFDEF USE_RX}
  TPlanAniGif = class(TPlanBox)
    private
      FImage : TGIFImage;
      FStdAniGifName : String;
      function GetDelayTime(Index: Integer): Cardinal;
      procedure SetStdAniGifName(const Value : String);
      procedure SetGIFImage(Value:TGIFImage);
      function GetFrameIndex : Integer;
      procedure SetFrameIndex(const Value : Integer);
      function GetFrameCount : Integer;
    protected
      procedure BlinkTimer(Sender: TObject); override;
      function  GetFrameBitmap(Index: Integer; var TransColor: TColor): TBitmap;
      function  GetObjectTypeDescr : String; override;
    public
      property StdAniGifName : String read FStdAniGifName write SetStdAniGifName;
      property GIFImage : TGIFImage read FImage write SetGIFImage;
      property FrameIndex : Integer read GetFrameIndex write SetFrameIndex;
      property FrameCount : Integer read GetFrameCount;

      function    CanBeParent : Boolean; override;
      function    Clone(Parent : TPlanObject) : TPlanObject; override;
      constructor Create(Owner:TCustomControl); override;
      procedure   Animate;
      procedure   Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer); override;
      procedure   FillObjectProperties; override;
      procedure   Free; override;
      function    HasInternalProperty(const PropertyName : String) : Boolean; override;
      procedure   StartBlink(const Interval:Integer); override;
      procedure   StopBlink; override;
      procedure   Stop;
      procedure   UpdateObjectProperties; override;
  end;
{$ENDIF}

  procedure SelectRectRgn(ACanvas : TCanvas; SelectColor : TColor; p1, p2 : TPoint);

implementation

  uses O2Plan;

  const
    MaxDelayTime = 10000;
    MinDelayTime = 50;

procedure SelectRectRgn(ACanvas : TCanvas; SelectColor : TColor; p1, p2 : TPoint);
begin
  with ACanvas do begin
    Pen.Mode := pmCopy;
    Pen.Color := SelectColor;
    Pen.Width := 1;
    Brush.Color := SelectColor;
    Brush.Style := bsSolid;
    Rectangle(p1.X - MarginDelta, p1.Y - MarginDelta, p1.X + MarginDelta, p1.Y + MarginDelta);
    Rectangle(p2.X - MarginDelta, p1.Y - MarginDelta, p2.X + MarginDelta, p1.Y + MarginDelta);
    Rectangle(p2.X - MarginDelta, p2.Y - MarginDelta, p2.X + MarginDelta, p2.Y + MarginDelta);
    Rectangle(p1.X - MarginDelta, p2.Y - MarginDelta, p1.X + MarginDelta, p2.Y + MarginDelta);
  end;
end;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//    TPlanLayer
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanLayer.Create(AOwner : TCustomControl);
begin
  inherited Create;
  FObjectList:=TList.Create;
  FOwner:=AOwner;
  HRegion:=0;
  FReadOnly := False;
end;

procedure TPlanLayer.Free;
begin
  Clear;
  FObjectList.Free;
  inherited Free;
end;

function TPlanLayer.GetObject(Index:Integer):TPlanObject;
begin
  Result:=TPlanObject(FObjectList[Index]);
end;

function TPlanLayer.GetObjectsCount : Integer;
begin
  Result:=FObjectList.Count;
end;

procedure TPlanLayer.SetVisible(Value:Boolean);
begin
  if (FVisible = Value) then Exit;
  FVisible := Value;
  FOwner.Repaint;
end;

procedure TPlanLayer.SetReadOnly(const Value : Boolean);
  var op : TOxygenPlan;
begin
  FReadOnly := Value;
  op := FOwner as TOxygenPlan;
  if Value then if Assigned(op.ActiveObject) then if (op.ActiveObject.Layer = Self) then op.ActiveObject.UnSelect;
end;

function TPlanLayer.Add(PlanObject : TPlanObject):Integer;
begin
  Result:=FObjectList.Add(PlanObject);
end;

function TPlanLayer.IndexOf(PlanObject : TPlanObject):Integer;
begin
  Result:=FObjectList.IndexOf(PlanObject);
end;

procedure TPlanLayer.Delete(Index:Integer);
begin
  TPlanObject(FObjectList[Index]).Free;
end;

procedure TPlanLayer.Clear;
begin
  while (Count>0) do Delete(0);
  FObjectList.Clear;
end;

procedure TPlanLayer.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i  : Integer;
      po : TPlanObject;
begin
  if ((FOwner as TOxygenPlan).DisableRepaintCount > 0) or (not Visible) then Exit;
  for i := 1 to FObjectList.Count do begin
    po := TPlanObject(FObjectList[i-1]);
    if (po.FParent = nil) then po.Draw(ACanvas, CWidth, CHeight);
  end;
end;

function TPlanLayer.ObjectFromCoord(X, Y:Integer; var MarginType : TMarginType):TPlanObject;
  var i:Integer;
      po:TPlanObject;
      mt:TMarginType;
begin
  Result:=nil;
  MarginType := mtNone;
  if not Visible then Exit;
  for i:=Count downto 1 do begin
    if (Objects[i-1].HasMarginPoint(X,Y,mt) and (Objects[i-1].Visible or (not Objects[i-1].ReadOnly))) then begin
      Result:=Objects[i-1];
      MarginType:=mt;
      Exit;
    end
    else if (Objects[i-1].HasInternalPoint(X,Y) and (Objects[i-1].Visible or (not Objects[i-1].ReadOnly))) then begin
      Result:=Objects[i-1];
      MarginType := mtNone;
      po:=Objects[i-1].ObjectFromCoord(X,Y,mt);
      if (po<>nil) then begin
        Result:=po;
        MarginType := mt;
      end;
      Exit;
    end;
  end;
end;

procedure TPlanLayer.SetTopObject(PlanObject : TPlanObject);
  var i : Integer;
begin
  i := IndexOf(PlanObject);
  if (i = -1) or (PlanObject = nil) then Exit;
  FObjectList.Delete(i);
  FObjectList.Add(PlanObject);
  if (PlanObject.Parent <> nil) then PlanObject.Parent.SetTopObject(PlanObject);
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//    TPlanLayers
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanLayers.Create(AOwner : TCustomControl);
begin
  inherited Create;
  FLayersList:=TList.Create;
  FOwner:=AOwner;
end;

procedure TPlanLayers.Free;
begin
  Clear;
  FLayersList.Free;
  inherited Free;
end;

function TPlanLayers.NewLayer(const LName:String):TPlanLayer;
begin
  Result:=TPlanLayer.Create(FOwner);
  with Result do begin
    Name:=LName;
    Index:=Self.Count;
    FVisible := True;
  end;
end;

function TPlanLayers.Get(Index:Integer):TPlanLayer;
begin
  Result:=TPlanLayer(FLayersList[Index]);
end;

procedure TPlanLayers.Put(Index:Integer; Layer:TPlanLayer);
begin
  FLayersList[Index]:=Layer;
end;

function TPlanLayers.GetCount:Integer;
begin
  Result:=FLayersList.Count;
end;

function TPlanLayers.Add(const LayerName:String):Integer;
begin
  Result:=FLayersList.Add(NewLayer(LayerName));
end;

procedure TPlanLayers.Delete(Index:Integer);
begin
  TPlanLayer(FLayersList[Index]).Free;
  FLayersList.Delete(Index);
end;

function TPlanLayers.IndexOf(const LayerName:String):Integer;
  var i:Integer;
begin
  Result:=-1;
  for i:=1 to FLayersList.Count do if (UpperCase(TPlanLayer(FLayersList[i-1]).Name)=UpperCase(LayerName)) then begin
    Result:=i-1;
    Exit;
  end;
end;

procedure TPlanLayers.Clear;
begin
  while (Count>0) do Delete(0);
  FLayersList.Clear;
end;

function TPlanLayers.LayerByName(const LayerName:String):TPlanLayer;
  var i:Integer;
begin
  Result:=nil;
  for i:=1 to Count do if (AnsiUpperCase(Layers[i-1].Name)=AnsiUpperCase(LayerName)) then begin
    Result:=Layers[i-1];
    Exit;
  end;
end;

procedure TPlanLayers.AddLayersDescr(sl:TStrings);
  var i:Integer;
      pl:TPlanLayer;
      s:String;
begin
  for i:=1 to Count do begin
    pl:=TPlanLayer(FLayersList[i-1]);
    if pl.Visible then s:='1' else s:='0';
    if pl.ReadOnly then s:=s+',1' else s:=s+',0';
    sl.Add(cmLayer+pl.Name+','{+IntToStr(pl.Index)+','}+s);
  end;
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanObject
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanObject.Create(Owner : TCustomControl);
  var op : TOxygenPlan;
begin
  if not (Owner is TOxygenPlan) then raise EOxygenPlanException.Create(sInvalidOwner);
  inherited Create;
  op := (Owner as TOxygenPlan);
  FDestroying := False;
  FChildren := TList.Create;
  FProperties := TStringList.Create;
  FPen := TPen.Create;
  FBrush := TBrush.Create;
  FPen.Assign(op.Pen);
  FBrush.Assign(op.Brush);
  FID := 0;
  FName:='';
  FOwner := Owner;
  FLayer := op.CurrentLayer;
  FParent := nil;
  FSelected := False;
  FSelectColor := -1;
  FVisible := True;
  FCanResize := True;
{$IFDEF USE_RX}
  FBlinkTimerHandle := 0;
  FBlinkType := btNone;
  FBlinkPen := TPen.Create;
  FSavePen := TPen.Create;
  FBlinkBrush := TBrush.Create;
  FSaveBrush := TBrush.Create;
  FBlinkIndex := 0;
{$ENDIF}
  FReadOnly := False;
  FCursor := crDefault;
  FGradientRecord := TGradientRecord.Create;
  with FGradientRecord do begin
    Active := False;
    StartColor := clBlack;
    EndColor := clBlack;
    Direction := fdTopToBottom;
    Colors := 255;
  end;
  FLayer.Add(Self);
end;

function TPlanObject.GetObjectTypeDescr : String;
begin
  Result := FProperties.Values['ObjectType'];
end;

procedure TPlanObject.DrawChildren(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i   : Integer;
      rgn : HRGN;
      FWasClippingRgn : Boolean;
      op  : TOxygenPlan;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if ((not CanBeParent) or (FChildren.Count = 0)) then Exit;
  if (poChildrenClipping in op.Options) then begin
    rgn := CreateRectRgn(0,0,0,0);
    i := GetClipRgn(ACanvas.Handle,rgn); // Old clipping region
    FWasClippingRgn := (i = 1);
    FClippingRgn := GetClippingRgn;
    if FWasClippingRgn then CombineRgn(FClippingRgn,rgn,FClippingRgn,RGN_AND);
    SelectClipRgn(ACanvas.Handle,FClippingRgn);
    DeleteObject(FClippingRgn);
    for i := 1 to FChildren.Count do TPlanObject(FChildren[i-1]).Draw(ACanvas, CWidth, CHeight);
    if FWasClippingRgn then SelectClipRgn(ACanvas.Handle,rgn) else SelectClipRgn(ACanvas.Handle,0);
    DeleteObject(rgn);
  end
  else for i := 1 to FChildren.Count do TPlanObject(FChildren[i-1]).Draw(ACanvas, CWidth, CHeight);
end;

function TPlanObject.GetClippingRgn : HRGN;
  var r:TRect;
begin
  r := GetDisplayRect;
  Result:=CreateRectRgnIndirect(r);
end;

procedure TPlanObject.SetReadOnly(const Value : Boolean);
begin
  FReadOnly:=Value;
  if Value then if Assigned((FOwner as TOxygenPlan).ActiveObject) then if ((FOwner as TOxygenPlan).ActiveObject = Self) then UnSelect;
end;

procedure TPlanObject.SetProperties(Value : TStrings);
  var i : Integer;
begin
  FProperties.Assign(Value);
  for i := 1 to FProperties.Count do FProperties[i-1] := LTrim(FProperties[i-1]);
end;

{$IFDEF USE_RX}
procedure TPlanObject.BlinkTimer(Sender: TObject);
begin
  FBlinkIndex := (FBlinkIndex xor 1);
  case FBlinkType of
    btNone  : Exit;
    btPen   : if (FBlinkIndex = 1) then FPen.Assign(FBlinkPen) else FPen.Assign(FSavePen);
    btBrush : if (FBlinkIndex = 1) then FBrush.Assign(FBlinkBrush) else FBrush.Assign(FSaveBrush);
  end;
  (FOwner as TOxygenPlan).DoObjectBlink(Self);
  if ((FLayer <> (FOwner as TOxygenPlan).CurrentLayer) or ((FBlinkType = btBrush) and (FBrush.Style = bsClear)) or
      (FBlinkType = btPen)) then NeedRepaint else Draw((FOwner as TOxygenPlan).Canvas, FOwner.Width, FOwner.Height);
end;
{$ENDIF}

procedure TPlanObject.SetPen(Value:TPen);
begin
  FPen.Assign(Value);
  FSelectColor:=-1;
  NeedRepaint;
end;

procedure TPlanObject.SetBrush(Value:TBrush);
begin
  FBrush.Assign(Value);
  NeedRepaint;
end;

procedure TPlanObject.SetLayer(Value : TPlanLayer);
  var i  : Integer;
      op : TOxygenPlan;
begin
  if ((FLayer = Value) or (Value = nil)) then Exit;
  op := FOwner as TOxygenPlan;
  if (op.ActiveObject = Self) then op.ActiveObject := nil;
  Unselect;
  op.DisableRepaint;
  try
    if (FLayer <> nil) then begin
      i := FLayer.IndexOf(Self);
      if (i >= 0) then FLayer.FObjectList.Delete(i);
    end;
    FLayer := Value;
    FLayer.Add(Self);
  finally
    op.EnableRepaint;
  end;
end;

procedure TPlanObject.SetName(Value : String);
  var i:Integer;
begin
  if (Value <> '') then for i:=1 to FLayer.Count do if ((UpperCase(FLayer.Objects[i-1].Name)=UpperCase(Value)) and (FLayer.Objects[i-1]<>Self)) then begin
    EOxygenPlanException.Create('Object named "'+Value+'" already exists');
    Exit;
  end;
  FName:=Value;
end;

function TPlanObject.GetObject(Index:Integer):TPlanObject;
begin
  Result := TPlanObject(FChildren[Index]);
end;

function TPlanObject.GetObjectsCount:Integer;
  var i:Integer;
begin
  Result := FChildren.Count;
  for i:=1 to FChildren.Count do Inc(Result,TPlanObject(FChildren[i-1]).ObjectsCount);
end;

procedure TPlanObject.SetParent(Value:TPlanObject);
begin
  if (FParent<>nil) then try FParent.FChildren.Delete(FParent.FChildren.IndexOf(Self)) except end;
  FParent:=Value;
  if (FParent<>nil) then FParent.FChildren.Add(Self);
end;

function TPlanObject.GetChildrenCount:Integer;
begin
  Result:=FChildren.Count;
end;

procedure TPlanObject.SetSelected(Value:Boolean);
begin
  if Value then Select else UnSelect;
end;

function TPlanObject.GetSelectedByColor:Boolean;
begin
  Result := (FSelectColor <> -1);
end;

procedure TPlanObject.SetVisible(const Value:Boolean);
begin
  if (FVisible = Value) then Exit;
  FVisible := Value;
  NeedRepaint;
end;

{$IFDEF USE_RX}
function TPlanObject.GetBlinking : Boolean;
begin
  Result:=(FBlinkTimerHandle<>0);
end;
{$ENDIF}

procedure TPlanObject.SaveCoords;
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].SaveCoords;
end;

procedure TPlanObject.EditMove(APoint:TPoint);
  var i : Integer;
begin
  for i := 1 to ChildrenCount do Objects[i-1].EditMove(APoint);
end;

procedure TPlanObject.Move(const DeltaX, DeltaY : Integer);
  var i : Integer;
begin
  for i := 1 to ChildrenCount do Objects[i-1].Move(DeltaX, DeltaY);
end;

procedure TPlanObject.Select;
begin
  if {FSelected or }(csDestroying in FOwner.ComponentState) then Exit;
  FSelected := True;
  SaveCoords;
  with (FOwner as TOxygenPlan) do begin
    ActiveObject := Self;
    SelectedObjects.Add(Self);
  end;
  NeedRepaint;
end;

procedure TPlanObject.UnSelect;
begin
  if (not FSelected) or (csDestroying in FOwner.ComponentState) then Exit;
  FSelected := False;
  (FOwner as TOxygenPlan).SelectedObjects.DeleteObject(Self);
  NeedRepaint;
end;

procedure TPlanObject.ParentResized;
begin
end;

procedure TPlanObject.SelectByColor(const AColor:TColor);
begin
  FSelectColor := AColor;
  NeedRepaint;
end;

procedure TPlanObject.SetTopObject(PlanObject : TPlanObject);
  var i : Integer;
begin
  i := FChildren.IndexOf(PlanObject);
  if (i = -1) then Exit;
  FChildren.Delete(i);
  FChildren.Add(PlanObject);
end;

procedure TPlanObject.UnSelectColor;
begin
  FSelectColor:=-1;
  NeedRepaint;
end;

{$IFDEF USE_RX}
procedure TPlanObject.StartBlink(const Interval:Integer);
begin
  StopBlink;
  FSavePen.Assign(FPen);
  FSaveBrush.Assign(FBrush);
  FBlinkTimerHandle := (FOwner as TOxygenPlan).AddTimerEvent(BlinkTimer,Interval,True);
end;

procedure TPlanObject.StopBlink;
begin
  if (FBlinkTimerHandle=0) then Exit;
  (FOwner as TOxygenPlan).DeleteTimerEvent(FBlinkTimerHandle);
  FBlinkType:=btNone;
  FPen.Assign(FSavePen);
  FBrush.Assign(FSaveBrush);
  FBlinkTimerHandle:=0;
  NeedRepaint;
end;
{$ENDIF}

procedure TPlanObject.FillObjectProperties;
  var ms:TMemoryStream;
begin
  FProperties.Values['Name']:=FName;
  FProperties.Values['ID']:=IntToStr(FID);
  FProperties.Values['Layer']:=IntToStr(FLayer.Index);
  FProperties.Values['PenColor']:=EncodeColor(FPen.Color);
  FProperties.Values['PenStyle']:=IntToStr(Integer(FPen.Style));
  FProperties.Values['PenMode']:=IntToStr(Integer(FPen.Mode));
  FProperties.Values['PenWidth']:=IntToStr(FPen.Width);
  FProperties.Values['BrushColor']:=EncodeColor(FBrush.Color);
  FProperties.Values['BrushStyle']:=IntToStr(Integer(FBrush.Style));
  FProperties.Values['Cursor']:=IntToStr(Integer(FCursor));
  if (FBrush.Bitmap<>nil) then begin
    ms:=TMemoryStream.Create;
    try
      FBrush.Bitmap.SaveToStream(ms);
      Properties.Values['BrushBitmap']:=Stream2String(ms);
    except end;
    ms.Free;
  end;
  FProperties.Values['ReadOnly'] := IntToStr(Integer(FReadOnly));
  FProperties.Values['GradientStartColor'] := EncodeColor(FGradientRecord.StartColor);
  FProperties.Values['GradientEndColor'] := EncodeColor(FGradientRecord.EndColor);
  FProperties.Values['GradientDirection'] := IntToStr(Integer(FGradientRecord.Direction));
  FProperties.Values['GradientColors'] := IntToStr(FGradientRecord.Colors);
  FProperties.Values['GradientActive'] := IntToStr(Integer(FGradientRecord.Active));
  FProperties.Values['Visible'] := IntToStr(Integer(FVisible));
  if (FSelectColor <> -1) or (FProperties.Values['SelectColor'] <> '') then FProperties.Values['SelectColor'] := EncodeColor(FSelectColor);
end;

procedure TPlanObject.AddObjectDescr(const SpaceNum:Integer; sl:TStrings);
  var s:String;
      i:Integer;
begin
  s:=MakeStr(' ',SpaceNum);
  sl.Add(s + sStartObject + GetObjectTypeDescr);
  for i:=1 to FProperties.Count do FProperties[i-1]:=LTrim(FProperties[i-1]);

  FillObjectProperties;
  for i:=1 to FProperties.Count do sl.Add(s+'  '+LTrim(FProperties[i-1]));
  //sl.AddStrings(FProperties);
  for i:=1 to FChildren.Count do TPlanObject(FChildren[i-1]).AddObjectDescr(SpaceNum+2,sl);
  sl.Add(s + sEndObject);
end;

procedure TPlanObject.UpdateObjectProperties;
  var i  : Integer;
      s  : String;
      ms : TMemoryStream;
      b  : TBitmap;
begin
  for i:=1 to FProperties.Count do FProperties[i-1] := LTrim(StrF.ReplaceText(FProperties[i-1],' = ','='));
  FName := FProperties.Values['Name'];
  s:=FProperties.Values['ID'];
  if (s<>'') then try FID:=StrToInt(s) except end;
  if (FLayer<>nil) then begin
    i:=FLayer.IndexOf(Self);
    if (i>=0) then FLayer.FObjectList.Delete(i);
  end;
  s := FProperties.Values['Layer'];
  if (s <> '') then try FLayer := (FOwner as TOxygenPlan).Layers[StrToInt(s)] except FLayer := (FOwner as TOxygenPlan).Layers[0] end;
  FLayer.Add(Self);
  s := FProperties.Values['PenColor'];
  if (s <> '') then try FPen.Color := ReadColor(s) except end;
  s := FProperties.Values['PenStyle'];
  if (s <> '') then try FPen.Style := TPenStyle(StrToInt(s)) except end;
  s := FProperties.Values['PenMode'];
  if (s <> '') then try FPen.Mode := TPenMode(StrToInt(s)) except end;
  s:=FProperties.Values['PenWidth'];
  if (s <> '') then try FPen.Width := StrToInt(s) except end;
  s := FProperties.Values['BrushColor'];
  if (s <> '') then try FBrush.Color := ReadColor(s) except end;
  s := FProperties.Values['BrushStyle'];
  if (s <> '') then try FBrush.Style := TBrushStyle(StrToInt(s)) except end else FBrush.Style := bsClear;
  s := FProperties.Values['Cursor'];
  if (s <> '') then try FCursor := TCursor(StrToInt(s)) except end else FCursor := crDefault;
  s := FProperties.Values['BrushBitmap'];
  if (s <> '') then begin
    ms := TMemoryStream.Create;
    b := TBitmap.Create;
    try
      String2Stream(s,ms);
      b.LoadFromStream(ms);
      FBrush.Bitmap := b;
    except end;
    ms.Free;
  end;
  s := FProperties.Values['ReadOnly'];
  if (s <> '') then try ReadOnly := (StrToInt(s) <> 0) except end;
  s := FProperties.Values['GradientStartColor'];
  if (s <> '') then try FGradientRecord.StartColor := ReadColor(s) except end;
  s := FProperties.Values['GradientEndColor'];
  if (s <> '') then try FGradientRecord.EndColor := ReadColor(s) except end;
  s := FProperties.Values['GradientDirection'];
  if (s <> '') then try FGradientRecord.Direction := TGradientFillDirection(StrToInt(s)) except end;
  s := FProperties.Values['GradientColors'];
  if (s <> '') then try FGradientRecord.Colors := StrToInt(s) except end;
  s := FProperties.Values['GradientActive'];
  if (s <> '') then try FGradientRecord.Active := (StrToInt(s) <> 0) except FGradientRecord.Active := False end;
  s := FProperties.Values['Visible'];
  if (s <> '') then try FVisible := (StrToInt(s) <> 0) except FVisible := True end;
  s := FProperties.Values['SelectColor'];
  if (s <> '') then try FSelectColor := ReadColor(s) except end;
end;

function TPlanObject.ObjectFromCoord(X,Y:Integer; var MarginType : TMarginType):TPlanObject;
  var i  : Integer;
      po : TPlanObject;
      mt : TMarginType;
begin
  Result := nil;
  MarginType := mtNone;
  if ((not Visible) and (FOwner as TOxygenPlan).ReadOnly) then Exit;
  for i:=FChildren.Count downto 1 do begin
    if Objects[i-1].HasMarginPoint(X,Y,mt) then begin
      Result:=Objects[i-1];
      MarginType:=mt;
      Exit;
    end
    else if Objects[i-1].HasInternalPoint(X,Y) then begin
      Result:=Objects[i-1];
      MarginType := mtNone;
      po:=Objects[i-1].ObjectFromCoord(X,Y,mt);
      if (po<>nil) then begin
        Result:=po;
        MarginType := mt;
      end;
      Exit;
    end;
  end;
end;

function TPlanObject.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  s := UpperCase(PropertyName);
  Result := (s = 'NAME') or (s = 'LAYER') or (s = 'ID') or (s = 'PENCOLOR') or (s = 'PENWIDTH') or (s = 'PENSTYLE') or
            (s = 'BRUSHCOLOR') or (s = 'BRUSHSTYLE') or (s = 'BRUSHBITMAP') or (s = 'READONLY') or (s = 'CURSOR') or
            (s = 'PENMODE') or (s = 'GRADIENTSTARTCOLOR') or (s = 'GRADIENTENDCOLOR') or (s = 'GRADIENTDIRECTION') or
            (s = 'GRADIENTCOLORS') or (s = 'GRADIENTACTIVE') or (s = 'VISIBLE') or (s = 'SELECTCOLOR');
end;

procedure TPlanObject.Free;
  var i : Integer;
begin
  FVisible := False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying := True;
{$IFDEF USE_RX}
  StopBlink;
{$ENDIF}
  with (FOwner as TOxygenPlan) do try
    if (ActiveObject = Self) then ActiveObject := nil;
    i := SelectedObjects.IndexOf(Self);
    if (i >= 0) then SelectedObjects.Delete(i);
  except end;
  for i:=FChildren.Count downto 1 do TPlanObject(FChildren[i-1]).Free;
  FChildren.Free;
  i:=FLayer.IndexOf(Self);
  if (i>=0) then FLayer.FObjectList.Delete(i);
  FPen.Free;
  //if (FBrush.Bitmap<>nil) then FBrush.Bitmap.Free;
  FBrush.Free;
  FProperties.Free;
  FGradientRecord.Free;
{$IFDEF USE_RX}
  FBlinkPen.Free;
  FSavePen.Free;
  FBlinkBrush.Free;
  FSaveBrush.Free;
{$ENDIF}
  SetParent(nil);
  inherited Free;
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanLine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanLine.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  Fx1:=0;
  Fy1:=0;
  Fx2:=0;
  Fy2:=0;
end;

function TPlanLine.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanLine.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanLine) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanLine.SetX1(Value:Integer);
begin
  if (Fx1 = Value) then Exit;
  NeedRepaint;
  Fx1 := Value;
  NeedRepaint;
end;

procedure TPlanLine.SetX2(Value:Integer);
begin
  if (Fx2 = Value) then Exit;
  NeedRepaint;
  Fx2 := Value;
  NeedRepaint;
end;

procedure TPlanLine.SetY1(Value:Integer);
begin
  if (Fy1 = Value) then Exit;
  NeedRepaint;
  Fy1 := Value;
  NeedRepaint;
end;

procedure TPlanLine.SetY2(Value:Integer);
begin
  if (Fy2 = Value) then Exit;
  NeedRepaint;
  Fy2 := Value;
  NeedRepaint;
end;

function TPlanLine.GetScrPoint1:TPoint;
begin
  Result := Point(Trunc(Fx1*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
end;

function TPlanLine.GetScrPoint2:TPoint;
begin
  Result := Point(Trunc(Fx2*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
end;

procedure TPlanLine.SetScrPoint1(p:TPoint);
  var sp : TPoint;
begin
  sp := ScrPoint1;
  if (sp.X <> p.X) and (FOwner.Width > 0) then Fx1 := Trunc(p.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
  if (sp.Y <> p.Y) and (FOwner.Height > 0) then Fy1 := Trunc(p.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height));
end;

procedure TPlanLine.SetScrPoint2(p:TPoint);
  var sp : TPoint;
begin
  sp := ScrPoint2;
  if (sp.X <> p.X) and (FOwner.Width > 0) then Fx2 := Trunc(p.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
  if (sp.Y <> p.Y) and (FOwner.Height > 0) then Fy2 := Trunc(p.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height));
end;

procedure TPlanLine.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
      op : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    PolyLine([p1, p2]);
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      ACanvas.Pen.Color := op.SelectColor;
      ACanvas.Pen.Width := 1;
      ACanvas.Brush.Color := op.SelectColor;
      ACanvas.Brush.Style := bsSolid;
      Rectangle(p1.X - MarginDelta, p1.Y - MarginDelta, p1.X + MarginDelta, p1.Y + MarginDelta);
      Rectangle(p2.X - MarginDelta, p2.Y - MarginDelta, p2.X + MarginDelta, p2.Y + MarginDelta);
    end;
  end;
end;

function TPlanLine.HasInternalPoint(X,Y:Integer):Boolean;
begin
  Result := PointOnLine(Point(X,Y), ScrPoint1, ScrPoint2, Pen.Width);
end;

function TPlanLine.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'X1') or (s = 'X2') or (s = 'Y1') or (s = 'Y2');
end;

function TPlanLine.HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean;
  var p1,p2:TPoint;
begin
  MarginType:=mtNone;
  p1:=ScrPoint1;
  p2:=ScrPoint2;
  Result:=PointInRect(Point(X,Y),Rect(p1.X-MarginDelta,p1.Y-MarginDelta,p1.X+MarginDelta,p1.Y+MarginDelta));
  if Result then begin
    if (p1.X<p2.X) then MarginType:=mtLeft else MarginType:=mtTop;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p2.X-MarginDelta,p2.Y-MarginDelta,p2.X+MarginDelta,p2.Y+MarginDelta));
  if Result then if (p1.X<p2.X) then MarginType:=mtRight else MarginType:=mtBottom;
end;

procedure TPlanLine.EditMove(APoint:TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
      clr:HRGN;
      p1,p2:TPoint;
      op : TOxygenPlan;
begin
  r := GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1:=CreateRectRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  p1 := Point(ScrSave1.X + (APoint.X - op.StartCoord.X), ScrSave1.Y + (APoint.Y - op.StartCoord.Y));
  p2 := Point(ScrSave2.X + (APoint.X - op.StartCoord.X), ScrSave2.Y + (APoint.Y - op.StartCoord.Y));
  if (FParent <> nil) and (poChildrenClipping in op.Options) then begin
    clr := FParent.GetClippingRgn;
    if not (PtInRegion(clr,p1.X,p1.Y) or PtInRegion(clr,p2.X,p2.Y)) then begin
      DeleteObject(rgn1);
      DeleteObject(clr);
      Exit;
    end;
    DeleteObject(clr);
  end;
  ScrPoint1:=p1;
  ScrPoint2:=p2;

  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2:=CreateRectRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;

  inherited EditMove(APoint);
end;

procedure TPlanLine.EditResize(APoint : TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
begin
  if not FCanResize then Exit;
  r:=GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1:=CreateRectRgnIndirect(r);

  if ((FOwner as TOxygenPlan).ActiveMarginType in [mtLeft,mtTop]) then ScrPoint1:=APoint else ScrPoint2:=APoint;

  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2:=CreateRectRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;
end;

procedure TPlanLine.Move(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx1, DeltaX);
  Inc(Fx2, DeltaX);
  Inc(Fy1, DeltaY);
  Inc(Fy2, DeltaY);
  NeedRepaint;
  inherited Move(DeltaX, DeltaY);
end;

procedure TPlanLine.Resize(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx2, DeltaX);
  Inc(Fy2, DeltaY);
  NeedRepaint;
end;

function TPlanLine.CanBeParent : Boolean;
begin
  Result:=False;
end;

function TPlanLine.GetObjectTypeDescr : String;
begin
  Result:=ocLine;
end;

procedure TPlanLine.FillObjectProperties;
begin
  inherited FillObjectProperties;
  FProperties.Values['X1']:=IntToStr(Fx1);
  FProperties.Values['Y1']:=IntToStr(Fy1);
  FProperties.Values['X2']:=IntToStr(Fx2);
  FProperties.Values['Y2']:=IntToStr(Fy2);
end;

procedure TPlanLine.SaveCoords;
begin
  ScrSave1:=ScrPoint1;
  ScrSave2:=ScrPoint2;
  inherited SaveCoords;
end;

procedure TPlanLine.NeedRepaint;
  var p1,p2:TPoint;
      r:TRect;
begin
  if FDestroying then Exit;
  p1:=ScrPoint1;
  p2:=ScrPoint2;
  r:=NormalizeRect(p1.X,p1.Y,p2.X,p2.Y);
  r:=Rect(r.Left - MarginDelta - Pen.Width, r.Top - MarginDelta - Pen.Width,
          r.Right + MarginDelta + Pen.Width, r.Bottom + MarginDelta + Pen.Width);
  Windows.InvalidateRect(FOwner.Handle,@r,False);
end;

function TPlanLine.GetDisplayRect:TRect;
begin
  Result.TopLeft := ScrPoint1;
  Result.BottomRight := ScrPoint2;
  Result := NormalizeRect(Result.Left,Result.Top,Result.Right,Result.Bottom);
  Dec(Result.Top, Pen.Width div 2);
  Dec(Result.Left, Pen.Width div 2);
  Inc(Result.Bottom, Pen.Width div 2);
  Inc(Result.Right, Pen.Width div 2);
end;

function TPlanLine.AddPoint(const CoordX,CoordY:Integer; CallHandler:TCallHandler):Boolean;
begin
  Result := ((FOwner as TOxygenPlan).CurDrawingPointNum < 2);
  if not Result then Exit;
  NeedRepaint;
  case (FOwner as TOxygenPlan).CurDrawingPointNum of
    0: begin ScrPoint1:=Point(CoordX,CoordY); ScrPoint2:=Point(CoordX,CoordY); end;
    1: ScrPoint2:=Point(CoordX,CoordY);
  end;
  NeedRepaint;
end;

function TPlanLine.GetClippingRgn : HRGN;
begin
  Result := CreateRectRgn(0,0,0,0);
end;

procedure TPlanLine.UpdateObjectProperties;
  var s:String;
begin
  inherited UpdateObjectProperties;
  s:=FProperties.Values['X1'];
  if (s<>'') then try Fx1:=StrToInt(s) except end;
  s:=FProperties.Values['Y1'];
  if (s<>'') then try Fy1:=StrToInt(s) except end;
  s:=FProperties.Values['X2'];
  if (s<>'') then try Fx2:=StrToInt(s) except end;
  s:=FProperties.Values['Y2'];
  if (s<>'') then try Fy2:=StrToInt(s) except end;
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanBox
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanBox.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  Fx1:=0;
  Fy1:=0;
  Fx2:=0;
  Fy2:=0;
end;

function TPlanBox.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanBox.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanBox) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanBox.SetX1(Value:Integer);
begin
  if (Fx1=Value) then Exit;
  NeedRepaint;
  Fx1:=Value;
  NeedRepaint;
end;

procedure TPlanBox.SetX2(Value:Integer);
begin
  if (Fx2=Value) then Exit;
  NeedRepaint;
  Fx2:=Value;
  NeedRepaint;
end;

procedure TPlanBox.SetY1(Value:Integer);
begin
  if (Fy1=Value) then Exit;
  NeedRepaint;
  Fy1:=Value;
  NeedRepaint;
end;

procedure TPlanBox.SetY2(Value:Integer);
begin
  if (Fy2=Value) then Exit;
  NeedRepaint;
  Fy2:=Value;
  NeedRepaint;
end;

function TPlanBox.GetScrPoint1:TPoint;
begin
  Result:=Point(Trunc(Fx1*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
end;

function TPlanBox.GetScrPoint2:TPoint;
begin
  Result:=Point(Trunc(Fx2*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
end;

procedure TPlanBox.SetScrPoint1(p:TPoint);
  var sp : TPoint;
begin
  sp := ScrPoint1;
  if (sp.X <> p.X) and (FOwner.Width > 0) then Fx1 := Trunc(p.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
  if (sp.Y <> p.Y) and (FOwner.Height > 0) then Fy1 := Trunc(p.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height));
end;

procedure TPlanBox.SetScrPoint2(p:TPoint);
  var sp : TPoint;
begin
  sp := ScrPoint2;
  if (sp.X <> p.X) and (FOwner.Width > 0) then Fx2 := Trunc(p.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
  if (sp.Y <> p.Y) and (FOwner.Height > 0) then Fy2 := Trunc(p.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height));
end;

procedure TPlanBox.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
      op : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    if FGradientRecord.Active then begin
      PlanUtils.GradientFillRect(ACanvas,Rect(p1.X + Pen.Width, p1.Y + Pen.Width, p2.X - Pen.Width, p2.Y - Pen.Width),
                                 FGradientRecord.StartColor,FGradientRecord.EndColor,FGradientRecord.Direction,
                                 FGradientRecord.Colors);
      Brush.Style := bsClear;
    end;
    Rectangle(p1.X, p1.Y, p2.X, p2.Y);

    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then SelectRectRgn(ACanvas,op.SelectColor,p1,p2);
  end;
  DrawChildren(ACanvas, CWidth, CHeight);
end;

function TPlanBox.HasInternalPoint(X,Y:Integer):Boolean;
  var p1,p2:TPoint;
begin
  p1:=ScrPoint1;
  p2:=ScrPoint2;
  Result:=PointInRect(Point(X,Y),NormalizeRect(p1.X,p1.Y,p2.X,p2.Y));
end;

function TPlanBox.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'X1') or (s = 'X2') or (s = 'Y1') or (s = 'Y2');
end;

function TPlanBox.HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean;
  var p1,p2:TPoint;
begin
  MarginType:=mtNone;
  p1:=ScrPoint1;
  p2:=ScrPoint2;
  Result:=PointInRect(Point(X,Y),Rect(p1.X-MarginDelta,p1.Y-MarginDelta,p1.X+MarginDelta,p1.Y+MarginDelta));
  if Result then begin
    MarginType:=mtTopLeft;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p2.X-MarginDelta,p1.Y-MarginDelta,p2.X+MarginDelta,p1.Y+MarginDelta));
  if Result then begin
    MarginType:=mtTopRight;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p2.X-MarginDelta,p2.Y-MarginDelta,p2.X+MarginDelta,p2.Y+MarginDelta));
  if Result then begin
    MarginType:=mtBottomRight;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p1.X-MarginDelta,p2.Y-MarginDelta,p1.X+MarginDelta,p2.Y+MarginDelta));
  if Result then begin
    MarginType:=mtBottomLeft;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),NormalizeRect(p1.X,p1.Y-MarginDelta,p2.X,p1.Y+MarginDelta));
  if Result then begin
    MarginType:=mtTop;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p2.X-MarginDelta,p1.Y,p2.X+MarginDelta,p2.Y));
  if Result then begin
    MarginType:=mtRight;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),NormalizeRect(p1.X,p2.Y-MarginDelta,p2.X,p2.Y+MarginDelta));
  if Result then begin
    MarginType:=mtBottom;
    Exit;
  end;
  Result:=PointInRect(Point(X,Y),Rect(p1.X-MarginDelta,p1.Y,p1.X+MarginDelta,p2.Y));
  if Result then begin
    MarginType:=mtLeft;
    Exit;
  end;
end;

procedure TPlanBox.EditMove(APoint : TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
      clr:HRGN;
      p1,p2:TPoint;
      op : TOxygenPlan;
begin
  r := GetDisplayRect;
  Dec(r.Left,MarginDelta*2 + Pen.Width);
  Dec(r.Top,MarginDelta*2 + Pen.Width);
  Inc(r.Right,MarginDelta*2 + Pen.Width);
  Inc(r.Bottom,MarginDelta*2 + Pen.Width);
  rgn1 := CreateRectRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  p1:=Point(ScrSave1.X+(APoint.X-op.StartCoord.X),ScrSave1.Y+(APoint.Y-op.StartCoord.Y));
  p2:=Point(ScrSave2.X+(APoint.X-op.StartCoord.X),ScrSave2.Y+(APoint.Y-op.StartCoord.Y));
  if (FParent <> nil) and (poChildrenClipping in op.Options) then begin
    clr:=FParent.GetClippingRgn;
    if not (PtInRegion(clr,p1.X,p1.Y) or PtInRegion(clr,p2.X,p2.Y)) then begin
      DeleteObject(rgn1);
      DeleteObject(clr);
      Exit;
    end;
    DeleteObject(clr);
  end;
  ScrPoint1:=p1;
  ScrPoint2:=p2;

  r := GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2:=CreateRectRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;

  inherited EditMove(APoint);
end;

procedure TPlanBox.EditResize(APoint:TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
      op : TOxygenPlan;
begin
  if not FCanResize then Exit;
  r:=GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1:=CreateRectRgnIndirect(r);

  op := FOwner  as TOxygenPlan;
  case op.ActiveMarginType of
    mtLeft:        ScrPoint1:=Point(APoint.X,ScrPoint1.Y);
    mtTopLeft:     ScrPoint1:=APoint;
    mtTop:         ScrPoint1:=Point(ScrPoint1.X,APoint.Y);
    mtTopRight:    begin ScrPoint1:=Point(ScrPoint1.X,APoint.Y); ScrPoint2:=Point(APoint.X,ScrPoint2.Y); end;
    mtRight:       ScrPoint2:=Point(APoint.X,ScrPoint2.Y);
    mtBottomRight: ScrPoint2:=APoint;
    mtBottom:      ScrPoint2:=Point(ScrPoint2.X,APoint.Y);
    mtBottomLeft:  begin ScrPoint1:=Point(APoint.X,ScrPoint1.Y); ScrPoint2:=Point(ScrPoint2.X,APoint.Y); end;
  end;
  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2:=CreateRectRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;
end;

procedure TPlanBox.Move(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx1, DeltaX);
  Inc(Fx2, DeltaX);
  Inc(Fy1, DeltaY);
  Inc(Fy2, DeltaY);
  NeedRepaint;
  inherited Move(DeltaX, DeltaY);
end;

procedure TPlanBox.Resize(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx2, DeltaX);
  Inc(Fy2, DeltaY);
  NeedRepaint;
end;

function TPlanBox.CanBeParent : Boolean;
begin
  Result:=True;
end;

function TPlanBox.GetObjectTypeDescr : String;
begin
  Result := ocBox;
end;

procedure TPlanBox.FillObjectProperties;
begin
  inherited FillObjectProperties;
  FProperties.Values['X1']:=IntToStr(Fx1);
  FProperties.Values['Y1']:=IntToStr(Fy1);
  FProperties.Values['X2']:=IntToStr(Fx2);
  FProperties.Values['Y2']:=IntToStr(Fy2);
end;

procedure TPlanBox.SaveCoords;
begin
  ScrSave1:=ScrPoint1;
  ScrSave2:=ScrPoint2;
  inherited SaveCoords;
end;

procedure TPlanBox.NeedRepaint;
  var p1,p2:TPoint;
      r:TRect;
begin
  if FDestroying then Exit;
  p1:=ScrPoint1;
  p2:=ScrPoint2;
  r:=NormalizeRect(p1.X,p1.Y,p2.X,p2.Y);
  r:=Rect(r.Left - MarginDelta - Pen.Width, r.Top - MarginDelta - Pen.Width,
          r.Right + MarginDelta + Pen.Width, r.Bottom + MarginDelta + Pen.Width);
  Windows.InvalidateRect(FOwner.Handle,@r,False);
end;

function TPlanBox.GetDisplayRect:TRect;
begin
  Result.TopLeft:=ScrPoint1;
  Result.BottomRight:=ScrPoint2;
  Result := NormalizeRect(Result.Left,Result.Top,Result.Right,Result.Bottom);
  Dec(Result.Top, Pen.Width div 2);
  Dec(Result.Left, Pen.Width div 2);
  Inc(Result.Bottom, Pen.Width div 2);
  Inc(Result.Right, Pen.Width div 2);
end;

function TPlanBox.AddPoint(const CoordX,CoordY:Integer; CallHandler:TCallHandler):Boolean;
  var op : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  Result := (op.CurDrawingPointNum < 2);
  if not Result then Exit;
  NeedRepaint;
  case op.CurDrawingPointNum of
    0: begin ScrPoint1 := Point(CoordX,CoordY); ScrPoint2:=Point(CoordX,CoordY); end;
    1: ScrPoint2 := Point(CoordX,CoordY);
  end;
  NeedRepaint;
end;

procedure TPlanBox.UpdateObjectProperties;
  var s:String;
begin
  inherited UpdateObjectProperties;
  s:=FProperties.Values['X1'];
  if (s<>'') then try Fx1:=StrToInt(s) except end;
  s:=FProperties.Values['Y1'];
  if (s<>'') then try Fy1:=StrToInt(s) except end;
  s:=FProperties.Values['X2'];
  if (s<>'') then try Fx2:=StrToInt(s) except end;
  s:=FProperties.Values['Y2'];
  if (s<>'') then try Fy2:=StrToInt(s) except end;
end;


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanEllipse
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TPlanEllipse.GetClippingRgn : HRGN;
  var r : TRect;
begin
  r := GetDisplayRect;
  Inc(r.Left, MarginDelta + FPen.Width);
  Inc(r.Top, MarginDelta + FPen.Width);
  Dec(r.Right, MarginDelta + FPen.Width);
  Dec(r.Bottom, MarginDelta + FPen.Width);
  Result := CreateEllipticRgnIndirect(r);
end;

function TPlanEllipse.GetObjectTypeDescr : String;
begin
  Result := ocEllipse;
end;

function TPlanEllipse.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanEllipse.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanEllipse) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanEllipse.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1, p2 : TPoint;
      dx, dy : Integer;
      op : TOxygenPlan;
      rgn : HRGN;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount>0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  dx := (p2.X - p1.X) div 2;
  dy := (p2.Y - p1.Y) div 2;
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    if FGradientRecord.Active then begin
      rgn := GetClippingRgn;
      PlanUtils.GradientFillRgn(ACanvas, DisplayRect, rgn, FGradientRecord.StartColor, FGradientRecord.EndColor, FGradientRecord.Direction, FGradientRecord.Colors);
      DeleteObject(rgn);
      Brush.Style := bsClear;
    end;
    Ellipse(p1.X, p1.Y, p2.X, p2.Y);
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      ACanvas.Pen.Color := op.SelectColor;
      ACanvas.Brush.Color := op.SelectColor;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Pen.Width := 1;
      Rectangle(p1.X + dx - MarginDelta, p1.Y - MarginDelta, p1.X + dx + MarginDelta, p1.Y + MarginDelta);
      Rectangle(p2.X - MarginDelta, p1.Y + dy - MarginDelta, p2.X + MarginDelta, p1.Y + dy + MarginDelta);
      Rectangle(p1.X + dx - MarginDelta, p2.Y - MarginDelta, p1.X + dx + MarginDelta, p2.Y + MarginDelta);
      Rectangle(p1.X - MarginDelta, p1.Y + dy - MarginDelta, p1.X + MarginDelta, p1.Y + dy + MarginDelta);
    end;
  end;
  DrawChildren(ACanvas, CWidth, CHeight);
end;

function TPlanEllipse.HasInternalPoint(X, Y : Integer) : Boolean;
  var rgn : HRGN;
begin
  rgn := GetClippingRgn;
  Result := PtInRegion(rgn,X,Y);
  DeleteObject(rgn);
end;

function TPlanEllipse.HasMarginPoint(X,Y:Integer; var MarginType : TMarginType) : Boolean;
  var p1, p2 : TPoint;
      dx, dy : Integer;
begin
  MarginType := mtNone;
  Result := False;
  p1 := ScrPoint1;
  p2 := ScrPoint2;
  dx := (p2.X - p1.X) div 2;
  dy := (p2.Y - p1.Y) div 2;
  if PointInRect(Point(X,Y),Rect(p1.X + dx - MarginDelta, p1.Y - MarginDelta, p1.X + dx + MarginDelta, p1.Y + MarginDelta)) then begin
    Result := True;
    MarginType := mtTop;
  end
  else if PointInRect(Point(X,Y),Rect(p2.X - MarginDelta, p1.Y + dy - MarginDelta, p2.X + MarginDelta, p1.Y + dy + MarginDelta)) then begin
    Result := True;
    MarginType:=mtRight;
  end
  else if PointInRect(Point(X,Y),Rect(p1.X + dx - MarginDelta, p2.Y - MarginDelta, p1.X + dx + MarginDelta, p2.Y + MarginDelta)) then begin
    Result := True;
    MarginType:=mtBottom;
  end
  else if PointInRect(Point(X,Y),Rect(p1.X - MarginDelta, p1.Y + dy - MarginDelta, p1.X + MarginDelta, p1.Y + dy + MarginDelta)) then begin
    Result := True;
    MarginType:=mtLeft;
  end;
end;

procedure TPlanEllipse.EditMove(APoint:TPoint);
  var rgn1, rgn2 : HRGN;
      r : TRect;
      clr : HRGN;
      p1,p2 : TPoint;
      i : Integer;
      op : TOxygenPlan;
begin
  r := GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1 := CreateEllipticRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  p1 := Point(ScrSave1.X + (APoint.X - op.StartCoord.X), ScrSave1.Y + (APoint.Y - op.StartCoord.Y));
  p2 := Point(ScrSave2.X + (APoint.X - op.StartCoord.X), ScrSave2.Y + (APoint.Y - op.StartCoord.Y));
  if (FParent <> nil) and (poChildrenClipping in op.Options) then begin
    clr := FParent.GetClippingRgn;
    if not (PtInRegion(clr,p1.X,p1.Y) or PtInRegion(clr,p2.X,p2.Y)) then begin
      DeleteObject(rgn1);
      DeleteObject(clr);
      Exit;
    end;
    DeleteObject(clr);
  end;
  ScrPoint1 := p1;
  ScrPoint2 := p2;

  r := GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2 := CreateEllipticRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF) <> NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;

  for i := 1 to ChildrenCount do Objects[i-1].EditMove(APoint);
end;

procedure TPlanEllipse.NeedRepaint;
  var p1, p2 : TPoint;
      rgn : HRgn;
begin
  if FDestroying then Exit;
  p1 := ScrPoint1;
  p2 := ScrPoint2;
  rgn := CreateEllipticRgn(p1.X - MarginDelta - Pen.Width, p1.Y - MarginDelta - Pen.Width,
                           p2.X + MarginDelta + Pen.Width, p2.Y + MarginDelta + Pen.Width);
  Windows.InvalidateRgn(FOwner.Handle, rgn, False);
  DeleteObject(rgn);
end;

procedure TPlanEllipse.EditResize(APoint:TPoint);
  var rgn1, rgn2 : HRGN;
      r : TRect;
      op : TOxygenPlan;
begin
  if not FCanResize then Exit;
  r := GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1 := CreateEllipticRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  case op.ActiveMarginType of
    mtLeft   : if (APoint.X + MarginDelta*2 < ScrPoint2.X) then ScrPoint1 := Point(APoint.X,ScrPoint1.Y);
    mtTop    : if (APoint.Y + MarginDelta*2 < ScrPoint2.Y) then ScrPoint1 := Point(ScrPoint1.X,APoint.Y);
    mtRight  : if (APoint.X - MarginDelta*2 > ScrPoint1.X) then ScrPoint2 := Point(APoint.X,ScrPoint2.Y);
    mtBottom : if (APoint.Y - MarginDelta*2 > ScrPoint1.Y) then ScrPoint2 := Point(ScrPoint2.X,APoint.Y);
  end;
  r := GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2 := CreateEllipticRgnIndirect(r);
  if (CombineRgn(rgn1, rgn1, rgn2, RGN_DIFF) <> NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;
end;

procedure TPlanEllipse.Move(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx1, DeltaX);
  Inc(Fx2, DeltaX);
  Inc(Fy1, DeltaY);
  Inc(Fy2, DeltaY);
  NeedRepaint;
  inherited Move(DeltaX, DeltaY);
end;

procedure TPlanEllipse.Resize(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx2, DeltaX);
  Inc(Fy2, DeltaY);
  NeedRepaint;
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanCircle
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanCircle.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  Fx:=0;
  Fy:=0;
  FRadius:=0;
end;

function TPlanCircle.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanCircle.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanCircle) do begin
    X := X + 4;
    Y := Y + 4;
  end;
end;

procedure TPlanCircle.SetX(Value:Integer);
begin
  if (Fx=Value) then Exit;
  NeedRepaint;
  Fx:=Value;
  NeedRepaint;
end;

procedure TPlanCircle.SetY(Value:Integer);
begin
  if (Fy=Value) then Exit;
  NeedRepaint;
  Fy:=Value;
  NeedRepaint;
end;

procedure TPlanCircle.SetRadius(Value:Integer);
begin
  if (FRadius=Value) then Exit;
  NeedRepaint;
  FRadius:=Value;
  NeedRepaint;
end;

function TPlanCircle.GetScrCenter:TPoint;
begin
  Result := Point(Trunc(Fx*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
end;

function TPlanCircle.GetScrRadius:Integer;
begin
  Result := Trunc(FRadius*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX));
end;

procedure TPlanCircle.SetScrCenter(p:TPoint);
  var sc : TPoint;
begin
  sc := ScrCenter;
  if (sc.X <> p.X) and (FOwner.Width > 0) then Fx := Trunc(p.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
  if (sc.Y <> p.Y) and (FOwner.Height > 0) then Fy := Trunc(p.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height));
end;

procedure TPlanCircle.SetScrRadius(r:Integer);
begin
  if (FOwner.Width > 0) then FRadius := Trunc(r*((FOwner as TOxygenPlan).ScaleX/FOwner.Width));
end;

procedure TPlanCircle.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p : TPoint;
      r : Integer;
      op : TOxygenPlan;
      rgn : HRGN;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount>0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p := Point(Trunc(Fx*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  r := Trunc(FRadius*(CWidth/(FOwner as TOxygenPlan).ScaleX));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    if FGradientRecord.Active then begin
      rgn := GetClippingRgn;
      PlanUtils.GradientFillRgn(ACanvas, DisplayRect, rgn, FGradientRecord.StartColor, FGradientRecord.EndColor, FGradientRecord.Direction, FGradientRecord.Colors);
      DeleteObject(rgn);
      Brush.Style := bsClear;
    end;
    Ellipse(p.X - r, p.Y - r, p.X + r, p.Y + r);
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      ACanvas.Pen.Color := op.SelectColor;
      ACanvas.Brush.Color := op.SelectColor;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Pen.Width := 1;
      Rectangle(p.X - MarginDelta, p.Y - r - MarginDelta, p.X + MarginDelta, p.Y - r + MarginDelta);
      Rectangle(p.X + r - MarginDelta, p.Y - MarginDelta, p.X + r + MarginDelta, p.Y + MarginDelta);
      Rectangle(p.X - MarginDelta, p.Y + r - MarginDelta, p.X + MarginDelta, p.Y + r + MarginDelta);
      Rectangle(p.X - r - MarginDelta, p.Y - MarginDelta, p.X - r + MarginDelta, p.Y + MarginDelta);
    end;
  end;
  DrawChildren(ACanvas, CWidth, CHeight);
end;

function TPlanCircle.HasInternalPoint(X,Y:Integer):Boolean;
  var d:Integer;
begin
  d := Distance(ScrCenter,Point(X,Y));
  Result := (d<=ScrRadius);
end;

function TPlanCircle.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'X') or (s = 'Y') or (s = 'RADIUS');
end;

function TPlanCircle.HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean;
  var p:TPoint;
      r,d:Integer;
begin
  MarginType:=mtNone;
  p:=ScrCenter;
  r:=ScrRadius;
  d:=Distance(p,Point(X,Y));
  Result:=((d>r-MarginDelta) and (d<r+MarginDelta));
  if not Result then Exit;
  if PointInRect(Point(X,Y),Rect(p.X-MarginDelta,p.Y-r-MarginDelta,p.X+MarginDelta,p.Y-r+MarginDelta)) then begin
    MarginType:=mtTop;
    Exit;
  end;
  if PointInRect(Point(X,Y),Rect(p.X+r-MarginDelta,p.Y-MarginDelta,p.X+r+MarginDelta,p.Y+MarginDelta)) then begin
    MarginType:=mtRight;
    Exit;
  end;
  if PointInRect(Point(X,Y),Rect(p.X-MarginDelta,p.Y+r-MarginDelta,p.X+MarginDelta,p.Y+r+MarginDelta)) then begin
    MarginType:=mtBottom;
    Exit;
  end;
  if PointInRect(Point(X,Y),Rect(p.X-r-MarginDelta,p.Y-MarginDelta,p.X+r+MarginDelta,p.Y+MarginDelta)) then begin
    MarginType:=mtLeft;
    Exit;
  end;
  if (X<p.X) and (Y<p.Y) then begin
    MarginType:=mtTopLeft;
    Exit;
  end;
  if (X>p.X) and (Y<p.Y) then begin
    MarginType:=mtTopRight;
    Exit;
  end;
  if (X<p.X) and (Y>p.Y) then begin
    MarginType:=mtBottomLeft;
    Exit;
  end;
  if (X>p.X) and (Y>p.Y) then begin
    MarginType:=mtBottomRight;
    Exit;
  end;
end;

procedure TPlanCircle.EditMove(APoint:TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
      clr:HRGN;
      p:TPoint;
      op : TOxygenPlan;
begin
  r:=GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1:=CreateEllipticRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  p:=Point(ScrSaveCenter.X + (APoint.X - op.StartCoord.X), ScrSaveCenter.Y + (APoint.Y - op.StartCoord.Y));
  if (FParent <> nil) and (poChildrenClipping in op.Options) then begin
    clr:=FParent.GetClippingRgn;
    if not PtInRegion(clr,p.X,p.Y) then begin
      DeleteObject(rgn1);
      DeleteObject(clr);
      Exit;
    end;
    DeleteObject(clr);
  end;
  ScrCenter:=p;

  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2:=CreateEllipticRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;

  inherited EditMove(APoint);
end;

procedure TPlanCircle.EditResize(APoint:TPoint);
  var rgn1,rgn2:HRGN;
      r:TRect;
      op : TOxygenPlan;
begin
  if not FCanResize then Exit;
  r := GetDisplayRect;
  Dec(r.Left,MarginDelta + Pen.Width);
  Dec(r.Top,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  rgn1 := CreateEllipticRgnIndirect(r);

  op := FOwner as TOxygenPlan;
  if (op.ActiveMarginType <> mtNone) then ScrRadius := Distance(ScrCenter,APoint);

  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta + Pen.Width);
  Inc(r.Top,MarginDelta + Pen.Width);
  Dec(r.Right,MarginDelta + Pen.Width);
  Dec(r.Bottom,MarginDelta + Pen.Width);
  rgn2 := CreateEllipticRgnIndirect(r);
  if (CombineRgn(rgn1,rgn1,rgn2,RGN_DIFF)<>NULLREGION) then InvalidateRgn(FOwner.Handle,rgn1,False);
  DeleteObject(rgn1);
  DeleteObject(rgn2);
  NeedRepaint;
end;

procedure TPlanCircle.Move(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(Fx, DeltaX);
  Inc(Fy, DeltaY);
  NeedRepaint;
  inherited Move(DeltaX, DeltaY);
end;

procedure TPlanCircle.Resize(const DeltaX, DeltaY : Integer);
begin
  NeedRepaint;
  Inc(FRadius, DeltaX);
  NeedRepaint;
end;

function TPlanCircle.CanBeParent : Boolean;
begin
  Result:=True;
end;

function TPlanCircle.GetObjectTypeDescr : String;
begin
  Result:=ocCircle;
end;

procedure TPlanCircle.FillObjectProperties;
begin
  inherited FillObjectProperties;
  Properties.Values['X']:=IntToStr(Fx);
  Properties.Values['Y']:=IntToStr(Fy);
  Properties.Values['Radius']:=IntToStr(FRadius);
end;

procedure TPlanCircle.SaveCoords;
begin
  ScrSaveCenter:=ScrCenter;
  ScrSaveRadius:=ScrRadius;
  inherited SaveCoords;
end;

procedure TPlanCircle.NeedRepaint;
  var p:TPoint;
      rad:Integer;
      rgn:HRgn;
begin
  if FDestroying then Exit;
  p:=ScrCenter;
  rad:=Abs(ScrRadius)+FPen.WIdth;
  rgn:=CreateEllipticRgn(p.X - rad - MarginDelta - Pen.Width, p.Y - rad - MarginDelta - Pen.Width,
                         p.X + rad + MarginDelta + Pen.Width, p.Y + rad + MarginDelta + Pen.Width);
  Windows.InvalidateRgn(FOwner.Handle,rgn,False);
  DeleteObject(rgn);
end;

function TPlanCircle.GetDisplayRect:TRect;
  var p:TPoint;
      rad:Integer;
begin
  p:=ScrCenter;
  rad:=Abs(ScrRadius);
  Result:=Rect(p.X-rad,p.Y-rad,p.X+rad,p.Y+rad);
  Dec(Result.Top, Pen.Width div 2);
  Dec(Result.Left, Pen.Width div 2);
  Inc(Result.Bottom, Pen.Width div 2);
  Inc(Result.Right, Pen.Width div 2);
end;

function TPlanCircle.AddPoint(const CoordX,CoordY:Integer; CallHandler:TCallHandler):Boolean;
  var op : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  Result := (op.CurDrawingPointNum < 2);
  if not Result then Exit;
  NeedRepaint;
  case op.CurDrawingPointNum of
    0: begin ScrCenter := Point(CoordX,CoordY); ScrRadius := 1; end;
    1: ScrRadius := Distance(ScrCenter,Point(CoordX,CoordY));
  end;
  NeedRepaint;
end;

function TPlanCircle.GetClippingRgn : HRGN;
  var r:TRect;
begin
  r:=GetDisplayRect;
  Inc(r.Left,MarginDelta+FPen.Width);
  Inc(r.Top,MarginDelta+FPen.Width);
  Dec(r.Right,MarginDelta+FPen.Width);
  Dec(r.Bottom,MarginDelta+FPen.Width);
  Result:=CreateEllipticRgnIndirect(r);
end;

procedure TPlanCircle.UpdateObjectProperties;
  var s:String;
begin
  inherited UpdateObjectProperties;
  s:=FProperties.Values['X'];
  if (s<>'') then try Fx:=StrToInt(s) except end;
  s:=FProperties.Values['Y'];
  if (s<>'') then try Fy:=StrToInt(s) except end;
  s:=FProperties.Values['Radius'];
  if (s<>'') then try FRadius:=StrToInt(s) except end;
end;


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  TPlanText
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanText.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  FText := '';
  FSaveText := '';
  FBlinkText := '';
  FFont := TFont.Create;
  FSaveFont := TFont.Create;
  FBlinkFont := TFont.Create;
  FFont.Assign((FOwner as TOxygenPlan).Font);
  FAlignment := taCenter;
  FEscapement := 0;
  FPen.Mode := pmNop;
end;

function TPlanText.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanText.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanText) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanText.SetText(Value:String);
begin
  if (FText=Value) then Exit;
  NeedRepaint;
  FText:=Value;
  NeedRepaint;
end;

procedure TPlanText.SetFont(Value:TFont);
begin
  FFont.Assign(Value);
  NeedRepaint;
end;

procedure TPlanText.SetAlignment(Value:TAlignment);
begin
  if (FAlignment = Value) then Exit;
  FAlignment := Value;
  NeedRepaint;
end;

procedure TPlanText.SetEscapement(const Value : Integer);
begin
  if (Value < 0) or ((Value mod 360) = FEscapement) then Exit;
  FEscapement := Value mod 360;
  NeedRepaint;
end;

procedure TPlanText.Free;
begin
  FVisible:=False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying:=True;
{$IFDEF USE_RX}
  StopBlink;
  FSaveFont.Free;
  FBlinkFont.Free;
{$ENDIF}
  FFont.Free;
  inherited Free;
end;

procedure TPlanText.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2 : TPoint;
      fmt   : Word;
      r     : TRect;
      op : TOxygenPlan;
      rgn,FClippingRgn : HRGN;
      LogFont : TLogFont;
      x,y : Integer;
      angle : Extended;
      FWasClippingRgn : Boolean;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount>0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  r := NormalizeRect(p1.X, p1.Y, p2.X, p2.Y);
  Inc(r.Top, MarginDelta*2);
  Inc(r.Left, MarginDelta*2);
  Dec(r.Bottom, MarginDelta*2);
  Dec(r.Right, MarginDelta*2);
  case FAlignment of
    taRightJustify : fmt := DT_RIGHT + DT_WORDBREAK + DT_NOPREFIX;
    taCenter       : fmt := DT_CENTER + DT_WORDBREAK + DT_NOPREFIX;
    else  fmt := DT_LEFT + DT_WORDBREAK + DT_NOPREFIX;
  end;
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    Font.Assign(Self.FFont);
    with Font do Size := Trunc(Size*(FOwner.Width/op.ScaleX));
    if (FSelectColor <> -1) then Font.Color := FSelectColor;
    if (Font.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Font.Color := clReplaceWhite;
    if (FText = '') then begin
      Pen.Color := op.Pen.Color;
      Pen.Width := 1;
      Pen.Style := psDash;
      Pen.Mode  := pmCopy;
    end;
    if FGradientRecord.Active then begin
      FClippingRgn := GetClippingRgn;
      PlanUtils.GradientFillRect(ACanvas, Rect(p1.X + Pen.Width, p1.Y + Pen.Width, p2.X - Pen.Width, p2.Y - Pen.Width), FGradientRecord.StartColor, FGradientRecord.EndColor, FGradientRecord.Direction, FGradientRecord.Colors);
      DeleteObject(FClippingRgn);
      Brush.Style := bsClear;
    end;
    if (Pen.Mode <> pmNop) then Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    if (FEscapement = 0) then Windows.DrawText(ACanvas.Handle, PChar(FText), -1, r, fmt)
    else begin
      rgn := CreateRectRgn(0,0,0,0);
      FWasClippingRgn := (GetClipRgn(ACanvas.Handle,rgn) = 1); // Old clipping region
      FClippingRgn := GetClippingRgn;
      if FWasClippingRgn then CombineRgn(FClippingRgn,rgn,FClippingRgn,RGN_AND);
      SelectClipRgn(ACanvas.Handle,FClippingRgn);
      DeleteObject(FClippingRgn);

      Windows.GetObject(ACanvas.Font.Handle, SizeOf(TLogFont), @LogFont);
      LogFont.lfEscapement := FEscapement * 10;
      ACanvas.Font.Handle := CreateFontIndirect(LogFont);
      angle := FEscapement;
      angle := (angle * Pi)/180;
      if (FEscapement >= 0) and (FEscapement <= 90) then begin
        x := r.Left;
        y := r.Top + Trunc((r.Bottom - r.Top) * sin(angle));
      end
      else if (FEscapement > 90) and (FEscapement <= 180) then begin
        x := r.Left + Trunc((r.Right - r.Left) * Abs(cos(angle)));
        y := r.Bottom;
      end
      else if (FEscapement > 180) and (FEscapement <= 270) then begin
        x := r.Right;
        y := r.Bottom + Trunc((r.Bottom - r.Top) * sin(angle));
      end
      else if (FEscapement > 270) and (FEscapement < 360) then begin
        x := r.Right - Trunc((r.Right - r.Left) * Abs(cos(angle)));
        y := r.Top;
      end
      else begin
        x := r.Left;
        y := r.Top;
      end;
      TextOut(x, y, FText);
      if FWasClippingRgn then SelectClipRgn(ACanvas.Handle,rgn) else SelectClipRgn(ACanvas.Handle,0);
      DeleteObject(rgn);
    end;
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then SelectRectRgn(ACanvas,op.SelectColor,p1,p2);
  end;
end;

function TPlanText.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'CAPTION') or (s = 'FONTNAME') or (s = 'FONTSIZE') or (s = 'STYLE') or (s = 'ALIGNMENT') or (s = 'FONTCOLOR') or (s = 'ESCAPEMENT');
end;

{$IFDEF USE_RX}
procedure TPlanText.StartBlink(const Interval:Integer);
begin
  StopBlink;
  FSaveFont.Assign(FFont);
  FSaveText:=FText;
  inherited StartBlink(Interval);
end;

procedure TPlanText.StopBlink;
begin
  if (FBlinkTimerHandle=0) then Exit;
  inherited StopBlink;
  FFont.Assign(FSaveFont);
  FText:=FSaveText;
  NeedRepaint;
end;
{$ENDIF}

function TPlanText.CanBeParent : Boolean;
begin
  Result:=False;
end;

function TPlanText.GetObjectTypeDescr : String;
begin
  Result := ocText;
end;

procedure TPlanText.FillObjectProperties;
  var s:String;
begin
  inherited FillObjectProperties;
  case FAlignment of
    taLeftJustify  : s := 'Left';
    taRightJustify : s := 'Right';
    taCenter       : s := 'Center';
  end;
  FProperties.Values['Caption'] := FText;
  FProperties.Values['FontName'] := FFont.Name;
  FProperties.Values['FontCharset'] := IntToStr(FFont.Charset);
  FProperties.Values['FontSize'] := IntToStr(FFont.Size);
  FProperties.Values['FontColor'] := EncodeColor(FFont.Color);
  FProperties.Values['Style'] := GetStyleString(FFont);
  FProperties.Values['Alignment'] := s;
  FProperties.Values['Escapement'] := IntToStr(FEscapement);
end;

{$IFDEF USE_RX}
procedure TPlanText.BlinkTimer(Sender: TObject);
begin
  FBlinkIndex:=(FBlinkIndex xor 1);
  case FBlinkType of
    btNone  : Exit;
    btPen   : if (FBlinkIndex = 1) then FPen.Assign(FBlinkPen) else FPen.Assign(FSavePen);
    btBrush : if (FBlinkIndex = 1) then FBrush.Assign(FBlinkBrush) else FBrush.Assign(FSaveBrush);
    btFont  : if (FBlinkIndex = 1) then FFont.Assign(FBlinkFont) else FFont.Assign(FSaveFont);
    btText  : if (FBlinkIndex = 1) then FText:=FBlinkText else FText:=FSaveText;
  end;
  (FOwner as TOxygenPlan).DoObjectBlink(Self);
  if ((FLayer <> (FOwner as TOxygenPlan).CurrentLayer) or (FBrush.Style = bsClear) or
      (FBlinkType = btPen)) then NeedRepaint else Draw((FOwner as TOxygenPlan).Canvas, FOwner.Width, FOwner.Height);
end;
{$ENDIF}

procedure TPlanText.UpdateObjectProperties;
  var s : String;
begin
  inherited UpdateObjectProperties;
  FText := FProperties.Values['Caption'];
  FFont.Name := FProperties.Values['FontName'];
  s := FProperties.Values['FontCharset'];
  if (s <> '') then try FFont.Charset := StrToInt(s) except end;
  FFont.Style := [];
  s:=FProperties.Values['Style'];
  if (Pos('I',s) > 0) then with FFont do Style := Style + [fsItalic];
  if (Pos('B',s) > 0) then with FFont do Style := Style + [fsBold];
  if (Pos('U',s) > 0) then with FFont do Style := Style + [fsUnderline];
  if (Pos('S',s) > 0) then with FFont do Style := Style + [fsStrikeOut];
  try FFont.Size := StrToInt(FProperties.Values['FontSize']) except end;
  s := FProperties.Values['FontColor'];
  if (s <> '') then try FFont.Color := ReadColor(s) except end;
  s := FProperties.Values['Escapement'];
  if (s <> '') then try FEscapement := StrToInt(s) except FEscapement := 0 end;
  s := FProperties.Values['Alignment'];
  if (s = 'Left') then FAlignment := taLeftJustify
  else if (s = 'Right') then FAlignment := taRightJustify
  else if (s = 'Center') then FAlignment := taCenter;
end;


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanPicture
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanPicture.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  FPicture:=TPicture.Create;
{$IFDEF USE_RX}
  FSavePicture:=TPicture.Create;
  FBlinkPicture:=TPicture.Create;
{$ENDIF}
  FBrush.Color:=clBlack;
  FStdPictureName:='';
end;

function TPlanPicture.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanPicture.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanPicture) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanPicture.Free;
begin
  FVisible:=False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying:=True;
{$IFDEF USE_RX}
  StopBlink;
  FSavePicture.Free;
  FBlinkPicture.Free;
{$ENDIF}
  FPicture.Free;
  inherited Free;
end;

procedure TPlanPicture.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
      op : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FPicture = nil) or (FPicture.Graphic = nil) or FPicture.Graphic.Empty then begin
      Pen.Color := op.Pen.Color;
      Pen.Width := 1;
      Pen.Style := psDash;
      Pen.Mode  := pmCopy;
      Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    end
    else begin
      //if ((FPicture.Graphic is TBitmap) and (FPicture.Bitmap <> nil) and Transparent) then FPicture.Bitmap.TransparentColor := FPicture.Bitmap.Canvas.Pixels[0,0];
      if Assigned(FPicture) then if (FPicture.Graphic <> nil) then StretchDraw(Rect(p1.X, p1.Y, p2.X, p2.Y), FPicture.Graphic);
    end;
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then SelectRectRgn(ACanvas,op.SelectColor,p1,p2);
    if (FSelectColor <> -1) then begin
      ACanvas.Pen.Color := FSelectColor;
      ACanvas.Pen.Width := 2;
      ACanvas.Brush.Color := op.Color;
      ACanvas.Brush.Style := bsClear;
      Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    end;
  end;
  DrawChildren(ACanvas, CWidth, CHeight);
end;

{$IFDEF USE_RX}
procedure TPlanPicture.StartBlink(const Interval:Integer);
begin
  StopBlink;
  FSavePicture.Assign(FPicture);
  inherited StartBlink(Interval);
end;

procedure TPlanPicture.StopBlink;
begin
  if (FBlinkTimerHandle=0) then Exit;
  inherited StopBlink;
  FPicture.Assign(FSavePicture);
  NeedRepaint;
end;
{$ENDIF}

function TPlanPicture.GetObjectTypeDescr : String;
begin
  Result:=ocPicture;
end;

procedure TPlanPicture.SetStdPictureName(const Value : String);
  var i : Integer;
begin
  if (FStdPictureName = Value) then Exit;
  FStdPictureName:='';
  i := (FOwner as TOxygenPlan).PictureIndex(AnsiUpperCase(Value));
  if (i>=0) then begin
    FPicture.Assign((FOwner as TOxygenPlan).Pictures[i]);
    FStdPictureName := AnsiUpperCase(Value);
  end
  else (FOwner as TOxygenPlan).LoadStdPicture(Value,Self);
end;

function TPlanPicture.GetTransparent : Boolean;
begin
  Result := False;
  if (FPicture <> nil) and (FPicture.Graphic <> nil) then Result := FPicture.Graphic.Transparent;
end;

procedure TPlanPicture.SetTransparent(const Value : Boolean);
begin
  if (FPicture <> nil) and (FPicture.Graphic <> nil) then FPicture.Graphic.Transparent := Value;
  if ((FPicture.Graphic is TBitmap) and (FPicture.Bitmap <> nil) and Value) then begin
    FPicture.Bitmap.TransparentColor := FPicture.Bitmap.Canvas.Pixels[0,0];
    FPicture.Bitmap.Transparent := True;
  end;
  FProperties.Values['Transparent'] := IntToStr(Integer(Value));
end;

procedure TPlanPicture.FillObjectProperties;
  var s:String;
      ms:TMemoryStream;
      pic:TPicture;
begin
  inherited FillObjectProperties;
  if (FStdPictureName <> '') then begin
    FProperties.Values['Graphic']:='%'+FStdPictureName+'%';
    FProperties.Values['Transparent']:=IntToStr(Integer(FPicture.Graphic.Transparent));
  end
  else begin
    pic:=TPicture.Create;
    pic.Assign(FPicture);
    if ((pic.Graphic=nil) or pic.Graphic.Empty) then s:=''
    else begin
      ms:=TMemoryStream.Create;
      pic.Graphic.SaveToStream(ms);
      if (poSavePictures in (FOwner as TOxygenPlan).Options) then FProperties.Values['Graphic'] := Stream2String(ms)
      else FProperties.Values['Graphic'] := '';
      ms.Free;
      FProperties.Values['Transparent']:=IntToStr(Integer(pic.Graphic.Transparent));
    end;
    FProperties.Values['Type']:=GetPictureType(pic);
    pic.Free;
  end;
end;

{$IFDEF USE_RX}
procedure TPlanPicture.BlinkTimer(Sender: TObject);
begin
  FBlinkIndex:=(FBlinkIndex xor 1);
  case FBlinkType of
    btNone    : Exit;
    btPen     : if (FBlinkIndex=1) then FPen.Assign(FBlinkPen) else FPen.Assign(FSavePen);
    btBrush   : if (FBlinkIndex=1) then FBrush.Assign(FBlinkBrush) else FBrush.Assign(FSaveBrush);
    btPicture : if (FBlinkIndex=1) then FPicture.Assign(FBlinkPicture) else FPicture.Assign(FSavePicture);
  end;
  (FOwner as TOxygenPlan).DoObjectBlink(Self);
  if (FLayer = (FOwner as TOxygenPlan).CurrentLayer) then begin
    if (FPicture.Graphic<>nil) then if not FPicture.Graphic.Transparent then Draw((FOwner as TOxygenPlan).Canvas, FOwner.Width, FOwner.Height) else NeedRepaint;
  end
  else NeedRepaint;
end;
{$ENDIF}

function TPlanPicture.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'GRAPHIC') or (s = 'TYPE') or (s = 'TRANSPARENT');
end;

procedure TPlanPicture.UpdateObjectProperties;
  var s,gr:String;
      ms:TMemoryStream;
      i:Integer;
begin
  inherited UpdateObjectProperties;

  FPicture.Assign(nil);
  FStdPictureName:='';
  gr := FProperties.Values['Graphic'];
  if (Trim(gr) = '') then (FOwner as TOxygenPlan).LoadPicture(Self)
  else if (Length(gr)>2) then if ((gr[1]='%') and (gr[Length(gr)]='%')) then begin
    gr:=AnsiUpperCase(System.Copy(gr,2,Length(gr)-2));
    i := (FOwner as TOxygenPlan).PictureIndex(gr);
    if (i>=0) then begin
      FPicture.Assign((FOwner as TOxygenPlan).Pictures[i]);
      FStdPictureName := gr;
      try FPicture.Graphic.Transparent := (StrToInt(Properties.Values['Transparent'])<>0) except end;
    end
    else (FOwner as TOxygenPlan).LoadStdPicture(gr,Self);
  end
  else begin
    s := FProperties.Values['Type'];
    ms := TMemoryStream.Create;
    String2Stream(gr,ms);
    LoadPictureFromStream(s, FPicture, ms);
    ms.Free;
    if not ((FPicture.Graphic = nil) or FPicture.Graphic.Empty) then try
      FPicture.Graphic.Transparent:=(StrToInt(Properties.Values['Transparent'])<>0);
    except end;
    if ((FPicture.Graphic is TBitmap) and (FPicture.Bitmap <> nil) and Transparent) then FPicture.Bitmap.TransparentColor := FPicture.Bitmap.Canvas.Pixels[0,0];
  end;
end;

{$IFDEF USE_RX}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TPlanAniPicture
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanAniPicture.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  FImageIndex:=0;
  FImageCount:=1;
  FInterval:=1000;
  FCurBitmap:=TBitmap.Create;
end;

function TPlanAniPicture.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanAniPicture.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanAniPicture) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanAniPicture.Free;
begin
  FVisible:=False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying:=True;
  StopBlink;
  FCurBitmap.Free;
  inherited Free;
end;

procedure TPlanAniPicture.SetImageIndex(const Value : Integer);
begin
  if ((FImageIndex = Value) or (Value > (FImageCount-1)) or (Value < 0)) then Exit;
  FImageIndex:=Value;
  NeedRepaint;
end;

procedure TPlanAniPicture.SetImageCount(const Value : Integer);
begin
  if ((FImageCount = Value) or (not Assigned(FPicture)) or (FPicture.Graphic = nil) or (Value < 1)) then Exit;
  FImageCount:=Value;
  FCurBitmap.Width:=FPicture.Width div FImageCount;
  FCurBitmap.Height:=FPicture.Height;
  FCurBitmap.Transparent:=FPicture.Graphic.Transparent;
end;

function TPlanAniPicture.CanBeParent : Boolean;
begin
  Result:=False;
end;

procedure TPlanAniPicture.FillObjectProperties;
begin
  inherited FillObjectProperties;
  Properties.Values['ImageCount']:=IntToStr(FImageCount);
  Properties.Values['Interval']:=IntToStr(FInterval);
end;

function TPlanAniPicture.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'IMAGECOUNT') or (s = 'INTERVAL');
end;

procedure TPlanAniPicture.BlinkTimer(Sender: TObject);
begin
  if Assigned(FPicture) then if (FPicture.Graphic <> nil) then if (FPicture.Graphic is TBitmap) then begin
    if (FImageIndex < FImageCount - 1) then Inc(FImageIndex) else FImageIndex := 0;
    (FOwner as TOxygenPlan).DoObjectBlink(Self);
    if (FLayer = (FOwner as TOxygenPlan).CurrentLayer) then Draw((FOwner as TOxygenPlan).Canvas, FOwner.Width, FOwner.Height) else NeedRepaint;
  end;
end;

function TPlanAniPicture.GetObjectTypeDescr : String;
begin
  Result:=ocAniPicture;
end;

procedure TPlanAniPicture.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
      op : TOxygenPlan;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FPicture = nil) or (FPicture.Graphic = nil) or FPicture.Graphic.Empty or (not (FPicture.Graphic is TBitmap)) then begin
      Pen.Color := op.Pen.Color;
      Pen.Width := 1;
      Pen.Style := psDash;
      Pen.Mode  := pmCopy;
      Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    end
    else if Assigned(FPicture) then if (FPicture.Graphic<>nil) then if (FPicture.Graphic is TBitmap) then begin
      FCurBitmap.Canvas.CopyRect(Rect(0,0,FCurBitmap.Width,FCurBitmap.Height),FPicture.Bitmap.Canvas,
                                 Rect(FImageIndex*FCurBitmap.Width,0,(FImageIndex+1)*FCurBitmap.Width,FCurBitmap.Height));
      StretchDraw(Rect(p1.X, p1.Y, p2.X, p2.Y), FCurBitmap);
    end;
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then SelectRectRgn(ACanvas,op.SelectColor,p1,p2);
    if (FSelectColor <> -1) then begin
      ACanvas.Pen.Color := FSelectColor;
      ACanvas.Pen.Width := 2;
      ACanvas.Brush.Color := op.Color;
      ACanvas.Brush.Style := bsClear;
      Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    end;
  end;
end;

procedure TPlanAniPicture.Animate(const Interval:Integer);
begin
  StartBlink(Interval);
end;

procedure TPlanAniPicture.StartBlink(const Interval:Integer);
begin
  StopBlink;
  if (FImageCount = 0) then Exit;
  FInterval := Interval;
  FBlinkTimerHandle := (FOwner as TOxygenPlan).AddTimerEvent(BlinkTimer,Interval,True);
end;

procedure TPlanAniPicture.StopBlink;
begin
  if (FBlinkTimerHandle = 0) then Exit;
  FImageIndex := 0;
  (FOwner as TOxygenPlan).DeleteTimerEvent(FBlinkTimerHandle);
  FBlinkType := btNone;
  FBlinkTimerHandle := 0;
  if FVisible then NeedRepaint;
end;

procedure TPlanAniPicture.Stop;
begin
  StopBlink;
end;

procedure TPlanAniPicture.UpdateObjectProperties;
  var s:String;
begin
  inherited UpdateObjectProperties;
  s:=Properties.Values['Interval'];
  if (s='') then FInterval:=100 else try FInterval:=StrToInt(s) except end;

  s:=Properties.Values['ImageCount'];
  if (s='') then SetImageCount(1) else try SetImageCount(StrToInt(s)) except end;
  if ((poAutoAnimate in (FOwner as TOxygenPlan).Options) and (FImageCount>0)) then StartBlink(FInterval);
end;
{$ENDIF}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanPolygon
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanPolygon.Create(Owner:TCustomControl);
begin
  inherited Create(Owner);
  FPointCount:=0;
  FPointArray := nil;
  FScrPointArray := nil;
  FScrSaveArray := nil;
  FDrawArray := nil;
  FCurMovingPointIndex:=-1;
end;

function TPlanPolygon.Clone(Parent : TPlanObject) : TPlanObject;
  var i : Integer;
      p : TPoint;
begin
  Result := TPlanPolygon.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanPolygon) do for i := 1 to PointCount do begin
    p := Points[i-1];
    Points[i-1] := Point(p.X + 4, p.Y + 4);
  end;
end;

procedure TPlanPolygon.SetPointCount(Value:Integer);
  var pa:PPointArray;
      i:Integer;
begin
  if (Value = FPointCount) then Exit;
  GetMem(pa,FPointCount*SizeOf(TPoint));
  System.Move(FPointArray^,pa^,FPointCount*SizeOf(TPoint));
  FreeMem(FPointArray,FPointCount*SizeOf(TPoint));
  if (FScrPointArray<>nil) then FreeMem(FScrPointArray,FPointCount*SizeOf(TPoint));
  if (FScrSaveArray<>nil) then FreeMem(FScrSaveArray,FPointCount*SizeOf(TPoint));
  if (FDrawArray <> nil) then FreeMem(FDrawArray,FPointCount*SizeOf(TPoint));
  GetMem(FPointArray,Value*SizeOf(TPoint));
  for i:=1 to Value do if (FPointCount>=i) then FPointArray^[i-1]:=pa^[i-1] else FPointArray^[i-1]:=Point(0,0);
  FreeMem(pa,FPointCount*SizeOf(TPoint));
  FPointCount:=Value;
  GetMem(FScrPointArray,FPointCount*SizeOf(TPoint));
  GetMem(FScrSaveArray,FPointCount*SizeOf(TPoint));
  GetMem(FDrawArray,FPointCount*SizeOf(TPoint));
  for i:=1 to FPointCount do
      FScrPointArray^[i-1]:=Point(Trunc(FPointArray^[i-1].X*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),
                                  Trunc(FPointArray^[i-1].Y*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)))
end;

function  TPlanPolygon.GetPoint(Index:Integer):TPoint;
begin
  Result:=Point(-1,-1);
  if (Index>(FPointCount-1)) then Exit;
  Result:=FPointArray^[Index];
end;

procedure TPlanPolygon.SetPoint(Index:Integer; Value:TPoint);
begin
  if (Index>(FPointCount-1)) then Exit;
  NeedRepaint;
  FPointArray^[Index]:=Value;
  FScrPointArray^[Index]:=Point(Trunc(FPointArray^[Index].X*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(FPointArray^[Index].Y*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
  NeedRepaint;
end;

procedure TPlanPolygon.ParentResized;
  var i:Integer;
begin
  if (FPointCount=0) then Exit;
  if (FScrPointArray = nil) then GetMem(FScrPointArray,FPointCount*SizeOf(TPoint));
  if (FScrSaveArray = nil) then GetMem(FScrSaveArray,FPointCount*SizeOf(TPoint));
  if (FDrawArray = nil) then GetMem(FDrawArray,FPointCount*SizeOf(TPoint));
  for i:=1 to FPointCount do
      FScrPointArray^[i-1]:=Point(Trunc(FPointArray^[i-1].X*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),
                                  Trunc(FPointArray^[i-1].Y*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)))
end;

procedure TPlanPolygon.DeletePoint(const Index : Integer);
  var pa:PPointArray;
begin
  if ((Index >= FPointCount) or (FPointCount <= 3)) then Exit;
  if (Index<FPointCount-1) then begin
    pa:=Addr(FPointArray^[Index+1]);
    System.Move(pa^,FPointArray^[Index],(FPointCount-1-Index)*SizeOf(TPoint));
  end;
  SetPointCount(FPointCount-1);
  NeedRepaint;
end;

procedure TPlanPolygon.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i  : Integer;
      op : TOxygenPlan;
      rgn : HRGN;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  for i := 1 to FPointCount do FDrawArray^[i-1] := Point(Trunc(FPointArray^[i-1].X*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(FPointArray^[i-1].Y*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    if FGradientRecord.Active then begin
      rgn := GetClippingRgn;
      PlanUtils.GradientFillRgn(ACanvas,DisplayRect, rgn, FGradientRecord.StartColor,FGradientRecord.EndColor,FGradientRecord.Direction, FGradientRecord.Colors);
      DeleteObject(rgn);
      Brush.Style := bsClear;
    end;
    Windows.Polygon(ACanvas.Handle, FDrawArray^, FPointCount);

    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      ACanvas.Pen.Color := op.SelectColor;
      ACanvas.Brush.Color := op.SelectColor;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Pen.Width := 1;
      for i := 1 to FPointCount do Rectangle(FDrawArray^[i-1].X - MarginDelta, FDrawArray^[i-1].Y - MarginDelta, FDrawArray^[i-1].X + MarginDelta, FDrawArray^[i-1].Y + MarginDelta);
    end;
  end;
  DrawChildren(ACanvas, CWidth, CHeight);
end;

procedure TPlanPolygon.Free;
begin
  FVisible:=False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying:=True;
{$IFDEF USE_RX}
  StopBlink;
{$ENDIF}
  if (FPointArray<>nil) then FreeMem(FPointArray,FPointCount*SizeOf(TPoint));
  if (FScrPointArray<>nil) then FreeMem(FScrPointArray,FPointCount*SizeOf(TPoint));
  if (FScrSaveArray<>nil) then FreeMem(FScrSaveArray,FPointCount*SizeOf(TPoint));
  if (FDrawArray <> nil) then FreeMem(FDrawArray,FPointCount*SizeOf(TPoint));
  FPointCount:=0;
  FPointArray:=nil;
  FScrPointArray:=nil;
  FScrSaveArray:=nil;
  FDrawArray := nil;
  inherited Free;
end;

function TPlanPolygon.HasInternalPoint(X,Y:Integer):Boolean;
  var rgn:HRgn;
begin
  rgn:=CreatePolygonRgn(FScrPointArray^,FPointCount,WINDING);
  Result:=PtInRegion(rgn,X,Y);
  Windows.DeleteObject(rgn);
end;

function TPlanPolygon.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'POINTCOUNT');
  if Result then Exit;
  Result := (s[1] = 'X') or (s[1] = 'Y');
  if Result then try
    StrToInt(System.Copy(s,2,255));
  except
    Result := False;
  end;
end;

function TPlanPolygon.HasMarginPoint(X,Y:Integer; var MarginType:TMarginType):Boolean;
  var i:Integer;
      p:TPoint;
begin
  Result:=False;
  for i:=1 to FPointCount do begin
    p:=FScrPointArray^[i-1];
    if PointInRect(Point(X,Y),Rect(p.X-MarginDelta,p.Y-MarginDelta,p.X+MarginDelta,p.Y+MarginDelta)) then begin
      FCurMovingPointIndex:=i-1;
      Result:=True;
      MarginType:=mtMovePoint;
      Exit;
    end;
  end;
end;

procedure TPlanPolygon.EditMove(APoint:TPoint);
  var i,dx,dy : Integer;
      tmp : PPointArray;
      clr : HRGN;
      f   : Boolean;
      op  : TOxygenPlan;
begin
  op := FOwner as TOxygenPlan;
  NeedRepaint;
  dx := APoint.X - op.StartCoord.X;
  dy := APoint.Y - op.StartCoord.Y;

  GetMem(tmp, FPointCount*SizeOf(TPoint));
  for i := 1 to FPointCount do tmp^[i-1] := Point(FScrSaveArray^[i-1].X+dx,FScrSaveArray^[i-1].Y+dy);
  if (FParent <> nil) and (poChildrenClipping in op.Options) then begin
    clr:=FParent.GetClippingRgn;
    f:=False;
    for i:=1 to FPointCount do if PtInRegion(clr,tmp^[i-1].X,tmp^[i-1].Y) then begin
      f:=True;
      Break;
    end;
    if not f then begin
      DeleteObject(clr);
      FreeMem(tmp,FPointCount*SizeOf(TPoint));
      Exit;
    end;
    DeleteObject(clr);
  end;
  System.Move(tmp^,FScrPointArray^,FPointCount*SizeOf(TPoint));
  FreeMem(tmp,FPointCount*SizeOf(TPoint));
  if (FOwner.Width > 0) and (FOwner.Height > 0) then for i := 1 to FPointCount do
    FPointArray^[i-1]:=Point(Trunc(FScrPointArray^[i-1].X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width)),
                             Trunc(FScrPointArray^[i-1].Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height)));
  NeedRepaint;

  inherited EditMove(APoint);
end;

procedure TPlanPolygon.EditResize(APoint:TPoint);
begin
  if not FCanResize then Exit;
  if (FCurMovingPointIndex=-1) then Exit;
  NeedRepaint;
  FScrPointArray[FCurMovingPointIndex] := APoint;
  if (FOwner.Width > 0) and (FOwner.Height > 0) then FPointArray^[FCurMovingPointIndex] := Point(Trunc(APoint.X*((FOwner as TOxygenPlan).ScaleX/FOwner.Width)),Trunc(APoint.Y*((FOwner as TOxygenPlan).ScaleY/FOwner.Height)));
  NeedRepaint;
end;

procedure TPlanPolygon.Move(const DeltaX, DeltaY : Integer);
  var i : Integer;
begin
  NeedRepaint;
  for i := 1 to FPointCount do begin
    Inc(FPointArray^[i-1].X, DeltaX);
    Inc(FPointArray^[i-1].Y, DeltaY);
    FScrPointArray^[i-1] := Point(Trunc(FPointArray^[i-1].X*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),
                                  Trunc(FPointArray^[i-1].Y*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)))
  end;
  NeedRepaint;
  inherited Move(DeltaX, DeltaY);
end;

procedure TPlanPolygon.Resize(const DeltaX, DeltaY : Integer);
begin
  // do nothing
end;

function TPlanPolygon.CanBeParent : Boolean;
begin
  Result:=True;
end;

function TPlanPolygon.GetObjectTypeDescr : String;
begin
  Result := ocPolygon;
end;

procedure TPlanPolygon.FillObjectProperties;
  var i:Integer;
begin
  inherited FillObjectProperties;
  FProperties.Values['PointCount']:=IntToStr(FPointCount);
  for i:=1 to FPointCount do begin
    FProperties.Values['X'+IntToStr(i)]:=IntToStr(FPointArray^[i-1].X);
    FProperties.Values['Y'+IntToStr(i)]:=IntToStr(FPointArray^[i-1].Y);
  end;
end;

procedure TPlanPolygon.SaveCoords;
begin
  if (FPointCount=0) then Exit;
  System.Move(FScrPointArray^,FScrSaveArray^,FPointCount*SizeOf(TPoint));
  inherited SaveCoords;
end;

procedure TPlanPolygon.NeedRepaint;
  var r:TRect;
begin
  if FDestroying then Exit;
  if (FPointCount=0) then Exit;
  r:=DisplayRect;
  Dec(r.Top,MarginDelta + Pen.Width);
  Dec(r.Left,MarginDelta + Pen.Width);
  Inc(r.Bottom,MarginDelta + Pen.Width);
  Inc(r.Right,MarginDelta + Pen.Width);
  Windows.InvalidateRect(FOwner.Handle,@r,False);
end;

function TPlanPolygon.GetDisplayRect:TRect;
  var i:Integer;
begin
  Result:=Rect(FScrPointArray^[0].X,FScrPointArray^[0].Y,FScrPointArray^[0].X,FScrPointArray^[0].Y);
  for i:=1 to FPointCount do begin
    if (FScrPointArray^[i-1].X < Result.Left) then Result.Left:=FScrPointArray^[i-1].X;
    if (FScrPointArray^[i-1].X > Result.Right) then Result.Right:=FScrPointArray^[i-1].X;
    if (FScrPointArray^[i-1].Y < Result.Top) then Result.Top:=FScrPointArray^[i-1].Y;
    if (FScrPointArray^[i-1].Y > Result.Bottom) then Result.Bottom:=FScrPointArray^[i-1].Y;
  end;
  Dec(Result.Top, Pen.Width div 2);
  Dec(Result.Left, Pen.Width div 2);
  Inc(Result.Bottom, Pen.Width div 2);
  Inc(Result.Right, Pen.Width div 2);
end;

function TPlanPolygon.AddPoint(const CoordX,CoordY:Integer; CallHandler:TCallHandler):Boolean;
  var i  : Integer;
      op : TOxygenPlan;
begin
  if ((FPointCount > 3) and (CallHandler = chMouseDown)) then Result:=(Distance(FScrPointArray^[0],Point(CoordX,CoordY)) > Cardinal(MarginDelta*2)) else Result:=True;
  if not Result then begin
    DeletePoint(FPointCount-1);
    Exit;
  end;
  op := FOwner as TOxygenPlan;
  i := op.CurDrawingPointNum;
  NeedRepaint;
  PointCount := i + 1;
  FScrPointArray^[i] := Point(CoordX,CoordY);
  if (FOwner.Width > 0) and (FOwner.Height > 0) then
    Points[i] := Point(Trunc(FScrPointArray^[i].X*(op.ScaleX/FOwner.Width)),Trunc(FScrPointArray^[i].Y*(op.ScaleY/FOwner.Height)));
end;

function TPlanPolygon.GetClippingRgn : HRGN;
begin
  Result := CreatePolygonRgn(FScrPointArray^,FPointCount,WINDING);
end;

procedure TPlanPolygon.UpdateObjectProperties;
  var i:Integer;
      s:String;
begin
  inherited UpdateObjectProperties;
  PointCount:=0;
  s:=FProperties.Values['PointCount'];
  if (s<>'') then try PointCount:=StrToInt(s) except end;
  for i:=1 to FPointCount do try
    FPointArray^[i-1]:=Point(StrToInt(FProperties.Values['X'+IntToStr(i)]),StrToInt(Properties.Values['Y'+IntToStr(i)]));
    FScrPointArray^[i-1]:=Point(Trunc(FPointArray^[i-1].X*(FOwner.Width/(FOwner as TOxygenPlan).ScaleX)),Trunc(FPointArray^[i-1].Y*(FOwner.Height/(FOwner as TOxygenPlan).ScaleY)));
  except end;
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanPolyLine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TPlanPolyLine.GetObjectTypeDescr : String;
begin
  Result := ocPolyLine;
end;

function TPlanPolyLine.GetClippingRgn : HRGN;
begin
  Result := CreateRectRgn(0,0,0,0);
end;

function TPlanPolyLine.CanBeParent : Boolean;
begin
  Result := False;
end;

function TPlanPolyLine.Clone(Parent : TPlanObject) : TPlanObject;
  var i : Integer;
      p : TPoint;
begin
  Result := TPlanPolyLine.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanPolyLine) do for i := 1 to PointCount do begin
    p := Points[i-1];
    Points[i-1] := Point(p.X + 4, p.Y + 4);
  end;
end;

procedure TPlanPolyLine.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i  : Integer;
      op : TOxygenPlan;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  for i := 1 to FPointCount do FDrawArray^[i-1] := Point(Trunc(FPointArray^[i-1].X*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(FPointArray^[i-1].Y*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    Windows.PolyLine(ACanvas.Handle, FDrawArray^, FPointCount);
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      ACanvas.Pen.Color := op.SelectColor;
      ACanvas.Brush.Color := op.SelectColor;
      ACanvas.Brush.Style := bsSolid;
      ACanvas.Pen.Width := 1;
      for i:=1 to FPointCount do Rectangle(FDrawArray^[i-1].X - MarginDelta, FDrawArray^[i-1].Y - MarginDelta, FDrawArray^[i-1].X + MarginDelta, FDrawArray^[i-1].Y + MarginDelta);
    end;
  end;
end;

function TPlanPolyLine.HasInternalPoint(X,Y:Integer):Boolean;
  var i : Integer;
begin
  Result := False;
  for i := 1 to PointCount - 1 do begin
    Result := PointOnLine(Point(X,Y), FScrPointArray^[i-1], FScrPointArray^[i], Pen.Width);
    if Result then Exit;
  end;
end;

function TPlanPolyLine.AddPoint(const CoordX,CoordY : Integer; CallHandler : TCallHandler) : Boolean;
  var i  : Integer;
      op : TOxygenPlan;
begin
  if ((FPointCount > 3) and (CallHandler = chMouseDown)) then Result := (Distance(FScrPointArray^[FPointCount - 2],Point(CoordX,CoordY)) > Cardinal(MarginDelta*2)) else Result:=True;
  if not Result then Exit;
  op := FOwner as TOxygenPlan;
  i := op.CurDrawingPointNum;
  NeedRepaint;
  PointCount := i + 1;
  FScrPointArray^[i] := Point(CoordX,CoordY);
  if (FOwner.Width > 0) and (FOwner.Height > 0) then
    Points[i] := Point(Trunc(FScrPointArray^[i].X*(op.ScaleX/FOwner.Width)),Trunc(FScrPointArray^[i].Y*(op.ScaleY/FOwner.Height)));
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanTraceLine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
function TPlanTraceLine.GetObjectTypeDescr : String;
begin
    Result := ocTraceLine;
end;

function TPlanTraceLine.Clone(Parent : TPlanObject) : TPlanObject;
  var i : Integer;
      p : TPoint;
begin
  Result := TPlanTraceLine.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanTraceLine) do for i := 1 to PointCount do begin
    p := Points[i-1];
    Points[i-1] := Point(p.X + 4, p.Y + 4);
  end;
end;

procedure TPlanTraceLine.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var i  : Integer;
      op : TOxygenPlan;
      p,np : TPoint;
      tgbeta,sinbeta : Extended;
      a,b,h,dx,dy : Integer;
      al, aw : Integer;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) or (FPointCount = 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  for i := 1 to FPointCount do FDrawArray^[i-1] := Point(Trunc(FPointArray^[i-1].X*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(FPointArray^[i-1].Y*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then ACanvas.Pen.Color := FSelectColor;
    if (Pen.Color = clWhite) and (FOwner as TOxygenPlan).Printing then Pen.Color := clReplaceWhite;
    Windows.PolyLine(ACanvas.Handle, FDrawArray^, FPointCount);
    Brush.Style := bsSolid;
    Brush.Color := Pen.Color;
    p := FScrPointArray^[0];
    Ellipse(p.X - MarginDelta - Pen.Width, p.Y - MarginDelta - Pen.Width, p.X + MarginDelta + Pen.Width, p.Y + MarginDelta + Pen.Width);
    {if (FPointCount > 3) then} for i := 1 to FPointCount-1 do begin // draw arrows
      p := FDrawArray^[i-1];
      p.Y := FOwner.Height - p.Y;
      np := FDrawArray^[i];
      np.Y := FOwner.Height - np.Y;
      h := Trunc(Sqrt(Sqr(np.X - p.X) + Sqr(np.Y - p.Y)));
      al := Trunc(iArrowLength * (CWidth/(FOwner as TOxygenPlan).ScaleX));
      aw := Trunc(iArrowWidth * (CWidth/(FOwner as TOxygenPlan).ScaleX));
      if (h > al) then begin
        if ((np.X - p.X) <> 0) then tgbeta := (np.Y - p.Y)/(np.X - p.X) else tgbeta := 4e4900;
        sinbeta := (np.Y - p.Y)/h;
        b := Trunc(np.Y - iArrowLength*sinbeta);
        if (tgbeta <> 0) then a := np.X - Trunc((al*sinbeta)/tgbeta) else a := np.X - 10;
        b := FOwner.Height - b;
        dy := Trunc(((np.X-p.X)*aw)/h);
        dx := Trunc(((np.Y-p.Y)*aw)/h);
        np.Y := FOwner.Height - np.Y;
        Polygon([np,Point(a + dx, b + dy),Point(a - dx, b - dy){,np}]);
      end;
    end;
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then begin
      Pen.Color := op.SelectColor;
      Brush.Color := op.SelectColor;
      Brush.Style := bsSolid;
      Pen.Width := 1;
      for i := 1 to FPointCount do Rectangle(FDrawArray^[i-1].X - MarginDelta, FDrawArray^[i-1].Y - MarginDelta, FDrawArray^[i-1].X + MarginDelta, FDrawArray^[i-1].Y + MarginDelta);
    end;
  end;
end;

function TPlanTraceLine.GetDisplayRect : TRect;
begin
  Result := inherited GetDisplayRect;
  Dec(Result.Left, MarginDelta);
  Dec(Result.Top, MarginDelta);
  Inc(Result.Right, MarginDelta);
  Inc(Result.Bottom, MarginDelta);
end;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//   TPlanCompositeObject
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
procedure TPlanCompositeObject.SetPen(Value:TPen);
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].SetPen(Value);
  inherited SetPen(Value);
end;

procedure TPlanCompositeObject.SetBrush(Value:TBrush);
  var i:Integer;
begin
  for i := 1 to ChildrenCount do Objects[i-1].SetBrush(Value);
  inherited SetBrush(Value);
end;

procedure TPlanCompositeObject.SetSelected(Value:Boolean);
  var i:Integer;
begin
  for i := 1 to ChildrenCount do Objects[i-1].SetSelected(Value);
  inherited SetSelected(Value);
end;

procedure TPlanCompositeObject.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
begin
  if ((FOwner as TOxygenPlan).DisableRepaintCount>0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and (FOwner as TOxygenPlan).ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  ACanvas.Pen.Assign(FPen);
  ACanvas.Pen.Style:=psDash;
  ACanvas.Brush.Style:=bsClear;
  ACanvas.Rectangle(p1.X,p1.Y,p2.X,p2.Y);
  DrawChildren(ACanvas, CWidth, CHeight);
end;

procedure TPlanCompositeObject.Select;
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].Select;
  FSelected:=True;
  NeedRepaint;
end;

procedure TPlanCompositeObject.UnSelect;
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].UnSelect;
  inherited Unselect;
end;

procedure TPlanCompositeObject.SelectByColor(const AColor:TColor);
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].SelectByColor(AColor);
end;

procedure TPlanCompositeObject.UnSelectColor;
  var i:Integer;
begin
  for i:=1 to ChildrenCount do Objects[i-1].UnSelectColor;
end;

function TPlanCompositeObject.CanBeParent : Boolean;
begin
  Result:=True;
end;

function TPlanCompositeObject.GetObjectTypeDescr : String;
begin
  Result:=ocCompositeObject;
end;


{$IFDEF USE_RX}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TPlanAniGif
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
constructor TPlanAniGif.Create(Owner : TCustomControl);
begin
  inherited Create(Owner);
  FImage:=TGIFImage.Create;
  FStdAniGifName:='';
end;

function TPlanAniGif.Clone(Parent : TPlanObject) : TPlanObject;
begin
  Result := TPlanAniGif.Create(Owner);
  FillObjectProperties;
  Result.Parent := Parent;
  Result.Properties.AddStrings(FProperties);
  Result.Properties.Values['Name'] := '';
  Result.Properties.Values['ID'] := '0';
  Result.UpdateObjectProperties;
  with (Result as TPlanAniGif) do begin
    X1 := X1 + 4;
    Y1 := Y1 + 4;
    X2 := X2 + 4;
    Y2 := Y2 + 4;
  end;
end;

procedure TPlanAniGif.Free;
begin
  FVisible:=False;
  if not (csDestroying in FOwner.ComponentState) then try NeedRepaint except end;
  FDestroying:=True;
  StopBlink;
  FImage.Free;
  inherited Free;
end;

function TPlanAniGif.GetDelayTime(Index: Integer): Cardinal;
  var FFrameIndex:Integer;
begin
  FFrameIndex:=FImage.FrameIndex;
  if ((FFrameIndex >= 0) and (FFrameIndex < FImage.Count) and (FImage.Count > 1)) then begin
    Result := FImage.Frames[FFrameIndex].AnimateInterval;
    if (Result < MinDelayTime) then Result := MinDelayTime else if (Result > MaxDelayTime) then Result := MaxDelayTime;
  end
  else Result := 0;
end;

procedure TPlanAniGif.SetStdAniGifName(const Value : String);
  var b:Boolean;
      i:Integer;
begin
  if (FStdAniGifName=Value) then Exit;
  b:=Blinking;
  StopBlink;
  FStdAniGifName:='';
  i := (FOwner as TOxygenPlan).AniGifIndex(AnsiUpperCase(Value));
  if (i>=0) then begin
    FImage.Assign((FOwner as TOxygenPlan).AniGifs[i]);
    FStdAniGifName:=Value;
  end
  else (FOwner as TOxygenPlan).LoadStdAniGif(Value,Self);
  if b then StartBlink(0);
end;

procedure TPlanAniGif.SetGIFImage(Value : TGIFImage);
  var b:Boolean;
begin
  b:=Blinking;
  StopBlink;
  FStdAniGifName:='';
  FImage.Assign(Value);
  if b then StartBlink(0);
end;

function TPlanAniGif.GetFrameIndex : Integer;
begin
  Result := GIFImage.FrameIndex;
end;

procedure TPlanAniGif.SetFrameIndex(const Value : Integer);
begin
  if (Blinking or (Value > (FrameCount - 1))) then Exit;
  GIFImage.FrameIndex := Value;
  NeedRepaint;
end;

function TPlanAniGif.GetFrameCount : Integer;
begin
  Result := GIFImage.Count;
end;

procedure TPlanAniGif.BlinkTimer(Sender: TObject);
  var FFrameIndex:Integer;
      FTimer : TRxTimerEvent;
begin
  if FImage.Empty then begin
    StopBlink;
    Exit;
  end;
  FFrameIndex := FImage.FrameIndex;
  if (FFrameIndex < FImage.Count-1) then Inc(FFrameIndex) else FFrameIndex:=0;
  if (FImage.FrameIndex <> FFrameIndex) then begin
    FImage.FrameIndex := FFrameIndex;
    (FOwner as TOxygenPlan).DoObjectBlink(Self);
    if ((FLayer = (FOwner as TOxygenPlan).CurrentLayer) and (FImage.Frames[0].TransparentColor = clNone)) then Draw((FOwner as TOxygenPlan).Canvas, FOwner.Width, FOwner.Height)
    else NeedRepaint;
  end;
  if (FFrameIndex >= 0) and (FImage.Count > 0) then begin
    FTimer := (FOwner as TOxygenPlan).TimerByHandle(FBlinkTimerHandle);
    FTimer.Interval := GetDelayTime(FFrameIndex);
    FTimer.Enabled:=True;
  end;
end;

function TPlanAniGif.CanBeParent : Boolean;
begin
  Result:=False;
end;

function TPlanAniGif.GetFrameBitmap(Index: Integer; var TransColor: TColor): TBitmap;
  var I, Last, First: Integer;
      SavePal: HPalette;
begin
  Index := Min(Index, FImage.Count - 1);
  Result := TBitmap.Create;
  Result.Canvas.Lock;
  try
    with Result do begin
      Width := FImage.ScreenWidth;
      Height := FImage.ScreenHeight;
      Last := Index;
      First := Max(0, Last);
      SavePal := 0;
      if FImage.Palette <> 0 then begin
        SavePal := SelectPalette(Canvas.Handle, FImage.Palette, False);
        RealizePalette(Canvas.Handle);
      end;
      if (FImage.Frames[FImage.FrameIndex].TransparentColor <> clNone) then begin
        TransColor := GetNearestColor(Canvas.Handle,ColorToRGB(FImage.Frames[FImage.FrameIndex].TransparentColor));
        Canvas.Brush.Color := PaletteColor(TransColor);
      end
      else if ((FImage.BackgroundColor<>clNone) and FImage.Transparent) then Canvas.Brush.Color:=PaletteColor(FImage.BackgroundColor) else Canvas.Brush.Color:=PaletteColor(clWindow);
      Canvas.FillRect(Bounds(0, 0, Width, Height));
      while (First>0) do begin
        if (FImage.ScreenWidth=FImage.Frames[First].Width) and (FImage.ScreenHeight=FImage.Frames[First].Height) then begin
          if (FImage.Frames[First].TransparentColor = clNone) or ((FImage.Frames[First].DisposalMethod = dmRestoreBackground) and (First < Last)) then Break;
        end;
        Dec(First);
      end;
      for I := First to Last - 1 do begin
        with FImage.Frames[I] do case DisposalMethod of
          dmUndefined, dmLeave: Draw(Canvas, Bounds(Origin.X, Origin.Y, Width, Height), True);
          dmRestoreBackground: if (I>First) then Canvas.FillRect(Bounds(Origin.X, Origin.Y, Width, Height));
          dmRestorePrevious: begin { do nothing } end;
        end;
      end;
      with FImage.Frames[Last] do Draw(Canvas,Bounds(Origin.X,Origin.Y,Width,Height), True);
      if (TransColor <> clNone) and FImage.Transparent then begin
        TransparentColor:=PaletteColor(TransColor);
        Transparent:=True;
      end;
      if (FImage.Palette<>0) then SelectPalette(Canvas.Handle, SavePal, False);
    end;
    Result.Canvas.Unlock;
  except
    Result.Canvas.Unlock;
    Result.Free;
    raise;
  end;
end;

function TPlanAniGif.GetObjectTypeDescr : String;
begin
  Result:=ocAniGif;
end;

procedure TPlanAniGif.FillObjectProperties;
  var ms:TMemoryStream;
      gif:TGifImage;
      s:String;
begin
  inherited FillObjectProperties;
  if (FStdAniGifName<>'') then Properties.Values['Graphic']:='%'+FStdAniGifName+'%'
  else begin
    gif:=TGifImage.Create;
    gif.Assign(FImage);
    if ((gif.Count=0) or gif.Empty) then s:=''
    else begin
      ms:=TMemoryStream.Create;
      gif.SaveToStream(ms);
      if (poSavePictures in (FOwner as TOxygenPlan).Options) then FProperties.Values['Graphic'] := Stream2String(ms)
      else FProperties.Values['Graphic'] := '';
      ms.Free;
    end;
    gif.Free;
  end;
end;

function TPlanAniGif.HasInternalProperty(const PropertyName : String) : Boolean;
  var s : String;
begin
  Result := inherited HasInternalProperty(PropertyName);
  if Result then Exit;
  s := UpperCase(PropertyName);
  Result := (s = 'GRAPHIC');
end;


procedure TPlanAniGif.UpdateObjectProperties;
  var s:String;
      ms:TMemoryStream;
      i:Integer;
begin
  inherited UpdateObjectProperties;

  FImage.Clear;
  FStdAniGifName:='';
  s := Properties.Values['Graphic'];
  if (Trim(s) = '') then (FOwner as TOxygenPlan).LoadAniGif(Self)
  else if (Length(s)>2) then if ((s[1]='%') and (s[Length(s)]='%')) then begin
    s := AnsiUpperCase(System.Copy(s,2,Length(s)-2));
    i := (FOwner as TOxygenPlan).AniGifIndex(s);
    if (i>=0) then begin
      FImage.Assign((FOwner as TOxygenPlan).AniGifs[i]);
      FStdAniGifName := s;
    end
    else (FOwner as TOxygenPlan).LoadStdAniGif(s,Self);
  end
  else begin
    ms:=TMemoryStream.Create;
    String2Stream(s,ms);
    try FImage.LoadFromStream(ms) except end;
    ms.Free;
  end;
  if ((poAutoAnimate in (FOwner as TOxygenPlan).Options) and (FImage.Count > 0)) then StartBlink(0);
end;

procedure TPlanAniGif.Draw(ACanvas : TCanvas; const CWidth, CHeight : Integer);
  var p1,p2:TPoint;
      Frame: TBitmap;
      TransColor: TColor;
      op : TOxygenPlan;
begin
  op := FOwner  as TOxygenPlan;
  if (op.DisableRepaintCount > 0) then Exit;
  if not FLayer.Visible then Exit;
  if ((not FVisible) and op.ReadOnly) then Exit;
  p1 := Point(Trunc(Fx1*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy1*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  p2 := Point(Trunc(Fx2*(CWidth/(FOwner as TOxygenPlan).ScaleX)),Trunc(Fy2*(CHeight/(FOwner as TOxygenPlan).ScaleY)));
  if FImage.Empty or (FImage.Count = 0) then with ACanvas do begin
    Pen.Color := op.Pen.Color;
    Pen.Width := 1;
    Pen.Style := psDash;
    Pen.Mode  := pmCopy;
    Brush.Assign(Self.FBrush);
    Rectangle(p1.X, p1.Y, p2.X, p2.Y);
  end
  else if ((FImage.Count>0) and (not FImage.Empty) and (FImage.ScreenWidth > 0) and (FImage.ScreenHeight > 0)) then begin
    TransColor := clNone;
    Frame := GetFrameBitmap(FImage.FrameIndex, TransColor);
    Frame.Canvas.Lock;
    try
      if (TransColor = clNone) or (not FImage.Transparent) then ACanvas.StretchDraw(Rect(p1.X,p1.Y,p2.X,p2.Y), Frame)
      else StretchBitmapRectTransparent(ACanvas, p1.X, p1.Y, p2.X-p1.X, p2.Y-p1.Y,Bounds(0,0,Frame.Width,Frame.Height),Frame,TransColor);
    finally
      Frame.Canvas.Unlock;
    end;
    Frame.Free;
  end;

  with ACanvas do begin
    Pen.Assign(Self.FPen);
    Brush.Assign(Self.FBrush);
    if (FSelectColor <> -1) then begin
      Pen.Color := FSelectColor;
      Pen.Width := 2;
      Brush.Color := op.Color;
      Brush.Style := bsClear;
      Rectangle(p1.X, p1.Y, p2.X, p2.Y);
    end;
    if (FSelected and (not (op.ReadOnly or FLayer.ReadOnly or FReadOnly))) then SelectRectRgn(ACanvas,op.SelectColor,p1,p2);
  end;
end;

procedure TPlanAniGif.Animate;
begin
  StartBlink(0);
end;

procedure TPlanAniGif.StartBlink(const Interval:Integer);
  var intv:Integer;
begin
  StopBlink;
  if ((FImage.Count = 0) or FImage.Empty) then Exit;
  intv := GetDelayTime(0);
  FBlinkTimerHandle := (FOwner as TOxygenPlan).AddTimerEvent(BlinkTimer,intv,True);
end;

procedure TPlanAniGif.StopBlink;
begin
  if (FBlinkTimerHandle = 0) then Exit;
  FImage.FrameIndex := 0;
  (FOwner as TOxygenPlan).DeleteTimerEvent(FBlinkTimerHandle);
  FBlinkType := btNone;
  FBlinkTimerHandle := 0;
  if FVisible then NeedRepaint;
end;

procedure TPlanAniGif.Stop;
begin
  StopBlink;
end;

{$ENDIF}

end.
