{===============================================================}
{    Oxygen Plan Library, v1.3  1999-2001 Oxygen Software      }
{---------------------------------------------------------------}
{        Written by Oleg Fyodorov, delphi@oxygensoftware.com    }
{                http://www.oxygensoftware.com                  }
{===============================================================}

unit PlanUtils;

{$OPTIMIZATION OFF}
{$Include O2Plan.inc}

  interface

    uses Windows, Classes, SysUtils, Graphics, {$IFDEF USE_RX} RxGif, {$ENDIF} jpeg, Controls, Messages;

    const
      MarginDelta    : Integer = 2;
      iArrowLength   : Integer = 15;      // TPlanTraceLine arrow length
      iArrowWidth    : Integer = 5;       // TPlanTraceLine 1/2 arrow width
      clReplaceWhite : TColor = clBlack;  // Replacement color for clWhite when Printing = True

    resourcestring  
      // Plan colors
      pcBlack   = 'Black';
      pcMaroon  = 'Maroon';
      pcGreen   = 'Green';
      pcOlive   = 'Olive';
      pcNavy    = 'Navy';
      pcPurple  = 'Purple';
      pcTeal    = 'Teal';
      pcGray    = 'Gray';
      pcSilver  = 'Silver';
      pcRed     = 'Red';
      pcLime    = 'Lime';
      pcYellow  = 'Yellow';
      pcBlue    = 'Blue';
      pcFuchsia = 'Fuchsia';
      pcAqua    = 'Aqua';
      pcWhite   = 'White';

    type
      TGradientFillDirection = (fdTopToBottom, fdBottomToTop, fdLeftToRight, fdRightToLeft);

    function Distance(const p1,p2 : TPoint) : Cardinal;
    function NormalizeRect(const x1,y1,x2,y2 : Integer) : TRect;
    function ScreenToDecart(const p : TPoint; const ScrHeight : Integer) : TPoint;
    function DecartToScreen(const p : TPoint; const ScrHeight : Integer) : TPoint;
    function PointOnLine(const p : TPoint; const m1,m2 : TPoint; const PenWidth : Integer) : Boolean;
    function PointInRect(const p : TPoint; const R : TRect) : Boolean;
    function PointInPolyRgn(const p : TPoint; const Points: array of TPoint) : Boolean;
    function Hex2Dec(const S : String) : Longint;
    function MakeStr(C : Char; N : Integer) : String;
    function Min(A, B: Longint): Longint;
    function Max(A, B: Longint): Longint;
    function ReadColor(s : String) : TColor;
    function EncodeColor(c : TColor) : String;
    function GetStyleString(AFont : TFont) : String;
    function Stream2String(ms : TMemoryStream) : String;
    procedure String2Stream(const s : String; var ms : TMemoryStream);
    function GetPictureType(Picture : TPicture) : String;
    function LoadPictureFromStream(const PictureType : String; Picture : TPicture; Stream : TStream) : Boolean;
    function GetPictureStreamString(Picture : TPicture) : String;
    procedure CopyParentImage(Control: TControl; Dest: TCanvas);
    procedure GradientFillRect(Canvas: TCanvas; ARect: TRect; StartColor, EndColor: TColor; Direction: TGradientFillDirection; Colors: Byte);
    procedure GradientFillRgn(Canvas: TCanvas; ARect : TRect; ARgn: HRGN; StartColor, EndColor: TColor; Direction: TGradientFillDirection; Colors: Byte);


  var iScreenWidth, iScreenHeight : Integer;

implementation

function Distance(const p1,p2 : TPoint) : Cardinal;
begin
  Result:=Trunc(Sqrt(Sqr(p1.X - p2.X) + Sqr(p1.Y - p2.Y)));
end;

function NormalizeRect(const x1,y1,x2,y2 : Integer) : TRect;
begin
  Result := Rect(x1,y1,x2,y2);
  if (x1 > x2) then if (y1 > y2) then Result := Rect(x2,y2,x1,y1) else Result:=Rect(x2,y1,x1,y2)
               else if (y1 > y2) then Result := Rect(x1,y2,x2,y1);
  Result := Rect(Result.Left - MarginDelta, Result.Top - MarginDelta, Result.Right + MarginDelta, Result.Bottom + MarginDelta);
end;

function ScreenToDecart(const p:TPoint; const ScrHeight:Integer):TPoint;
begin
  Result := Point(p.X,ScrHeight - p.Y);
end;

function DecartToScreen(const p:TPoint; const ScrHeight:Integer):TPoint;
begin
  Result := Point(p.X,ScrHeight-p.Y);
end;

function PointOnLine(const p : TPoint; const m1,m2 : TPoint; const PenWidth : Integer) : Boolean;
  var l2,d12,d22 : {$IFNDEF VER100} Int64 {$ELSE} Integer {$ENDIF};
      dist2 : Extended;
begin
  l2 := (m1.X - m2.X)*(m1.X - m2.X) + (m1.Y - m2.Y)*(m1.Y - m2.Y);
  if (l2 = 0) then begin
    Result := ((Abs(p.X - m1.X) <= (MarginDelta + PenWidth)) and (Abs(p.Y - m1.Y) <= (MarginDelta + PenWidth)));
    Exit;
  end;
  d12 := (m1.X - p.X)*(m1.X - p.X) + (m1.Y - p.Y)*(m1.Y - p.Y);
  d22 := (m2.X - p.X)*(m2.X - p.X) + (m2.Y - p.Y)*(m2.Y - p.Y);
  dist2 := Abs(4*d12*d22 - (l2 - d12 - d22)*(l2 - d12 - d22)) / (4*l2);
  Result := (dist2 <= (MarginDelta + PenWidth)*(MarginDelta + PenWidth)) and PointInRect(p,NormalizeRect(m1.X,m1.Y,m2.X,m2.Y));
end;

function PointInRect(const P: TPoint; const R: TRect): Boolean;
begin
  with R do
    Result := (Left <= P.X) and (Top <= P.Y) and
      (Right >= P.X) and (Bottom >= P.Y);
end;

function PointInPolyRgn(const P: TPoint; const Points: array of TPoint): Boolean;
type
  PPoints = ^TPoints;
  TPoints = array[0..0] of TPoint;
var
  Rgn: HRgn;
begin
  Rgn := CreatePolygonRgn(PPoints(@Points)^, High(Points) + 1, WINDING);
  try
    Result := PtInRegion(Rgn, P.X, P.Y);
  finally
    DeleteObject(Rgn);
  end;
end;

function Hex2Dec(const S: string): Longint;
var
  HexStr: string;
begin
  if Pos('$', S) = 0 then HexStr := '$' + S
  else HexStr := S;
  Result := StrToIntDef(HexStr, 0);
end;

function MakeStr(C: Char; N: Integer): string;
begin
  if N < 1 then Result := ''
  else begin
    SetLength(Result, N);
    FillChar(Result[1], Length(Result), C);
  end;
end;

function Min(A, B: Longint): Longint;
begin
  if A < B then Result := A else Result := B;
end;

function Max(A, B: Longint): Longint;
begin
  if A > B then Result := A else Result := B;
end;

function ReadColor(s : String):TColor;
begin
  Result := clWhite;
  if (UpperCase(s)=UpperCase(pcBlack)) then   Result:=clBlack
  else if (UpperCase(s)=UpperCase(pcMaroon)) then  Result:=clMaroon
  else if (UpperCase(s)=UpperCase(pcGreen)) then   Result:=clGreen
  else if (UpperCase(s)=UpperCase(pcOlive)) then   Result:=clOlive
  else if (UpperCase(s)=UpperCase(pcNavy)) then    Result:=clNavy
  else if (UpperCase(s)=UpperCase(pcPurple)) then  Result:=clPurple
  else if (UpperCase(s)=UpperCase(pcTeal)) then    Result:=clTeal
  else if (UpperCase(s)=UpperCase(pcGray)) then    Result:=clGray
  else if (UpperCase(s)=UpperCase(pcSilver)) then  Result:=clSilver
  else if (UpperCase(s)=UpperCase(pcRed)) then     Result:=clRed
  else if (UpperCase(s)=UpperCase(pcLime)) then    Result:=clLime
  else if (UpperCase(s)=UpperCase(pcYellow)) then  Result:=clYellow
  else if (UpperCase(s)=UpperCase(pcBlue)) then    Result:=clBlue
  else if (UpperCase(s)=UpperCase(pcFuchsia)) then Result:=clFuchsia
  else if (UpperCase(s)=UpperCase(pcAqua)) then    Result:=clAqua
  else if (UpperCase(s)=UpperCase(pcWhite)) then   Result:=clWhite
  else try Result:=StrToInt(s) except end;
end;

function EncodeColor(c:TColor):String;
begin
  Result:=IntToStr(c);
  if (c=clBlack) then     Result:=pcBlack;
  if (c=clMaroon) then  Result:=pcMaroon;
  if (c=clGreen) then   Result:=pcGreen;
  if (c=clOlive) then   Result:=pcOlive;
  if (c=clNavy) then    Result:=pcNavy;
  if (c=clPurple) then  Result:=pcPurple;
  if (c=clTeal) then    Result:=pcTeal;
  if (c=clGray) then    Result:=pcGray;
  if (c=clSilver) then  Result:=pcSilver;
  if (c=clRed) then     Result:=pcRed;
  if (c=clLime) then    Result:=pcLime;
  if (c=clYellow) then  Result:=pcYellow;
  if (c=clBlue) then    Result:=pcBlue;
  if (c=clFuchsia) then Result:=pcFuchsia;
  if (c=clAqua) then    Result:=pcAqua;
  if (c=clWhite) then   Result:=pcWhite;
end;

function GetStyleString(AFont:TFont):String;
begin
  Result:='';
  if fsItalic in AFont.Style then Result:=Result+'I';
  if fsBold in AFont.Style then Result:=Result+'B';
  if fsUnderline in AFont.Style then Result:=Result+'U';
  if fsStrikeout in AFont.Style then Result:=Result+'S';
end;

function Stream2String(ms:TMemoryStream):String;
  var b:Byte;
begin
  Result:='';
  ms.Position:=0;
  while (ms.Position<ms.Size) do try
    ms.Read(b,1);
    Result:=Result+IntToHex(b,2);
  except end;
end;

procedure String2Stream(const s:String; var ms:TMemoryStream);
  var b:Byte;
      st:String;
begin
  st:=s;
  ms.Clear;
  while (Length(st)>0) do try
    b:=Byte(Hex2Dec(System.Copy(st,1,2)));
    ms.Write(b,1);
    System.Delete(st,1,2);
  except end;
  ms.Position:=0;
end;

function GetPictureType(Picture : TPicture) : String;
  var g : TGraphic;
begin
  Result:='';
  if ((Picture = nil) or (Picture.Graphic = nil)) then Exit;
  g:=Picture.Graphic;
  if (g is TIcon) then Result:='Icon'
  else if (g is TMetafile) then Result:='Metafile'
{$IFDEF USE_RX}
  else if (g is TGIFImage) then Result:='GIF'
{$ENDIF}
  else if (g is TJPEGImage) then Result:='JPEG'
  else Result:='Bitmap';
end;

function LoadPictureFromStream(const PictureType : String; Picture : TPicture; Stream : TStream) : Boolean;
  var g  : TGraphic;
      pt : String;

  function TryFormat(gc : TGraphicClass) : Boolean;
  begin
    Result:=False;
    Stream.Position:=0;
    g:=gc.Create;
    try
      g.LoadFromStream(Stream);
      Picture.Assign(g);
      Result:=True;
      g.Free;
    except
      g.Free;
    end;
  end;

begin
  Result:=False;
  if ((Picture = nil) or (Stream.Size = 0)) then Exit;
  Stream.Position:=0;
  pt:=UpperCase(Trim(PictureType));
  if (pt = '') then Result:=TryFormat(TBitmap) or TryFormat(TMetafile) or TryFormat(TIcon) or TryFormat(TJPEGImage) {$IFDEF USE_RX} or TryFormat(TGIFImage) {$ENDIF}
  else if (pt = 'BITMAP') then Result:=TryFormat(TBitmap)
  else if (pt = 'METAFILE') then Result:=TryFormat(TMetafile)
  else if (pt = 'ICON') then Result:=TryFormat(TIcon)
{$IFDEF USE_RX}
  else if (pt = 'GIF') then Result:=TryFormat(TGIFImage)
{$ENDIF}
  else if (pt = 'JPEG') then Result:=TryFormat(TJPEGImage);
end;

function GetPictureStreamString(Picture : TPicture) : String;
  var ms  : TMemoryStream;
      s   : String;
      pic : TPicture;
begin
  Result := '';
  if ((Picture.Graphic=nil) or Picture.Graphic.Empty) then Exit;
  pic:=TPicture.Create;
  pic.Assign(Picture);
  ms:=TMemoryStream.Create;
  Picture.Graphic.SaveToStream(ms);
  s:=GetPictureType(pic);
  Result:=s+','+IntToStr(Integer(Picture.Graphic.Transparent))+','+Stream2String(ms);
  ms.Free;
  Picture.Assign(pic);
  pic.Free;
end;

type
  TParentControl = class(TWinControl);

procedure CopyParentImage(Control: TControl; Dest: TCanvas);
  var i, Count, X, Y, SaveIndex: Integer;
      DC: HDC;
      R, SelfR, CtlR: TRect;
begin
  if (Control = nil) or (Control.Parent = nil) then Exit;
  Count := Control.Parent.ControlCount;
  DC := Dest.Handle;
  with Control.Parent do ControlState := ControlState + [csPaintCopy];
  try
    with Control do begin
      SelfR := Bounds(Left, Top, Width, Height);
      X := -Left; Y := -Top;
    end;
    { Copy parent control image }
    SaveIndex := SaveDC(DC);
    try
      SetViewportOrgEx(DC, X, Y, nil);
      IntersectClipRect(DC, 0, 0, Control.Parent.ClientWidth, Control.Parent.ClientHeight);
      with TParentControl(Control.Parent) do begin
        Perform(WM_ERASEBKGND, DC, 0);
        PaintWindow(DC);
      end;
    finally
      RestoreDC(DC, SaveIndex);
    end;
    { Copy images of graphic controls }
    for I := 0 to Count - 1 do begin
      if Control.Parent.Controls[I] = Control then Break
      else if (Control.Parent.Controls[I] <> nil) and (Control.Parent.Controls[I] is TGraphicControl) then begin
        with TGraphicControl(Control.Parent.Controls[I]) do begin
          CtlR := Bounds(Left, Top, Width, Height);
          if Bool(IntersectRect(R, SelfR, CtlR)) and Visible then begin
            ControlState := ControlState + [csPaintCopy];
            SaveIndex := SaveDC(DC);
            try
              SaveIndex := SaveDC(DC);
              SetViewportOrgEx(DC, Left + X, Top + Y, nil);
              IntersectClipRect(DC, 0, 0, Width, Height);
              Perform(WM_PAINT, DC, 0);
            finally
              RestoreDC(DC, SaveIndex);
              ControlState := ControlState - [csPaintCopy];
            end;
          end;
        end;
      end;
    end;
finally
    with Control.Parent do ControlState := ControlState - [csPaintCopy];
  end;
end;

{ Gradient fill procedure - displays a gradient beginning with a chosen    }
{ color and ending with another chosen color. Based on TGradientFill       }
{ component source code written by Curtis White, cwhite@teleport.com.      }
procedure GradientFillRect(Canvas: TCanvas; ARect: TRect; StartColor, EndColor: TColor; Direction: TGradientFillDirection; Colors: Byte);
  var
    StartRGB: array[0..2] of Byte;    { Start RGB values }
    RGBDelta: array[0..2] of Integer; { Difference between start and end RGB values }
    ColorBand: TRect;                 { Color band rectangular coordinates }
    I, Delta: Integer;
    Brush: HBrush;
    SavePen : TPen;
begin
  if IsRectEmpty(ARect) then Exit;
  SavePen := TPen.Create;
  SavePen.Assign(Canvas.Pen);
  if Colors < 2 then begin
    Brush := CreateSolidBrush(ColorToRGB(StartColor));
    FillRect(Canvas.Handle, ARect, Brush);
    DeleteObject(Brush);
    Exit;
  end;
  StartColor := ColorToRGB(StartColor);
  EndColor := ColorToRGB(EndColor);
  case Direction of
    fdTopToBottom, fdLeftToRight: begin
      { Set the Red, Green and Blue colors }
      StartRGB[0] := GetRValue(StartColor);
      StartRGB[1] := GetGValue(StartColor);
      StartRGB[2] := GetBValue(StartColor);
      { Calculate the difference between begin and end RGB values }
      RGBDelta[0] := GetRValue(EndColor) - StartRGB[0];
      RGBDelta[1] := GetGValue(EndColor) - StartRGB[1];
      RGBDelta[2] := GetBValue(EndColor) - StartRGB[2];
    end;
    fdBottomToTop, fdRightToLeft: begin
      { Set the Red, Green and Blue colors }
      { Reverse of TopToBottom and LeftToRight directions }
      StartRGB[0] := GetRValue(EndColor);
      StartRGB[1] := GetGValue(EndColor);
      StartRGB[2] := GetBValue(EndColor);
      { Calculate the difference between begin and end RGB values }
      { Reverse of TopToBottom and LeftToRight directions }
      RGBDelta[0] := GetRValue(StartColor) - StartRGB[0];
      RGBDelta[1] := GetGValue(StartColor) - StartRGB[1];
      RGBDelta[2] := GetBValue(StartColor) - StartRGB[2];
    end;
  end; {case}
  { Calculate the color band's coordinates }
  ColorBand := ARect;
  if Direction in [fdTopToBottom, fdBottomToTop] then begin
    Colors := Max(2, Min(Colors, ARect.Bottom - ARect.Top));
    Delta := (ARect.Bottom - ARect.Top) div Colors;
  end
  else begin
    Colors := Max(2, Min(Colors, ARect.Right - ARect.Left));
    Delta := (ARect.Right - ARect.Left) div Colors;
  end;
  with Canvas.Pen do begin { Set the pen style and mode }
    Style := psSolid;
    Mode := pmCopy;
  end;
  { Perform the fill }
  if Delta > 0 then begin
    for I := 0 to Colors do begin
      case Direction of
        { Calculate the color band's top and bottom coordinates }
        fdTopToBottom, fdBottomToTop: begin
          ColorBand.Top := ARect.Top + I * Delta;
          ColorBand.Bottom := ColorBand.Top + Delta;
        end;
        { Calculate the color band's left and right coordinates }
        fdLeftToRight, fdRightToLeft: begin
          ColorBand.Left := ARect.Left + I * Delta;
          ColorBand.Right := ColorBand.Left + Delta;
        end;
      end; {case}
      { Calculate the color band's color }
      Brush := CreateSolidBrush(RGB(
        StartRGB[0] + MulDiv(I, RGBDelta[0], Colors - 1),
        StartRGB[1] + MulDiv(I, RGBDelta[1], Colors - 1),
        StartRGB[2] + MulDiv(I, RGBDelta[2], Colors - 1)));
      FillRect(Canvas.Handle, ColorBand, Brush);
      DeleteObject(Brush);
    end;
  end;
  if Direction in [fdTopToBottom, fdBottomToTop] then Delta := (ARect.Bottom - ARect.Top) mod Colors
  else Delta := (ARect.Right - ARect.Left) mod Colors;
  if Delta > 0 then begin
    case Direction of
      { Calculate the color band's top and bottom coordinates }
      fdTopToBottom, fdBottomToTop: begin
        ColorBand.Top := ARect.Bottom - Delta;
        ColorBand.Bottom := ColorBand.Top + Delta;
      end;
      { Calculate the color band's left and right coordinates }
      fdLeftToRight, fdRightToLeft: begin
        ColorBand.Left := ARect.Right - Delta;
        ColorBand.Right := ColorBand.Left + Delta;
      end;
    end; {case}
    case Direction of
      fdTopToBottom, fdLeftToRight:
        Brush := CreateSolidBrush(EndColor);
      else {fdBottomToTop, fdRightToLeft }
        Brush := CreateSolidBrush(StartColor);
    end;
    FillRect(Canvas.Handle, ColorBand, Brush);
    DeleteObject(Brush);
  end;
  Canvas.Pen.Assign(SavePen);
  SavePen.Free;
end;

procedure GradientFillRgn(Canvas: TCanvas; ARect : TRect; ARgn: HRGN; StartColor, EndColor: TColor; Direction: TGradientFillDirection; Colors: Byte);
  var rgn : HRGN;
      i   : Integer;
      FWasClippingRgn : Boolean;
begin
  rgn := CreateRectRgn(0,0,0,0);
  i := GetClipRgn(Canvas.Handle,rgn); // Old clipping region
  FWasClippingRgn := (i = 1);
  SelectClipRgn(Canvas.Handle,ARgn);
  GradientFillRect(Canvas,ARect,StartColor,EndColor,Direction,Colors);
  if FWasClippingRgn then SelectClipRgn(Canvas.Handle,rgn) else SelectClipRgn(Canvas.Handle,0);
  DeleteObject(rgn);
end;

initialization
  iScreenWidth := GetSystemMetrics(SM_CXSCREEN);
  iScreenHeight := GetSystemMetrics(SM_CYSCREEN);
end.